Package: wru
Title: Who are You? Bayesian Prediction of Racial Category Using
        Surname, First Name, Middle Name, and Geolocation
Version: 3.1.0
Date: 2026-04-30
Authors@R: c(
    person("Kabir", "Khanna", , "kabirkhanna@gmail.com", role = "aut"),
    person("Brandon", "Bertelsen", , "brandon@bertelsen.ca", role = c("aut", "cre")),
    person("Santiago", "Olivella", , "olivella@unc.edu", role = "aut"),
    person("Evan", "Rosenman", , "etrrosenman@gmail.com", role = "aut"),
    person("Alexander", "Rossell Hayes", , "alexander@rossellhayes.com", role = "aut"),
    person("Kosuke", "Imai", , "imai@harvard.edu", role = "aut"),
    person("Noah", "Dasanaike", , "noahdasanaike@g.harvard.edu", role = "aut")
  )
Description: Predicts individual race/ethnicity using surname, first name,
    middle name, geolocation, and other attributes, such as gender and
    age. The method utilizes Bayes' Rule (with optional measurement error
    correction) to compute the posterior probability of each racial
    category for any given individual. The package implements methods
    described in Imai and Khanna (2016) "Improving Ecological Inference by
    Predicting Individual Ethnicity from Voter Registration Records"
    Political Analysis <DOI:10.1093/pan/mpw001> and Imai, Olivella, and
    Rosenman (2022) "Addressing census data problems in race imputation
    via fully Bayesian Improved Surname Geocoding and name supplements"
    <DOI:10.1126/sciadv.adc9824>.  The package also incorporates the data
    described in Rosenman, Olivella, and Imai (2023) "Race and ethnicity
    data for first, middle, and surnames"
    <DOI:10.1038/s41597-023-02202-2>.
License: GPL (>= 3)
URL: https://github.com/kosukeimai/wru
BugReports: https://github.com/kosukeimai/wru/issues
Depends: R (>= 4.1.0), utils
Imports: cli, dplyr, furrr, piggyback (>= 0.1.4), PL94171, purrr, Rcpp,
        rlang
Suggests: covr, future, reticulate, testthat (>= 3.0.0), tidycensus,
        tidyr
LinkingTo: Rcpp, RcppArmadillo
Config/testthat/edition: 3
Encoding: UTF-8
LazyData: yes
LazyDataCompression: xz
LazyLoad: yes
RoxygenNote: 7.3.1
NeedsCompilation: yes
Packaged: 2026-05-07 03:49:58 UTC; b
Author: Kabir Khanna [aut],
  Brandon Bertelsen [aut, cre],
  Santiago Olivella [aut],
  Evan Rosenman [aut],
  Alexander Rossell Hayes [aut],
  Kosuke Imai [aut],
  Noah Dasanaike [aut]
Maintainer: Brandon Bertelsen <brandon@bertelsen.ca>
Repository: CRAN
Date/Publication: 2026-05-07 11:40:12 UTC
Built: R 4.4.3; x86_64-apple-darwin20; 2026-05-07 12:55:10 UTC; unix
Archs: wru.so.dSYM

# Generated by roxygen2: do not edit by hand

export(as_fips_code)
export(as_state_abbreviation)
export(format_legacy_data)
export(get_census_data)
export(predict_race)
export(setup_ebisg)
import(PL94171)
importFrom(Rcpp,evalCpp)
importFrom(dplyr,coalesce)
importFrom(dplyr,pull)
importFrom(furrr,future_map_dfr)
importFrom(piggyback,pb_download)
importFrom(purrr,map_dfr)
importFrom(rlang,"%||%")
importFrom(stats,rmultinom)
importFrom(utils,setTxtProgressBar)
importFrom(utils,txtProgressBar)
useDynLib(wru, .registration=TRUE)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# wru: Who Are You? Bayesian Prediction of Racial Category Using Surname and Geolocation <img src="man/figures/logo.png?raw=TRUE" align="right" height="138" alt="Package logo" />

[![R-CMD-check](https://github.com/kosukeimai/wru/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/kosukeimai/wru/actions/workflows/R-CMD-check.yaml)
[![CRAN_Status_Badge](https://www.r-pkg.org/badges/version-last-release/wru)](https://cran.r-project.org/package=wru)
![CRAN downloads](http://cranlogs.r-pkg.org/badges/grand-total/wru)

This R package implements the methods proposed in Imai, K. and Khanna,
K. (2016). “[Improving Ecological Inference by Predicting Individual
Ethnicity from Voter Registration
Record](http://imai.princeton.edu/research/race.html).” Political
Analysis, Vol. 24, No. 2 (Spring), pp. 263-272. [doi:
10.1093/pan/mpw001](https://dx.doi.org/10.1093/pan/mpw001).

## Installation

You can install the released version of **wru** from
[CRAN](https://cran.r-project.org/package=wru) with:

``` r
install.packages("wru")
```

Or you can install the development version of **wru** from
[GitHub](https://github.com/kosukeimai/wru) with:

``` r
# install.packages("pak")
pak::pkg_install("kosukeimai/wru")
```

## Using wru

Here is a simple example that predicts the race/ethnicity of voters
based only on their surnames.

``` r
library(wru)
future::plan(future::multisession)
predict_race(voter.file = voters, surname.only = TRUE)
```

The above produces the following output, where the last five columns are
probabilistic race/ethnicity predictions (e.g., `pred.his` is the
probability of being Hispanic/Latino):

     VoterID    surname state CD county  tract block age sex party PID place    pred.whi    pred.bla     pred.his    pred.asi    pred.oth
           1     Khanna    NJ 12    021 004000  3001  29   0   Ind   0 74000 0.045110474 0.003067623 0.0068522723 0.860411906 0.084557725
           2       Imai    NJ 12    021 004501  1025  40   0   Dem   1 60900 0.052645440 0.001334812 0.0558160072 0.719376581 0.170827160
           3     Rivera    NY 12    061 004800  6001  33   0   Rep   2 51000 0.043285692 0.008204605 0.9136195794 0.024316883 0.010573240
           4    Fifield    NJ 12    021 004501  1025  27   0   Dem   1 60900 0.895405704 0.001911388 0.0337464844 0.011079323 0.057857101
           5       Zhou    NJ 12    021 004501  1025  28   1   Dem   1 60900 0.006572555 0.001298962 0.0005388581 0.982365594 0.009224032
           6   Ratkovic    NJ 12    021 004000  1025  35   0   Ind   0 60900 0.861236727 0.008212824 0.0095395642 0.011334635 0.109676251
           7    Johnson    NY  9    061 014900  4000  25   0   Dem   1 51000 0.543815322 0.344128607 0.0272403940 0.007405765 0.077409913
           8      Lopez    NJ 12    021 004501  1025  33   0   Rep   2 60900 0.038939877 0.004920643 0.9318797791 0.012154125 0.012105576
           9 Wantchekon    NJ 12    021 004501  1025  50   0   Rep   2 60900 0.330697188 0.194700665 0.4042849478 0.021379541 0.048937658
          10      Morse    DC  0    001 001301  3005  29   1   Rep   2 50000 0.866360147 0.044429853 0.0246568086 0.010219712 0.054333479

### Using geolocation

In order to predict race/ethnicity based on surnames *and* geolocation,
a user needs to provide a valid U.S. Census API key to access the census
statistics. You can request a U.S. Census API key from [the U.S. Census
API key signup page](http://api.census.gov/data/key_signup.html). Once
you have an API key, you can use the package to download relevant Census
geographic data on demand and condition race/ethnicity predictions on
geolocation (county, tract, block, or place).

First, you should save your census key to your `.Rprofile` or
`.Renviron`. Below is an example procedure:

    usethis::edit_r_environ()
    # Edit the file with the following:
    CENSUS_API_KEY=YourKey
    # Save and close the file
    # Restart your R session

The following example predicts the race/ethnicity of voters based on
their surnames, census tract of residence (`census.geo = "tract"`), and
party registration (`party = "PID"`). Note that a valid API key must be
stored in a `CENSUS_API_KEY` environment variable or provided with the
`census.key` argument in order for the function to download the relevant
tract-level data.

``` r
library(wru)
predict_race(voter.file = voters, census.geo = "tract", party = "PID")
```

     VoterID    surname state CD county  tract block age sex party PID place    pred.whi     pred.bla     pred.his   pred.asi    pred.oth
           1     Khanna    NJ 12    021 004000  3001  29   0   Ind   0 74000 0.021711601 0.0009552652 2.826779e-03 0.93364592 0.040860431
           2       Imai    NJ 12    021 004501  1025  40   0   Dem   1 60900 0.015364583 0.0002320815 9.020240e-03 0.90245186 0.072931231
           3     Rivera    NY 12    061 004800  6001  33   0   Rep   2 51000 0.092415538 0.0047099965 7.860806e-01 0.09924761 0.017546300
           4    Fifield    NJ 12    021 004501  1025  27   0   Dem   1 60900 0.854810748 0.0010870744 1.783931e-02 0.04546436 0.080798514
           5       Zhou    NJ 12    021 004501  1025  28   1   Dem   1 60900 0.001548762 0.0001823506 7.031116e-05 0.99501901 0.003179566
           6   Ratkovic    NJ 12    021 004000  1025  35   0   Ind   0 60900 0.852374629 0.0052590592 8.092435e-03 0.02529163 0.108982246
           7    Johnson    NY  9    061 014900  4000  25   0   Dem   1 51000 0.831282563 0.0613242553 1.059715e-02 0.01602557 0.080770461
           8      Lopez    NJ 12    021 004501  1025  33   0   Rep   2 60900 0.062022518 0.0046691402 8.218906e-01 0.08321206 0.028205698
           9 Wantchekon    NJ 12    021 004501  1025  50   0   Rep   2 60900 0.396500218 0.1390722877 2.684107e-01 0.11018413 0.085832686
          10      Morse    DC  0    001 001301  3005  29   1   Rep   2 50000 0.861168219 0.0498449102 1.131154e-02 0.01633532 0.061340015

In `predict_race()`, the `census.geo` options are “county”, “tract”,
“block” and “place”. Here is an example of prediction based on census
statistics collected at the level of “place”:

``` r
predict_race(voter.file = voters, census.geo = "place", party = "PID")
```

     VoterID    surname state CD county  tract block age sex party PID place    pred.whi     pred.bla     pred.his   pred.asi    pred.oth
           1     Khanna    NJ 12    021 004000  3001  29   0   Ind   0 74000 0.042146148 0.0620484276 9.502254e-02 0.55109761 0.249685278
           2       Imai    NJ 12    021 004501  1025  40   0   Dem   1 60900 0.018140322 0.0002204255 1.026018e-02 0.90710894 0.064270133
           3     Rivera    NY 12    061 004800  6001  33   0   Rep   2 51000 0.015528660 0.0092292671 9.266893e-01 0.04182290 0.006729825
           4    Fifield    NJ 12    021 004501  1025  27   0   Dem   1 60900 0.879537890 0.0008997896 1.768379e-02 0.03982601 0.062052518
           5       Zhou    NJ 12    021 004501  1025  28   1   Dem   1 60900 0.001819394 0.0001723242 7.957542e-05 0.99514078 0.002787926
           6   Ratkovic    NJ 12    021 004000  1025  35   0   Ind   0 60900 0.834942701 0.0038157857 4.933723e-03 0.04021245 0.116095337
           7    Johnson    NY  9    061 014900  4000  25   0   Dem   1 51000 0.290386744 0.5761904554 4.112613e-02 0.01895885 0.073337820
           8      Lopez    NJ 12    021 004501  1025  33   0   Rep   2 60900 0.065321588 0.0039558641 8.339387e-01 0.07461133 0.022172551
           9 Wantchekon    NJ 12    021 004501  1025  50   0   Rep   2 60900 0.428723819 0.1209683869 2.796062e-01 0.10142953 0.069272098
          10      Morse    DC  0    001 001301  3005  29   1   Rep   2 50000 0.716211008 0.1899554127 1.867133e-02 0.01025241 0.064909839

### Using surname embeddings (eBISG)

About 10% of US voters have surnames that do not appear in the Census
surname list. Standard BISG has no name-specific information for these
voters and falls back on a name-independent imputation (the column means
of the matched-surname probabilities), which gives every unmatched name
the same prediction. The `eBISG` model fills this gap: it embeds each
unmatched surname with a pre-trained sentence-transformer (E5-Large by
default), then runs a small neural network trained on Census 2020 data
to predict race-conditional probabilities for that specific name. For
names that *are* in the Census list, predictions are identical to
standard BISG.

`eBISG` is opt-in. It depends on Python with `sentence-transformers` and
`torch`; the first call additionally downloads about 17 MB of
pre-trained MLP weights from the package’s GitHub release.

Run this once per machine to set up the environment:

``` r
library(wru)
setup_ebisg()
# Installs sentence-transformers + torch into the 'r-ebisg' virtualenv
# and downloads the MLP weight files.
```

Now consider a small voter file with one Census-listed surname (`Smith`)
and two surnames that are not in the Census list (`Rajasekhar`, a Tamil
name, and `Schwarzenegger`, an Austrian-German name):

``` r
demo <- data.frame(
  surname = c("Smith", "Rajasekhar", "Schwarzenegger"),
  state   = "NJ",
  stringsAsFactors = FALSE
)

predict_race(voter.file = demo, surname.only = TRUE, model = "BISG")
```

             surname state    pred.whi    pred.bla    pred.his    pred.asi    pred.oth
    1          Smith    NJ 0.663005990 0.232872914 0.028090777 0.006953018 0.069077301
    2     Rajasekhar    NJ 0.663005990 0.232872914 0.028090777 0.006953018 0.069077301
    3 Schwarzenegger    NJ 0.663005990 0.232872914 0.028090777 0.006953018 0.069077301

The two unmatched surnames receive the *same* prediction; standard BISG
cannot distinguish them. With eBISG:

``` r
predict_race(voter.file = demo, surname.only = TRUE, model = "eBISG")
```

             surname state    pred.whi    pred.bla    pred.his    pred.asi    pred.oth
    1          Smith    NJ 0.663005990 0.232872914 0.028090777 0.006953018 0.069077301
    2     Rajasekhar    NJ 0.528512168 0.020206366 0.032476816 0.411517171 0.007287581
    3 Schwarzenegger    NJ 0.987577004 0.004130978 0.005649397 0.000511518 0.002131135

`Smith` (matched) is unchanged. `Rajasekhar` is now correctly tilted
toward Asian (~41%) and `Schwarzenegger` toward White (~99%), reflecting
the name-specific signal the embedding model recovers.

eBISG also composes with geolocation just like standard BISG:

``` r
predict_race(voter.file = voters, census.geo = "tract", model = "eBISG")
```

To use a different sentence-transformer (for example, one of the newer
top entries on the [MTEB
leaderboard](https://huggingface.co/spaces/mteb/leaderboard)), pass a
list to `ebisg.model` with the HuggingFace model ID and a path to a
matching MLP checkpoint trained against that encoder:

``` r
predict_race(
  voter.file = voters, surname.only = TRUE, model = "eBISG",
  ebisg.model = list(
    transformer  = "your-org/your-encoder",
    dim          = 768L,
    surname_mlp  = "/path/to/your-surname-mlp.pt"
  )
)
```

For methodological details, see Dasanaike and Imai (2026),
“Embedding-Supplemented Bayesian Improved Surname Geocoding.”

### Downloading census data

It is also possible to pre-download Census geographic data, which can
save time when running `predict_race()`. The example dataset `voters`
includes people in DC, NJ, and NY. The following example subsets voters
in DC and NJ, and then uses `get_census_data()` to download census
geographic data in these two states (a valid API key must be stored in a
`CENSUS_API_KEY` environment variable or provided with the `key`
argument). Census data is assigned to an object named `census.dc.nj`.
The `predict_race()` statement predicts the race/ethnicity of voters in
DC and NJ using the pre-downloaded census data
(`census.data = census.dc.nj`). This example conditions race/ethnicity
predictions on voters’ surnames, block of residence
(`census.geo = "block"`), age (`age = TRUE`), and party registration
(`party = "PID"`).

Please note that the input parameters `age` and `sex` must have the same
values in `get_census_data()` and `predict_race()`, i.e., `TRUE` in both
or `FALSE` in both. In this case, predictions are conditioned on age but
not sex, so `age = TRUE` and `sex = FALSE` in both the
`get_census_data()` and `predict_race()` statements.

``` r
library(wru)
voters.dc.nj <- voters[voters$state %in% c("DC", "NJ"), ]
census.dc.nj <- get_census_data(state = c("DC", "NJ"), age = TRUE, sex = FALSE)
predict_race(voter.file = voters.dc.nj, census.geo = "block", census.data = census.dc.nj, age = TRUE, sex = FALSE, party = "PID")
```

This produces the same result as the following statement, which
downloads census data during evaluation rather than using pre-downloaded
data:

``` r
predict_race(voter.file = voters.dc.nj, census.geo = "block", age = TRUE, sex = FALSE, party = "PID")
```

Using pre-downloaded Census data may be useful for the following
reasons:

- You can save a lot of time in future runs of `predict_race()` if the
  relevant census data has already been saved;
- The machines used to run `predict_race()` may not have internet
  access;
- You can obtain timely snapshots of census geographic data that match
  your voter file.

Downloading data using `get_census_data()` may take a long time,
especially in large states or when using small geographic levels. If
block-level census data is not required, downloading census data at the
tract level will save time. Similarly, if tract-level data is not
required, county-level data may be specified in order to save time.

``` r
library(wru)
voters.dc.nj <- voters[voters$state %in% c("DC", "NJ"), ]
census.dc.nj2 <- get_census_data(state = c("DC", "NJ"), age = TRUE, sex = FALSE, census.geo = "tract")  
predict_race(voter.file = voters.dc.nj, census.geo = "tract", census.data = census.dc.nj2, party = "PID", age = TRUE, sex = FALSE)
predict_race(voter.file = voters.dc.nj, census.geo = "county", census.data = census.dc.nj2, age = TRUE, sex = FALSE)  # Pr(Race | Surname, County)
predict_race(voter.file = voters.dc.nj, census.geo = "tract", census.data = census.dc.nj2, age = TRUE, sex = FALSE)  # Pr(Race | Surname, Tract)
predict_race(voter.file = voters.dc.nj, census.geo = "county", census.data = census.dc.nj2, party = "PID", age = TRUE, sex = FALSE)  # Pr(Race | Surname, County, Party)
predict_race(voter.file = voters.dc.nj, census.geo = "tract", census.data = census.dc.nj2, party = "PID", age = TRUE, sex = FALSE)  # Pr(Race | Surname, Tract, Party)
```

#### Interact directly with the Census API

You can use `census_geo_api()` to manually construct a census object.
The example below creates a census object with county-level and
tract-level data in DC and NJ, while avoiding downloading block-level
data. Note that the `state` argument requires a vector of two-letter
state abbreviations.

``` r
census.dc.nj3 = list()

county.dc <- census_geo_api(state = "DC", geo = "county", age = TRUE, sex = FALSE)
tract.dc <- census_geo_api(state = "DC", geo = "tract", age = TRUE, sex = FALSE)
census.dc.nj3[["DC"]] <- list(state = "DC", county = county.dc, tract = tract.dc, age = TRUE, sex = FALSE)

tract.nj <- census_geo_api(state = "NJ", geo = "tract", age = TRUE, sex = FALSE)
county.nj <- census_geo_api(state = "NJ", geo = "county", age = TRUE, sex = FALSE)
census.dc.nj3[["NJ"]] <- list(state = "NJ", county = county.nj, tract = tract.nj, age = TRUE, sex = FALSE)
```

Note: The age and sex parameters must be consistent when creating the
Census object and using that Census object in the predict_race function.
If one of these parameters is TRUE in the Census object, it must also be
TRUE in the predict_race function.

After saving the data in censusObj2 above, we can condition
race/ethnicity predictions on different combinations of input variables,
without having to re-download the relevant Census data.

``` r
predict_race(voter.file = voters.dc.nj, census.geo = "county", census.data = census.dc.nj3, age = TRUE, sex = FALSE)  # Pr(Race | Surname, County)
predict_race(voter.file = voters.dc.nj, census.geo = "tract", census.data = census.dc.nj3, age = TRUE, sex = FALSE)  # Pr(Race | Surname, Tract)
predict_race(voter.file = voters.dc.nj, census.geo = "county", census.data = census.dc.nj3, party = "PID", age = TRUE, sex = FALSE)  # Pr(Race | Surname, County, Party)
predict_race(voter.file = voters.dc.nj, census.geo = "tract", census.data = census.dc.nj3, party = "PID", age = TRUE, sex = FALSE)  # Pr(Race | Surname, Tract, Party)
```

### Parallelization

For larger scale imputations, garbage collection can become a problem
and your machine(s) can quickly run out of memory (RAM). We recommended
using the `future.callr::callr` plan instead of `future::multisession`.
The `callr` plan instantiates a new session at every iteration of your
parallel loop or map. Although this has the negative effect of creating
more overhead, it also clears sticky memory elements that can grow to
eventual system failure when using `multisession`. You end up with a
process that is more stable, but slightly slower.

``` r
library(wru)
future::plan(future.callr::callr)
# ...
```

## Census Data

This package uses the Census Bureau Data API but is not endorsed or
certified by the Census Bureau.

U.S. Census Bureau (2021, October 8). Decennial Census API. Census.gov.
Retrieved from
<https://www.census.gov/data/developers/data-sets/decennial-census.html>

## A related song

[![Thumbnail of the music video for “Who Are You” by The
Who](https://img.youtube.com/vi/PNbBDrceCy8/maxresdefault.jpg)](https://www.youtube.com/watch?v=PNbBDrceCy8)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# wru: Who Are You? Bayesian Prediction of Racial Category Using Surname and Geolocation <img src="man/figures/logo.png?raw=TRUE" align="right" height="138" alt="Package logo" />

[![R-CMD-check](https://github.com/kosukeimai/wru/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/kosukeimai/wru/actions/workflows/R-CMD-check.yaml)
[![CRAN_Status_Badge](https://www.r-pkg.org/badges/version-last-release/wru)](https://cran.r-project.org/package=wru)
![CRAN downloads](http://cranlogs.r-pkg.org/badges/grand-total/wru)

This R package implements the methods proposed in Imai, K. and Khanna,
K. (2016). “[Improving Ecological Inference by Predicting Individual
Ethnicity from Voter Registration
Record](http://imai.princeton.edu/research/race.html).” Political
Analysis, Vol. 24, No. 2 (Spring), pp. 263-272. [doi:
10.1093/pan/mpw001](https://dx.doi.org/10.1093/pan/mpw001).

## Installation

You can install the released version of **wru** from
[CRAN](https://cran.r-project.org/package=wru) with:

``` r
install.packages("wru")
```

Or you can install the development version of **wru** from
[GitHub](https://github.com/kosukeimai/wru) with:

``` r
# install.packages("pak")
pak::pkg_install("kosukeimai/wru")
```

## Using wru

Here is a simple example that predicts the race/ethnicity of voters
based only on their surnames.

``` r
library(wru)
future::plan(future::multisession)
predict_race(voter.file = voters, surname.only = TRUE)
```

The above produces the following output, where the last five columns are
probabilistic race/ethnicity predictions (e.g., `pred.his` is the
probability of being Hispanic/Latino):

     VoterID    surname state CD county  tract block age sex party PID place    pred.whi    pred.bla     pred.his    pred.asi    pred.oth
           1     Khanna    NJ 12    021 004000  3001  29   0   Ind   0 74000 0.045110474 0.003067623 0.0068522723 0.860411906 0.084557725
           2       Imai    NJ 12    021 004501  1025  40   0   Dem   1 60900 0.052645440 0.001334812 0.0558160072 0.719376581 0.170827160
           3     Rivera    NY 12    061 004800  6001  33   0   Rep   2 51000 0.043285692 0.008204605 0.9136195794 0.024316883 0.010573240
           4    Fifield    NJ 12    021 004501  1025  27   0   Dem   1 60900 0.895405704 0.001911388 0.0337464844 0.011079323 0.057857101
           5       Zhou    NJ 12    021 004501  1025  28   1   Dem   1 60900 0.006572555 0.001298962 0.0005388581 0.982365594 0.009224032
           6   Ratkovic    NJ 12    021 004000  1025  35   0   Ind   0 60900 0.861236727 0.008212824 0.0095395642 0.011334635 0.109676251
           7    Johnson    NY  9    061 014900  4000  25   0   Dem   1 51000 0.543815322 0.344128607 0.0272403940 0.007405765 0.077409913
           8      Lopez    NJ 12    021 004501  1025  33   0   Rep   2 60900 0.038939877 0.004920643 0.9318797791 0.012154125 0.012105576
           9 Wantchekon    NJ 12    021 004501  1025  50   0   Rep   2 60900 0.330697188 0.194700665 0.4042849478 0.021379541 0.048937658
          10      Morse    DC  0    001 001301  3005  29   1   Rep   2 50000 0.866360147 0.044429853 0.0246568086 0.010219712 0.054333479

### Using geolocation

In order to predict race/ethnicity based on surnames *and* geolocation,
a user needs to provide a valid U.S. Census API key to access the census
statistics. You can request a U.S. Census API key from [the U.S. Census
API key signup page](http://api.census.gov/data/key_signup.html). Once
you have an API key, you can use the package to download relevant Census
geographic data on demand and condition race/ethnicity predictions on
geolocation (county, tract, block, or place).

First, you should save your census key to your `.Rprofile` or
`.Renviron`. Below is an example procedure:

    usethis::edit_r_environ()
    # Edit the file with the following:
    CENSUS_API_KEY=YourKey
    # Save and close the file
    # Restart your R session

The following example predicts the race/ethnicity of voters based on
their surnames, census tract of residence (`census.geo = "tract"`), and
party registration (`party = "PID"`). Note that a valid API key must be
stored in a `CENSUS_API_KEY` environment variable or provided with the
`census.key` argument in order for the function to download the relevant
tract-level data.

``` r
library(wru)
predict_race(voter.file = voters, census.geo = "tract", party = "PID")
```

     VoterID    surname state CD county  tract block age sex party PID place    pred.whi     pred.bla     pred.his   pred.asi    pred.oth
           1     Khanna    NJ 12    021 004000  3001  29   0   Ind   0 74000 0.021711601 0.0009552652 2.826779e-03 0.93364592 0.040860431
           2       Imai    NJ 12    021 004501  1025  40   0   Dem   1 60900 0.015364583 0.0002320815 9.020240e-03 0.90245186 0.072931231
           3     Rivera    NY 12    061 004800  6001  33   0   Rep   2 51000 0.092415538 0.0047099965 7.860806e-01 0.09924761 0.017546300
           4    Fifield    NJ 12    021 004501  1025  27   0   Dem   1 60900 0.854810748 0.0010870744 1.783931e-02 0.04546436 0.080798514
           5       Zhou    NJ 12    021 004501  1025  28   1   Dem   1 60900 0.001548762 0.0001823506 7.031116e-05 0.99501901 0.003179566
           6   Ratkovic    NJ 12    021 004000  1025  35   0   Ind   0 60900 0.852374629 0.0052590592 8.092435e-03 0.02529163 0.108982246
           7    Johnson    NY  9    061 014900  4000  25   0   Dem   1 51000 0.831282563 0.0613242553 1.059715e-02 0.01602557 0.080770461
           8      Lopez    NJ 12    021 004501  1025  33   0   Rep   2 60900 0.062022518 0.0046691402 8.218906e-01 0.08321206 0.028205698
           9 Wantchekon    NJ 12    021 004501  1025  50   0   Rep   2 60900 0.396500218 0.1390722877 2.684107e-01 0.11018413 0.085832686
          10      Morse    DC  0    001 001301  3005  29   1   Rep   2 50000 0.861168219 0.0498449102 1.131154e-02 0.01633532 0.061340015

In `predict_race()`, the `census.geo` options are “county”, “tract”,
“block” and “place”. Here is an example of prediction based on census
statistics collected at the level of “place”:

``` r
predict_race(voter.file = voters, census.geo = "place", party = "PID")
```

     VoterID    surname state CD county  tract block age sex party PID place    pred.whi     pred.bla     pred.his   pred.asi    pred.oth
           1     Khanna    NJ 12    021 004000  3001  29   0   Ind   0 74000 0.042146148 0.0620484276 9.502254e-02 0.55109761 0.249685278
           2       Imai    NJ 12    021 004501  1025  40   0   Dem   1 60900 0.018140322 0.0002204255 1.026018e-02 0.90710894 0.064270133
           3     Rivera    NY 12    061 004800  6001  33   0   Rep   2 51000 0.015528660 0.0092292671 9.266893e-01 0.04182290 0.006729825
           4    Fifield    NJ 12    021 004501  1025  27   0   Dem   1 60900 0.879537890 0.0008997896 1.768379e-02 0.03982601 0.062052518
           5       Zhou    NJ 12    021 004501  1025  28   1   Dem   1 60900 0.001819394 0.0001723242 7.957542e-05 0.99514078 0.002787926
           6   Ratkovic    NJ 12    021 004000  1025  35   0   Ind   0 60900 0.834942701 0.0038157857 4.933723e-03 0.04021245 0.116095337
           7    Johnson    NY  9    061 014900  4000  25   0   Dem   1 51000 0.290386744 0.5761904554 4.112613e-02 0.01895885 0.073337820
           8      Lopez    NJ 12    021 004501  1025  33   0   Rep   2 60900 0.065321588 0.0039558641 8.339387e-01 0.07461133 0.022172551
           9 Wantchekon    NJ 12    021 004501  1025  50   0   Rep   2 60900 0.428723819 0.1209683869 2.796062e-01 0.10142953 0.069272098
          10      Morse    DC  0    001 001301  3005  29   1   Rep   2 50000 0.716211008 0.1899554127 1.867133e-02 0.01025241 0.064909839

### Downloading census data

It is also possible to pre-download Census geographic data, which can
save time when running `predict_race()`. The example dataset `voters`
includes people in DC, NJ, and NY. The following example subsets voters
in DC and NJ, and then uses `get_census_data()` to download census
geographic data in these two states (a valid API key must be stored in a
`CENSUS_API_KEY` environment variable or provided with the `key`
argument). Census data is assigned to an object named `census.dc.nj`.
The `predict_race()` statement predicts the race/ethnicity of voters in
DC and NJ using the pre-downloaded census data
(`census.data = census.dc.nj`). This example conditions race/ethnicity
predictions on voters’ surnames, block of residence
(`census.geo = "block"`), age (`age = TRUE`), and party registration
(`party = "PID"`).

Please note that the input parameters `age` and `sex` must have the same
values in `get_census_data()` and `predict_race()`, i.e., `TRUE` in both
or `FALSE` in both. In this case, predictions are conditioned on age but
not sex, so `age = TRUE` and `sex = FALSE` in both the
`get_census_data()` and `predict_race()` statements.

``` r
library(wru)
voters.dc.nj <- voters[voters$state %in% c("DC", "NJ"), ]
census.dc.nj <- get_census_data(state = c("DC", "NJ"), age = TRUE, sex = FALSE)
predict_race(voter.file = voters.dc.nj, census.geo = "block", census.data = census.dc.nj, age = TRUE, sex = FALSE, party = "PID")
```

This produces the same result as the following statement, which
downloads census data during evaluation rather than using pre-downloaded
data:

``` r
predict_race(voter.file = voters.dc.nj, census.geo = "block", age = TRUE, sex = FALSE, party = "PID")
```

Using pre-downloaded Census data may be useful for the following
reasons:

- You can save a lot of time in future runs of `predict_race()` if the
  relevant census data has already been saved;
- The machines used to run `predict_race()` may not have internet
  access;
- You can obtain timely snapshots of census geographic data that match
  your voter file.

Downloading data using `get_census_data()` may take a long time,
especially in large states or when using small geographic levels. If
block-level census data is not required, downloading census data at the
tract level will save time. Similarly, if tract-level data is not
required, county-level data may be specified in order to save time.

``` r
library(wru)
voters.dc.nj <- voters[voters$state %in% c("DC", "NJ"), ]
census.dc.nj2 <- get_census_data(state = c("DC", "NJ"), age = TRUE, sex = FALSE, census.geo = "tract")  
predict_race(voter.file = voters.dc.nj, census.geo = "tract", census.data = census.dc.nj2, party = "PID", age = TRUE, sex = FALSE)
predict_race(voter.file = voters.dc.nj, census.geo = "county", census.data = census.dc.nj2, age = TRUE, sex = FALSE)  # Pr(Race | Surname, County)
predict_race(voter.file = voters.dc.nj, census.geo = "tract", census.data = census.dc.nj2, age = TRUE, sex = FALSE)  # Pr(Race | Surname, Tract)
predict_race(voter.file = voters.dc.nj, census.geo = "county", census.data = census.dc.nj2, party = "PID", age = TRUE, sex = FALSE)  # Pr(Race | Surname, County, Party)
predict_race(voter.file = voters.dc.nj, census.geo = "tract", census.data = census.dc.nj2, party = "PID", age = TRUE, sex = FALSE)  # Pr(Race | Surname, Tract, Party)
```

#### Interact directly with the Census API

You can use `census_geo_api()` to manually construct a census object.
The example below creates a census object with county-level and
tract-level data in DC and NJ, while avoiding downloading block-level
data. Note that the `state` argument requires a vector of two-letter
state abbreviations.

``` r
census.dc.nj3 = list()

county.dc <- census_geo_api(state = "DC", geo = "county", age = TRUE, sex = FALSE)
tract.dc <- census_geo_api(state = "DC", geo = "tract", age = TRUE, sex = FALSE)
census.dc.nj3[["DC"]] <- list(state = "DC", county = county.dc, tract = tract.dc, age = TRUE, sex = FALSE)

tract.nj <- census_geo_api(state = "NJ", geo = "tract", age = TRUE, sex = FALSE)
county.nj <- census_geo_api(state = "NJ", geo = "county", age = TRUE, sex = FALSE)
census.dc.nj3[["NJ"]] <- list(state = "NJ", county = county.nj, tract = tract.nj, age = TRUE, sex = FALSE)
```

Note: The age and sex parameters must be consistent when creating the
Census object and using that Census object in the predict_race function.
If one of these parameters is TRUE in the Census object, it must also be
TRUE in the predict_race function.

After saving the data in censusObj2 above, we can condition
race/ethnicity predictions on different combinations of input variables,
without having to re-download the relevant Census data.

``` r
predict_race(voter.file = voters.dc.nj, census.geo = "county", census.data = census.dc.nj3, age = TRUE, sex = FALSE)  # Pr(Race | Surname, County)
predict_race(voter.file = voters.dc.nj, census.geo = "tract", census.data = census.dc.nj3, age = TRUE, sex = FALSE)  # Pr(Race | Surname, Tract)
predict_race(voter.file = voters.dc.nj, census.geo = "county", census.data = census.dc.nj3, party = "PID", age = TRUE, sex = FALSE)  # Pr(Race | Surname, County, Party)
predict_race(voter.file = voters.dc.nj, census.geo = "tract", census.data = census.dc.nj3, party = "PID", age = TRUE, sex = FALSE)  # Pr(Race | Surname, Tract, Party)
```

### Parallelization

For larger scale imputations, garbage collection can become a problem
and your machine(s) can quickly run out of memory (RAM). We recommended
using the `future.callr::callr` plan instead of `future::multisession`.
The `callr` plan instantiates a new session at every iteration of your
parallel loop or map. Although this has the negative effect of creating
more overhead, it also clears sticky memory elements that can grow to
eventual system failure when using `multisession`. You end up with a
process that is more stable, but slightly slower.

``` r
library(wru)
future::plan(future.callr::callr)
# ...
```

## Census Data

This package uses the Census Bureau Data API but is not endorsed or
certified by the Census Bureau.

U.S. Census Bureau (2021, October 8). Decennial Census API. Census.gov.
Retrieved from
<https://www.census.gov/data/developers/data-sets/decennial-census.html>

## A related song

[![Thumbnail of the music video for “Who Are You” by The
Who](https://img.youtube.com/vi/PNbBDrceCy8/maxresdefault.jpg)](https://www.youtube.com/watch?v=PNbBDrceCy8)

# @staticimports pkg:stringstatic
#   str_pad

#' Census geo API helper functions
#'
#' @inheritParams census_geo_api
#'
#' @return
#' \describe{
#'   \item{`census_geo_api_names()`}{
#'     A named list of [character] vectors whose values correspond to columns
#'     of a Census API table and whose names represent the new columns they are
#'     used to calculate in [census_geo_api()].
#'   }
#'   \item{`census_geo_api_url()`}{
#'     A [character] string containing the base of the URL to a
#'     Census API table.
#'   }
#' }
#' @keywords internal
census_geo_api_names <- function(
    year = c("2020", "2010", "2000"),
    age = FALSE,
    sex = FALSE
) {
  year <- as.character(year)
  year <- rlang::arg_match(year)
  
  assert_boolean(age)
  assert_boolean(sex)
  
  if (year == "2020") {
    prefix <- "P12"
    separator <- "_"
    suffix <- "N"
    sex_codes <- c("_mal" = 2, "_fem" = 26)
    age_codes <- 1:23
    names(age_codes) <- paste0("_", age_codes)
  } else if (year %in% c("2010", "2000")) {
    prefix <- "PCT012"
    separator <- ""
    suffix <- ""
    sex_codes <- c("_mal" = 2, "_fem" = 106)
    age_codes <- list(
      "_1" = 1:5,
      "_2" = 6:10,
      "_3" = 11:15,
      "_4" = 16:18,
      "_5" = 19:20,
      "_6" = 21,
      "_7" = 22,
      "_8" = 23:25,
      "_9" = 26:30,
      "_10" = 31:35,
      "_11" = 36:40,
      "_12" = 41:45,
      "_13" = 46:50,
      "_14" = 51:55,
      "_15" = 56:60,
      "_16" = 61:62,
      "_17" = 63:65,
      "_18" = 66:67,
      "_19" = 68:70,
      "_20" = 71:75,
      "_21" = 76:80,
      "_22" = 81:85,
      "_23" = 86:103
    )
  }
  
  race_codes <- list(
    "_whi" = "I",
    "_bla" = "J",
    "_his" = "H",
    "_asi" = c("L", "M"),
    "_oth" = c("K", "N", "O")
  )
  
  numeric_codes <- if (age && sex) {
    age_sex_codes <- purrr::imap(
      sex_codes,
      function(sex_code, name) {
        codes <- purrr::map(
          age_codes,
          function(age_code) {
            str_pad(age_code + sex_code, 3, "left", pad = "0")
          }
        )
        names(codes) <- paste0(name, names(codes))
        codes
      }
    )
    
    do.call(c, unname(age_sex_codes))
  } else if (age) {
    purrr::map(
      age_codes,
      function(age_code) {
        unlist(
          purrr::map(
            sex_codes,
            function(sex_code) {
              str_pad(age_code + sex_code, 3, "left", pad = "0")
            }
          )
        )
      }
    )
  } else if (sex) {
    sex_codes[] <- str_pad(sex_codes, 3, "left", pad = "0")
    as.list(sex_codes)
  }
  
  numeric_codes <- c("001", numeric_codes)
  
  combinations <- expand.grid(
    prefix = prefix,
    race_codes = race_codes,
    separator = separator,
    numeric_codes = numeric_codes,
    suffix = suffix,
    KEEP.OUT.ATTRS = FALSE,
    stringsAsFactors = FALSE
  )
  
  vars <- purrr::pmap(
    combinations,
    function(prefix, race_codes, separator, numeric_codes, suffix) {
      inner_combinations <- expand.grid(
        prefix = prefix,
        race_codes = race_codes,
        separator = separator,
        numeric_codes = numeric_codes,
        suffix = suffix,
        KEEP.OUT.ATTRS = FALSE,
        stringsAsFactors = FALSE
      )
      
      apply(inner_combinations, 1, paste, collapse = "")
    }
  )
  
  names(vars) <- paste0(
    "r",
    names(combinations$numeric_codes),
    names(combinations$race_codes)
  )
  
  vars
}

census_geo_api_names_legacy <- function(year) {
  if (year == 2020) {
    return(
      list(
        r_whi = 'P2_005N',
        r_bla = 'P2_006N',
        r_his = 'P2_002N',
        r_asi = c('P2_008N', 'P2_009N'),
        r_oth = c('P2_007N', 'P2_010N', 'P2_011N')
      )
    )
  }
      
  list(
    r_whi = 'P005003',
    r_bla = 'P005004',
    r_his = 'P005010',
    r_asi = c('P005006', 'P005007'),
    r_oth = c('P005005', 'P005008', 'P005009')
  )
}

#' @rdname census_geo_api_names
census_geo_api_url <- function(year = c("2020", "2010", "2000")) {
  year <- as.character(year)
  year <- rlang::arg_match(year)
  
  if (year == "2020") return("https://api.census.gov/data/2020/dec/dhc?")
  paste0("https://api.census.gov/data/", year, "/dec/sf1?")
}

#' Census helper function.
#'
#' \code{census_helper_new} links user-input dataset with Census geographic data.
#'
#' This function allows users to link their geocoded dataset (e.g., voter file) 
#' with U.S. Census data (2010 or 2020). The function extracts Census Summary File data 
#' at the county, tract, block, or place level. Census data calculated are 
#' Pr(Geolocation | Race) where geolocation is county, tract, block, or place.
#'
#' @inheritParams get_census_data
#' @param voter.file An object of class \code{data.frame}. Must contain field(s) named 
#'  \code{\var{county}}, \code{\var{tract}}, \code{\var{block}}, and/or \code{\var{place}}
#'  specifying geolocation. These should be character variables that match up with 
#'  U.S. Census categories. County should be three characters (e.g., "031" not "31"), 
#'  tract should be six characters, and block should be four characters. 
#'  Place should be five characters if it is included.
#' @param states A character vector specifying which states to extract 
#'  Census data for, e.g. \code{c("NJ", "NY")}. Default is \code{"all"}, which extracts 
#'  Census data for all states contained in user-input data.
#' @param geo A character object specifying what aggregation level to use. 
#'  Use \code{"county"}, \code{"tract"}, \code{"block"}, or \code{"place"}. 
#'  Default is \code{"tract"}. Warning: extracting block-level data takes very long.
#' @param age A \code{TRUE}/\code{FALSE} object indicating whether to condition on
#'  age or not. If \code{FALSE} (default), function will return Pr(Geolocation | Race).
#'  If \code{TRUE}, function will return Pr(Geolocation, Age | Race).
#'  If \code{\var{sex}} is also \code{TRUE}, function will return Pr(Geolocation, Age, Sex | Race).
#' @param sex A \code{TRUE}/\code{FALSE} object indicating whether to condition on
#'  sex or not. If \code{FALSE} (default), function will return Pr(Geolocation | Race).
#'  If \code{TRUE}, function will return Pr(Geolocation, Sex | Race).
#'  If \code{\var{age}} is also \code{TRUE}, function will return Pr(Geolocation, Age, Sex | Race).
#' @param year A character object specifying the year of U.S. Census data to be downloaded.
#'  Use \code{"2010"}, or \code{"2020"}. Default is \code{"2020"}.
#' @param census.data A optional census object of class \code{list} containing 
#' pre-saved Census geographic data. Can be created using \code{get_census_data} function.
#' If \code{\var{census.data}} is provided, the \code{\var{year}} element must 
#' have the same value as the \code{\var{year}} option specified in this function 
#' (i.e., \code{"2010"} in both or \code{"2020"} in both). 
#' If \code{\var{census.data}} is provided, the \code{\var{age}} and the \code{\var{sex}} 
#' elements must be \code{FALSE}. This corresponds to the defaults of \code{census_geo_api}.
#' If \code{\var{census.data}} is missing, Census geographic data will be obtained via Census API. 
#' @param retry The number of retries at the census website if network interruption occurs.
#' @param use.counties A logical, defaulting to FALSE. Should census data be filtered by counties 
#' available in \var{census.data}?
#' @param skip_bad_geos Logical. Option to have the function skip any geolocations that are not present 
#' in the census data, returning a partial data set. Default is set to \code{FALSE}, which case it will 
#' break and provide error message with a list of offending geolocations.
#' @return Output will be an object of class \code{data.frame}. It will 
#'  consist of the original user-input data with additional columns of 
#'  Census data.
#'
#' @examples
#' \dontshow{data(voters)}
#' \dontrun{census_helper_new(voter.file = voters, states = "nj", geo = "block")}
#' \dontrun{census_helper_new(voter.file = voters, states = "all", geo = "tract")}
#' \dontrun{census_helper_new(voter.file = voters, states = "all", geo = "place",
#'  year = "2020")}
#'
#' @keywords internal

census_helper_new <- function(
    key = Sys.getenv("CENSUS_API_KEY"),
    voter.file,
    states = "all",
    geo = c("tract", "block", "block_group", "county", "place", "zcta"),
    age = FALSE,
    sex = FALSE,
    year = "2020",
    census.data = NULL,
    retry = 3,
    use.counties = FALSE,
    skip_bad_geos = FALSE
) {
  
  if ("precinct" %in% geo) {
    stop("Error: census_helper_new function does not currently support precinct-level data.")
  }
  
  geo <- tolower(geo)
  geo <- rlang::arg_match(geo)
  
  if(!(year %in% c("2000","2010","2020"))){
    stop("Interface only implemented for census years '2000', '2010', or '2020'.")
  }
  
  if (is.null(census.data) || (typeof(census.data) != "list")) {
    toDownload = TRUE
  } else {
    toDownload = FALSE
  }
  
  if (toDownload) {
    key <- validate_key(key)
  } 
  
  if (toupper(states) == "ALL") {
    states <- toupper(as.character(unique(voter.file$state)))
  }
  states <- as_state_abbreviation(states)
  
  df.out <- NULL
  
  for (s in 1:length(states)) {
    
    message(paste("State ", s, " of ", length(states), ": ", states[s], sep  = ""))
    state <- toupper(states[s])
    
    if (geo == "tract") {
      geo.merge <- c("county", "tract")
      if ((toDownload) || (is.null(census.data[[state]])) || (census.data[[state]]$year != year) || (census.data[[state]]$age != FALSE) || (census.data[[state]]$sex != FALSE)) {#} || (census.data[[state]]$age != age) || (census.data[[state]]$sex != sex)) {
        if(use.counties) {
          census <- census_geo_api(key, state, geo = "tract", age, sex, retry, 
                                   # Only those counties within the target state
                                   counties = unique(voter.file$county[voter.file$state == state]))
        } else {
          census <- census_geo_api(key, state, geo = "tract", age, sex, year, retry)
        }
      } else {
        census <- census.data[[toupper(state)]]$tract
      }
    } else if (geo == "block_group") {
      geo.merge <- c("county", "tract", "block_group")
      if ((toDownload) || (is.null(census.data[[state]])) || (census.data[[state]]$year != year) || (census.data[[state]]$age != FALSE) || (census.data[[state]]$sex != FALSE)) {#} || (census.data[[state]]$age != age) || (census.data[[state]]$sex != sex)) {
        if(use.counties) {
          census <- census_geo_api(key, state, geo = "block_group", age, sex, retry, 
                                   # Only those counties within the target state
                                   counties = unique(voter.file$county[voter.file$state == state]))
        } else {
          census <- census_geo_api(key, state, geo = "block_group", age, sex, year, retry)
        }
        
      } else {
        census <- census.data[[toupper(state)]]$block_group
      }
    } else if (geo == "block") {
      if(any(names(census.data) == "block_group")) {
        geo.merge <- c("county", "tract", "block_group", "block")
      } else {
        geo.merge <- c("county", "tract", "block")
      }

      if ((toDownload) || (is.null(census.data[[state]])) || (census.data[[state]]$year != year) || (census.data[[state]]$age != FALSE) || (census.data[[state]]$sex != FALSE)) {#} || (census.data[[state]]$age != age) || (census.data[[state]]$sex != sex)) {
        if(use.counties) {
          census <- census_geo_api(key, state, geo = "block", age, sex, retry, 
                                   # Only those counties within the target state
                                   counties = unique(voter.file$county[voter.file$state == state]))
        } else {
          census <- census_geo_api(key, state, geo = "block", age, sex, year, retry)
        }
        
      } else {
        census <- census.data[[toupper(state)]]$block
      }
    } else {
      geo.merge <- geo
      
      state_must_be_downloaded <- toDownload ||
        is.null(census.data[[state]]) ||
        census.data[[state]]$year != year ||
        # TODO: Why do we always redownload if sex or age == TRUE?
        census.data[[state]]$age != FALSE ||
        census.data[[state]]$sex != FALSE
      
      if (state_must_be_downloaded) {
        census <- census_geo_api(key, state, geo = geo, age, sex, year, retry)
      } else {
        census <- census.data[[state]][[geo]]
      }
    }
    
    census$state <- state
    
    ## Calculate Pr(Geolocation | Race)
    if (any(c("P2_005N", "P005003") %in% names(census))) {
      message(sprintf("NOTE: Legacy column names detected, loading Race values from Census Redistricting table for %s. Age, Sex, and ZCTA predictions will be unavailable.", year))
      # TODO: Add test that we get the same ratios with legacy and new tables for 2020
      # Old table: Redistricting (Pl-some numbers) (does not have age, sex, or ZCTAs)
      # New table: DHC (does have age, sex, and ZCTA)
      vars_ <- census_geo_api_names_legacy(year = year)
    } else {
      vars_ <- census_geo_api_names(year)
    }
    drop <- match(c("state", unlist(vars_)), names(census))
    
    geoPopulations <- rowSums(census[,names(census) %in% unlist(vars_)])
    
    census$r_whi <- rowSums(census[, vars_[["r_whi"]], drop = FALSE]) / (geoPopulations) #Pr(White | Geo)
    census$r_bla <- rowSums(census[, vars_[["r_bla"]], drop = FALSE]) / (geoPopulations) #Pr(Black | Geo)
    census$r_his <- rowSums(census[, vars_[["r_his"]], drop = FALSE]) / (geoPopulations) #Pr(Latino | Geo)
    census$r_asi <- rowSums(census[, vars_[["r_asi"]], drop = FALSE]) / (geoPopulations) #Pr(Asian or NH/PI | Geo)
    census$r_oth <- rowSums(census[, vars_[["r_oth"]], drop = FALSE]) / (geoPopulations) #Pr(AI/AN, Other, or Mixed | Geo)
    
    # check locations with zero people
    # get average without places with zero people, and assign that to zero locs.
    zero_ind <- which((geoPopulations - 0.0) < .Machine$double.eps)
    if (length(zero_ind)) {
      for (rcat in c("r_whi","r_bla","r_his","r_asi","r_oth")) {
        census[[rcat]][zero_ind] <- mean(census[[rcat]], na.rm = TRUE)
      }
    }
    
    voters.census <- merge(
      voter.file[toupper(voter.file$state) == toupper(states[s]), ],
      census[, -drop], by = geo.merge, all.x  = TRUE)
    
    #Check if geolocation missing from census object
    if(any(is.na(voters.census$r_whi))){
      miss_ind <- which(is.na(voters.census$r_whi))
      message("The following locations in the voter.file are not available in the census data.",
           paste0("(listed as ", paste0(c("state",geo.merge), collapse="-"),"):\n"),
           paste(do.call(paste, c(unique(voters.census[miss_ind, c("state",geo.merge)]),
                                  sep="-")),
                 collapse = ", "))
      if(skip_bad_geos == TRUE) {
        message("NOTE: Skipping unavailable geolocations. Returning partial data set.")
        voters.census <- voters.census[!is.na(voters.census$r_whi),]
      }
      else {
        stop("Stopping predictions. Please revise census data and/or verify the correct year is being supplied. To skip these rows use 'skip_bad_geos = TRUE'")
      }
    }
      
    # }
    
    keep.vars <- c(names(voter.file)[names(voter.file) != "agecat"], 
                   paste("r", c("whi", "bla", "his", "asi", "oth"), sep = "_"))
    df.out <- as.data.frame(rbind(df.out, voters.census[keep.vars]))
    
  }
  
  return(df.out)
}

# Note: must provide a valid U.S. Census API key for test cases that use U.S. Census statistics
# > usethis::edit_r_profile
# Sys.setenv("CENSUS_API_KEY" = "yourkey")
# For testing package coverage use: Sys.setenv("NOT_CRAN" = "TRUE")
options("piggyback.verbose" = FALSE)
options("wru_data_wd" = TRUE)

skip_if_not(nzchar(Sys.getenv("CENSUS_API_KEY")))
test_that("Fails if 'precinct' is set as the geo var",{
  skip_on_cran()
  set.seed(42)
  data(voters)
  future::plan(future::multisession)
  census <- readRDS(test_path("data/new_census_table_NJ_2020.rds"))
  expect_error(
  census_helper_new(
    voter.file = voters,
    states = "all",
    geo = "precinct",
    age = FALSE,
    sex = FALSE,
    year = "2020",
    census.data = census,
    retry = 3,
    use.counties = FALSE,
    skip_bad_geos = FALSE
  ),
  "Error: census_helper_new function does not currently support precinct-level data.")
})

skip_if_not(nzchar(Sys.getenv("CENSUS_API_KEY")))
test_that("helper returns verified census tract data",{
  skip_on_cran()
  set.seed(42)
  data(voters)
  future::plan(future::multisession)
  census <- readRDS(test_path("data/new_census_table_NJ_2020.rds"))
  x <- census_helper_new(
    voter.file = voters,
    states = "NJ",
    geo = "tract",
    age = FALSE,
    sex = FALSE,
    year = "2020",
    census.data = census,
    retry = 3,
    use.counties = FALSE,
    skip_bad_geos = FALSE
    )
  expect_equal(x[x$surname == "Lopez", "r_whi"], 0.767197, tolerance = 0.000001)
  expect_equal(x[x$surname == "Khanna", "r_whi"], 0.708026, tolerance = 0.000001)
  expect_equal(x[x$surname == "Lopez", "r_bla"], 0.09522743, tolerance = 0.000001)
  expect_equal(x[x$surname == "Khanna", "r_bla"], 0.09544469, tolerance = 0.000001)
})

skip_if_not(nzchar(Sys.getenv("CENSUS_API_KEY")))
test_that("New tables and legacy tables return equal race predictions",{
  skip_on_cran()
  set.seed(42)
  data(voters)
  future::plan(future::multisession)
  # legacy redistricting table
  census <- readRDS(test_path("data/census_test_nj_block_2020.rds"))
  x <- census_helper_new(
    voter.file = voters,
    states = "NJ",
    geo = "tract",
    age = FALSE,
    sex = FALSE,
    year = "2020",
    census.data = census,
    use.counties = FALSE
  )
  # use new table source
  new_census <- readRDS(test_path("data/new_census_table_NJ_2020.rds"))
  y <- census_helper_new(
    voter.file = voters,
    states = "NJ",
    geo = "tract",
    age = FALSE,
    sex = FALSE,
    year = "2020",
    census.data = new_census,
    use.counties = FALSE
  )
  expect_equal(x$r_whi, y$r_whi, tolerance = .01)
  # expect_equal(x$r_bla, y$r_bla, tolerance = .01)
  expect_equal(x$r_his, y$r_his, tolerance = .01)
  expect_equal(x$r_asi, y$r_asi, tolerance = .01)
  # expect_equal(x$r_oth, y$r_oth, tolerance = .01)
})

#' Race prediction function.
#'
#' \code{predict_race} makes probabilistic estimates of individual-level race/ethnicity.
#'
#' This function implements the Bayesian race prediction methods outlined in
#' Imai and Khanna (2015). The function produces probabilistic estimates of
#' individual-level race/ethnicity, based on surname, geolocation, and party.
#' @param voter.file An object of class \code{data.frame}.
#' Must contain a row for each individual being predicted,
#' as well as a field named \code{\var{surname}} containing each individual's surname.
#' If using geolocation in predictions, \code{\var{voter.file}} must contain a field named
#' \code{\var{state}}, which contains the two-character abbreviation for each individual's
#' state of residence (e.g., \code{"nj"} for New Jersey).
#' If using Census geographic data in race/ethnicity predictions,
#' \code{\var{voter.file}} must also contain at least one of the following fields:
#' \code{\var{county}}, \code{\var{tract}}, \code{\var{block_group}}, \code{\var{block}}, 
#' and/or \code{\var{place}}.
#' These fields should contain character strings matching U.S. Census categories.
#' County is three characters (e.g., \code{"031"} not \code{"31"}),
#' tract is six characters, block group is usually a single character and block
#'  is four characters. Place is five characters.
#' See below for other optional fields.
#' @param census.surname A \code{TRUE}/\code{FALSE} object. If \code{TRUE},
#'  function will call \code{merge_surnames} to merge in Pr(Race | Surname)
#'  from U.S. Census Surname List (2000, 2010, or 2020) and Spanish Surname List.
#'  If \code{FALSE}, user must provide a \code{name.dictionary} (see below).
#'  Default is \code{TRUE}.
#' @param surname.only A \code{TRUE}/\code{FALSE} object. If \code{TRUE}, race predictions will
#'  only use surname data and calculate Pr(Race | Surname). Default is \code{FALSE}.
#' @param census.geo An optional character vector specifying what level of
#' geography to use to merge in U.S. Census geographic data. Currently
#' \code{"county"}, \code{"tract"}, \code{"block_group"}, \code{"block"}, and \code{"place"} 
#' are supported.
#' Note: sufficient information must be in user-defined \code{\var{voter.file}} object.
#' If \code{\var{census.geo} = "county"}, then \code{\var{voter.file}}
#' must have column named \code{county}.
#' If \code{\var{census.geo} = "tract"}, then \code{\var{voter.file}}
#' must have columns named \code{county} and \code{tract}.
#' And if \code{\var{census.geo} = "block"}, then \code{\var{voter.file}}
#' must have columns named \code{county}, \code{tract}, and \code{block}.
#' If \code{\var{census.geo} = "place"}, then \code{\var{voter.file}}
#' must have column named \code{place}.
#' If `census.geo = "zcta"`, then `voter.file` must have column named `zcta`.
#' Specifying \code{\var{census.geo}} will call \code{census_helper} function
#' to merge Census geographic data at specified level of geography.
#' 
#' @param census.key A character object specifying user's Census API key.
#'   Required if `census.geo` is specified, because a valid Census API key is
#'   required to download Census geographic data.
#'   
#'   If [`NULL`], the default, attempts to find a census key stored in an
#'   [environment variable][Sys.getenv] named `CENSUS_API_KEY`.
#'   
#' @param census.data A list indexed by two-letter state abbreviations,
#' which contains pre-saved Census geographic data.
#' Can be generated using \code{get_census_data} function.
#' @param age An optional \code{TRUE}/\code{FALSE} object specifying whether to
#' condition race predictions on age (in addition to surname and geolocation).
#' Default is \code{FALSE}. Must be same as \code{\var{age}} in \code{\var{census.data}} object.
#' May only be set to \code{TRUE} if \code{census.geo} option is specified.
#' If \code{TRUE}, \code{\var{voter.file}} should include a numerical variable \code{\var{age}}.
#' @param sex optional \code{TRUE}/\code{FALSE} object specifying whether to
#' condition race predictions on sex (in addition to surname and geolocation).
#' Default is \code{FALSE}. Must be same as \code{\var{sex}} in \code{\var{census.data}} object.
#' May only be set to \code{TRUE} if \code{census.geo} option is specified.
#' If \code{TRUE}, \code{\var{voter.file}} should include a numerical variable \code{\var{sex}},
#' where \code{\var{sex}} is coded as 0 for males and 1 for females.
#' @param year An optional character vector specifying the year of U.S. Census geographic
#' data to be downloaded. Use \code{"2010"}, or \code{"2020"}. Default is \code{"2020"}.
#' @param party An optional character object specifying party registration field
#' in \code{\var{voter.file}}, e.g., \code{\var{party} = "PartyReg"}.
#' If specified, race/ethnicity predictions will be conditioned
#' on individual's party registration (in addition to geolocation).
#' Whatever the name of the party registration field in \code{\var{voter.file}},
#' it should be coded as 1 for Democrat, 2 for Republican, and 0 for Other.
#' @param retry The number of retries at the census website if network interruption occurs.
#' @param impute.missing Logical, defaults to TRUE. Should missing be imputed?
#' @param skip_bad_geos Logical. Option to have the function skip any geolocations that are not present 
#' in the census data, returning a partial data set. Default is set to \code{FALSE}, in which case it
#' will break and provide error message with a list of offending geolocations.
#' @param use.counties A logical, defaulting to FALSE. Should census data be filtered by counties 
#' available in \var{census.data}?
#' @param model Character string: "BISG" (default), "fBISG" (fully-Bayesian with
#' error correction), or "eBISG" (embedding-supplemented BISG, which uses
#' text embeddings to predict race probabilities for names not found
#' in Census surname lists). The eBISG model requires Python with
#' sentence-transformers and torch; run \code{\link{setup_ebisg}} to configure.
#' @param ebisg.model Character string (HuggingFace model ID) or named list
#' specifying which embedding model to use when \code{model = "eBISG"}.
#' The only built-in option is \code{"intfloat/multilingual-e5-large"}
#' (default; 1024-dim), which is keyed by its full HuggingFace ID for
#' consistency with the Python side. To use a different sentence-transformer,
#' pass a named list with elements \code{transformer} (HuggingFace model ID),
#' \code{dim} (embedding dimension), and \code{surname_mlp},
#' \code{firstname_mlp} (paths to custom .pt checkpoints).
#' @param name.dictionaries Optional named list of \code{data.frame}'s 
#' containing counts of names by race. Any of the following named elements 
#' are allowed: "surname", "first", "middle". When present, the objects must 
#' follow the same structure as \code{last_c}, \code{first_c},
#'\code{mid_c}, respectively.
#' @param names.to.use One of 'surname', 'surname, first', or  'surname, first,
#'  middle'. Defaults to 'surname'.
#' @param race.init Vector of initial race for each observation in voter.file.
#' Must be an integer vector, with 1=white, 2=black, 3=hispanic, 4=asian, and 
#' 5=other. Defaults to values obtained using \code{model="BISG_surname"}.
#' @param control List of control arguments only used when \code{model="fBISG"}, including
#' \describe{
#'   \item{iter}{Number of MCMC iterations. Defaults to 1000.}
#'   \item{burnin}{Number of iterations discarded as burnin. Defaults to half of \code{iter}.}
#'   \item{verbose}{Print progress information. Defaults to \code{TRUE}.}
#'   \item{me.correct}{Boolean. Should the model correct measurement error for \code{races|geo}? Defaults to \code{TRUE}.}
#'   \item{seed}{RNG seed. If \code{NULL}, a seed is generated and returned as an attribute for reproducibility.}
#' }
#'
#' @return Output will be an object of class \code{data.frame}. It will
#'  consist of the original user-input \code{voter.file} with additional columns with
#'  predicted probabilities for each of the five major racial categories:
#'  \code{\var{pred.whi}} for White,
#'  \code{\var{pred.bla}} for Black,
#'  \code{\var{pred.his}} for Hispanic/Latino,
#'  \code{\var{pred.asi}} for Asian/Pacific Islander, and
#'  \code{\var{pred.oth}} for Other/Mixed.
#'
#' @examples
#' \donttest{
#' #' data(voters)
#' try(predict_race(voter.file = voters, surname.only = TRUE))
#' \dontrun{
#' try(predict_race(voter.file = voters, census.geo = "tract"))
#' }
#' \dontrun{
#' try(predict_race(
#'   voter.file = voters, census.geo = "place", year = "2020"))
#' }
#' \dontrun{
#' CensusObj <- try(get_census_data(state = c("NY", "DC", "NJ")))
#' try(predict_race(
#'   voter.file = voters, census.geo = "tract", census.data = CensusObj, party = "PID")
#'   )
#' }
#' \dontrun{
#' CensusObj2 <- try(get_census_data(state = c("NY", "DC", "NJ"), age = T, sex = T))
#' try(predict_race(
#'   voter.file = voters, census.geo = "tract", census.data = CensusObj2, age = T, sex = T))
#' }
#' \dontrun{
#' CensusObj3 <- try(get_census_data(state = c("NY", "DC", "NJ"), census.geo = "place"))
#' try(predict_race(voter.file = voters, census.geo = "place", census.data = CensusObj3))
#' }
#' }

#' @export

predict_race <- function(
    voter.file,
    census.surname = TRUE,
    surname.only = FALSE,
    census.geo = c("tract", "block", "block_group", "county", "place", "zcta"),
    census.key = Sys.getenv("CENSUS_API_KEY"),
    census.data = NULL,
    age = FALSE,
    sex = FALSE,
    year = "2020",
    party = NULL,
    retry = 3,
    impute.missing = TRUE,
    skip_bad_geos = FALSE,
    use.counties = FALSE,
    model = "BISG",
    race.init = NULL,
    name.dictionaries = NULL,
    names.to.use = "surname",
    control = NULL,
    ebisg.model = "intfloat/multilingual-e5-large"
) {
  
  message("Predicting race for ", year)
  
  ## Check model type
  if (!(model %in% c("BISG", "fBISG", "eBISG"))) {
    stop(
      paste0(
        "'model' must be one of 'BISG' (for standard BISG results, or results",
        " with all name data without error correction), 'fBISG' (for the",
        " fully Bayesian/error correction model that accommodates all name data),",
        " or 'eBISG' (for embedding-supplemented BISG)."
      )
    )
  }
  
  if (any(unique(voter.file$state) %in% c("AS","GU","MP","PR","VI"))) {
    stop(
      paste0(
        "The wru package does not support US territories", 
        " please filter these from your voter.file data")
    )
  }
  
  census.geo <- tolower(census.geo)
  census.geo <- rlang::arg_match(census.geo)
  
  # block_group is missing, pull from block
  if((surname.only == FALSE) && !(missing(census.geo)) && (census.geo == "block_group") && !("block_group" %in% names(voter.file))) {
    voter.file$block_group <- substring(voter.file$block, 1, 1)
  }
  
  # Adjust voter.file with caseid for ordering at the end
  voter.file$caseid <- 1:nrow(voter.file)
  
  if (surname.only == FALSE && is.null(census.data)) {
    # Otherwise predict_race_new and predict_race_me will both
    # attempt to pull census_data
    census.key <- validate_key(census.key)
    voter.file$state <- toupper(voter.file$state)
    states <- unique(voter.file$state)
    county.list <- split(voter.file$county, voter.file$state)
    county.list <- lapply(county.list, function(x) unique(x))
    census.data <- get_census_data(
      census.key, states, age, 
      sex, year, census.geo, 
      retry, county.list
    )
  }
  
  if (model == "eBISG") {
    if (isTRUE(surname.only)) {
      warning("eBISG surname-only mode: embedding predictions will be used for unmatched surnames.")
    }
    preds <- predict_race_embedding(
      voter.file = voter.file,
      names.to.use = names.to.use,
      year = year,
      age = age, sex = sex,
      census.geo = census.geo,
      census.key = census.key,
      name.dictionaries = name.dictionaries,
      surname.only = surname.only,
      census.data = census.data,
      retry = retry,
      impute.missing = impute.missing,
      skip_bad_geos = skip_bad_geos,
      census.surname = census.surname,
      use.counties = use.counties,
      ebisg.model = ebisg.model
    )
  } else if((model == "BISG") | (surname.only==TRUE)){
    if((surname.only==TRUE) & (model == "fBISG")){
      warning("Surname-only model only available with model = BISG.")
    }
    preds <- predict_race_new(voter.file = voter.file,
                              names.to.use = names.to.use,
                              year = year,
                              age = age, sex = sex, # not implemented, default to F
                              census.geo = census.geo,
                              census.key = census.key,
                              name.dictionaries = name.dictionaries,
                              surname.only=surname.only,
                              census.data = census.data,
                              retry = retry,
                              impute.missing = impute.missing,
                              skip_bad_geos = skip_bad_geos,
                              census.surname = census.surname,
                              use.counties = use.counties)
  } else {
    ctrl <- list(
      iter = 1000,
      thin = 1,
      verbose = TRUE,
      seed = sample(1:1000, 1) 
    )
    ctrl$burnin <- floor(ctrl$iter / 2)
    ctrl[names(control)] <- control
    ctrl$usr_seed <- ifelse(is.null(control$seed), FALSE, TRUE)

    if (is.null(race.init)) {
      if(ctrl$verbose){
        message("Using `predict_race` to obtain initial race prediction priors with BISG model")
      }

      race.init <-  predict_race(voter.file = voter.file,
                                 names.to.use = names.to.use,
                                 year = year,
                                 age = age, sex = sex, # not implemented, default to F
                                 census.geo = census.geo,
                                 census.key = census.key,
                                 name.dictionaries = name.dictionaries,
                                 surname.only=surname.only,
                                 census.data = census.data,
                                 retry = retry,
                                 impute.missing = TRUE,
                                 skip_bad_geos = skip_bad_geos,
                                 census.surname = census.surname,
                                 use.counties = use.counties,
                                 model = "BISG",
                                 control = list(verbose=FALSE))

      race.init <- max.col(
        race.init[, paste0("pred.", c("whi", "bla", "his", "asi", "oth"))],
        ties.method = "random"
      )
    }
    if (any(is.na(race.init))) {
      stop("Some initial race values are NA.\n
         If you didn't provide initial values, check the results of calling predict_race() on the voter.file you want me to work on.\n
         The most likely reason for getting a missing race prediction is having a geolocation that does not match \n 
         locations on the census. If this problem persists, try impute.missing = TRUE or model = fBISG.")
    }
    
    preds <- predict_race_me(voter.file = voter.file,
                             names.to.use = names.to.use,
                             year = year, age = age, sex = age, 
                             census.geo = census.geo,
                             census.key = census.key,
                             name.dictionaries = name.dictionaries,
                             surname.only = surname.only,
                             census.data = census.data, retry = retry,
                             impute.missing = impute.missing,
                             census.surname = census.surname,
                             use.counties = use.counties, race.init = race.init,
                             ctrl = ctrl)
  }
  seed_attr <- attr(preds, "RNGseed")
  preds <- preds[order(preds$caseid),setdiff(names(preds), "caseid")]
  attr(preds, "RNGseed") <- seed_attr
  preds
}

# Note: must provide a valid U.S. Census API key for test cases that use U.S. Census statistics
# > usethis::edit_r_profile
# Sys.setenv("CENSUS_API_KEY" = "yourkey")
options("piggyback.verbose" = FALSE)
options("wru_data_wd" = TRUE)

test_that("Fails if model is set to anything other than BISG or fBISG", {
  skip_on_cran()
  set.seed(42)
  data(voters)
  expect_error(suppressMessages(predict_race(
    voter.file = voters,
    surname.only = TRUE,
    model = "tBISG")),
    "'model' must be one of 'BISG' \\(for standard BISG results, or results"
    )
})

test_that("Tests surname only predictions", {
  skip_on_cran()
  set.seed(42)
  data(voters)
  # Prediction using surname only
  x <- suppressMessages(predict_race(
    voter.file = voters,
    surname.only = TRUE))
  # Test and confirm prediction output is as expected
  expect_equal(dim(x), c(10, 20))
  expect_equal(sum(is.na(x)), 0)
  expect_equal(round(x[x$surname == "Khanna", "pred.whi"], 4), 0.045, tolerance = 0.01)
  expect_equal(round(x[x$surname == "Johnson", "pred.his"], 4), 0.0272, tolerance = 0.01)
})

test_that("Test BISG NJ at county level", {
  skip_on_cran()
  set.seed(42)
  data(voters)
  census <- readRDS(test_path("data/census_test_nj_block_2020.rds"))
  x <- suppressMessages(predict_race(
      voter.file = voters[voters$state == "NJ",],
      census.geo = "county",
      census.data = census
    ))

  expect_equal(as.character(x$VoterID), as.character(c(1, 2, 4, 5, 6, 8, 9)))
  expect_equal(dim(x), c(7, 20))
  expect_equal(sum(is.na(x)), 0L)
  expect_equal(sum(x$surname == "Johnson"), 0)
  expect_equal(round(x[x$surname == "Khanna", "pred.whi"], 4), 0.0181, tolerance = 0.01)
  expect_equal(round(x[x$surname == "Khanna", "pred.asi"], 4), 0.9444, tolerance = 0.01)
  expect_equal(round(x[x$surname == "Fifield", "pred.whi"], 4), 0.8664, tolerance = 0.01)
  expect_equal(round(x[x$surname == "Lopez", "pred.his"], 4), 0.9392, tolerance = 0.01)
})

test_that("Test fBISG NJ at tract level", {
  skip_on_cran()
  set.seed(42)
  data(voters)
  
  census <- readRDS(test_path("data/census_test_nj_block_2020.rds"))

  x <- suppressMessages(predict_race(
    voter.file = voters[voters$state == "NJ",],
    census.geo = "tract",
    census.data = census,
    model = "fBISG",
    control=list(verbose=FALSE)
  ))
  
  expect_equal(as.character(x$VoterID), as.character(c(1, 2, 4, 5, 6, 8, 9)))
  expect_equal(dim(x), c(7, 20))
  expect_equal(sum(is.na(x)), 0L)
  expect_equal(sum(x$surname == "Johnson"), 0)
  expect_equal(round(x[x$surname == "Khanna", "pred.whi"], 4), 0.031, tolerance = 0.01) 
  expect_equal(round(x[x$surname == "Lopez", "pred.his"], 4), 0.798, tolerance = 0.01) 
})

test_that("BISG NJ at block level", {
  skip_on_cran()
  set.seed(42)
  data(voters)
  census <- readRDS(test_path("data/census_test_nj_block_2020.rds"))
  voters <- dplyr::mutate(voters, block = dplyr::case_when(block == 1025 ~ "3001", TRUE ~ block))
  
  x <- suppressMessages(predict_race(
    voter.file = voters[voters$state == "NJ", ], 
    census.geo = "block", 
    census.key = NULL, 
    census.data = census, 
    use.counties = TRUE)
  )
  
  expect_equal(dim(x), c(7, 20))
  expect_equal(sum(is.na(x$pred.asi)), 0L)
  expect_true(!any(duplicated(x$surname)))
  expect_equal(x[x$surname == "Khanna", "pred.asi"], 0.8078, tolerance = 0.01)
  expect_equal(x[x$surname == "Zhou", "pred.asi"], 0.9926, tolerance = 0.1)
  expect_equal(x[x$surname == "Lopez", "pred.his"], 0.8605, tolerance = 0.1)
})

test_that("BISG NJ at block_group level", {
  skip_on_cran()
  set.seed(42)
  data(voters)
  census <- readRDS(test_path("data/census_test_nj_block_2020.rds"))
  
  voters <- voters[voters$state == "NJ", ]
  voters$block_group <- "1"
  
  x <- suppressMessages(predict_race(
    voter.file = voters, 
    census.geo = "block_group", 
    census.key = NULL, 
    census.data = census, 
    use.counties = TRUE)
  )
  
  expect_equal(dim(x), c(7, 21))
  expect_equal(sum(is.na(x$pred.asi)), 0)
  expect_true(!any(duplicated(x$surname)))
  expect_equal(x[x$surname == "Khanna", "pred.asi"], 0.9374, tolerance = 0.01)
  expect_equal(x[x$surname == "Zhou", "pred.asi"], 0.9954, tolerance = 0.01)
  expect_equal(x[x$surname == "Lopez", "pred.his"], 0.8361, tolerance = 0.01)
})

test_that("Fails on territories", {
  data(voters)
  voters$state <- "GU"
  expect_error(
    predict_race(voter.file = voters),
    "The wru package does not support US territories"
  )
}) 

test_that("Fails on missing geolocation if skip_bad_geos default is used", {
  skip_on_cran()
  set.seed(42)
  data(voters)
  census <- readRDS(test_path("data/census_test_nj_block_2020.rds"))
  expect_error(suppressMessages(predict_race(
    voter.file = voters[voters$state == "NJ", ], 
    census.geo = "block", 
    census.key = NULL, 
    census.data = census, 
    use.counties = TRUE)
  ),
  "Stopping predictions. Please revise"
  )
})

test_that("Skip_bad_geos option successfully returns working geolocations", {
  skip_on_cran()
  set.seed(42)
  data(voters)
  census <- readRDS(test_path("data/census_test_nj_block_2020.rds"))
  test_drop <- suppressMessages(predict_race(
    voter.file = voters[voters$state == "NJ", ], 
    census.geo = "block", 
    census.key = NULL, 
    census.data = census,
    skip_bad_geos = TRUE,
    use.counties = TRUE)
  )
  expect_equal(nrow(test_drop), 1)
  })

test_that("Handles zero-pop. geolocations", {
  skip_on_cran()
  set.seed(42)
  data(voters)
  census <- readRDS(test_path("data/census_test_nj_block_2020.rds"))
  census$NJ$county[6,grep("P005",colnames(census$NJ$county))] <- 0
  x <- suppressMessages(predict_race(
    voter.file = voters[voters$state == "NJ", ], 
    census.geo = "county", 
    census.key = NULL, 
    census.data = census, 
    use.counties = TRUE)
  )
  expect_equal(dim(x), c(7, 20))
  expect_equal(sum(is.na(x$pred.asi)), 0)
  expect_true(!any(duplicated(x$surname)))
  expect_equal(x[x$surname == "Khanna", "pred.asi"], 0.9444, tolerance = 0.01)
  expect_equal(x[x$surname == "Zhou", "pred.asi"], 0.9932, tolerance = 0.01)
  expect_equal(x[x$surname == "Lopez", "pred.his"], 0.9392, tolerance = 0.01)
})

test_that("Fixes for issue #68 work as expected", {
  skip_on_cran()
  set.seed(42)
  surname <- c("SULLIVAN")
  one <- predict_race(voter.file=data.frame(surname), surname.only=TRUE)
  
  surname <- c("SULLIVAN", "SULLIVAN")
  two <- predict_race(voter.file=data.frame(surname), surname.only=TRUE)
  
  surname <- c("SULLIVAN", "SULLIVAN", "SULLIVAN")
  three <- predict_race(voter.file=data.frame(surname), surname.only=TRUE)
  
  expect_equal(one$pred.whi, 0.8397254)
  expect_equal(two$pred.whi[1], 0.8397254)
  expect_equal(two$pred.whi[2], 0.8397254)
  
  expect_equal(three$pred.whi[1], 0.8397254)
  expect_equal(three$pred.whi[2], 0.8397254)
  expect_equal(three$pred.whi[3], 0.8397254)
})

## Regression tests for eBISG handling of unmatched surnames.
skip_on_cran()
##
## These guard against re-introducing the bug where merge_names'
## cleaning cascade overwrites lastname.match for non-hyphenated
## unmatched names (it leaves the literal string "NA" in that column),
## which previously caused predict_race_embedding to embed the token
## "NA" instead of the actual surname text — collapsing every distinct
## non-hyphenated unmatched surname to the same garbage prediction.
##
## We mock the embedding/MLP calls so the tests exercise the real
## predict_race_embedding code path without needing Python deps or a
## downloaded MLP checkpoint.

with_mocked_ebisg <- function(code) {
  embed_calls <- character(0)
  testthat::local_mocked_bindings(
    ensure_ebisg_python  = function(...) invisible(),
    ebisg_data_preflight = function(...) tempdir(),
    ebisg_embed_names    = function(names, ...) {
      embed_calls <<- c(embed_calls, names)
      ## Return a deterministic per-name vector so distinct inputs
      ## produce distinct (downstream) outputs.
      matrix(seq_along(names), nrow = length(names), ncol = 4L)
    },
    ebisg_predict_mlp = function(embeddings, ...) {
      n <- nrow(embeddings)
      ## Distinct softmax distributions per input row, normalized.
      raw <- matrix(seq_len(n * 6L), nrow = n, ncol = 6L)
      probs <- raw / rowSums(raw)
      ## Mirror the real ebisg_predict_mlp() contract: name the columns
      ## using the EBISG_RACE_COLS ordering from the Python helper.
      colnames(probs) <- c("whi", "bla", "his", "asi", "aian", "oth")
      probs
    },
    .package = "wru"
  )
  result <- code()
  list(result = result, embed_calls = embed_calls)
}

test_that("eBISG embeds the original surname text, not merge_names' scratchpad", {
  vf <- data.frame(
    surname = c("SMITH", "ZQXWVPLM", "QQXWPLZZ"),
    state   = "NJ",
    stringsAsFactors = FALSE
  )
  out <- with_mocked_ebisg(function() {
    predict_race(voter.file = vf, surname.only = TRUE, model = "eBISG")
  })

  ## The fix: original surnames flow into the embedding step. The bug
  ## would have sent the literal token "NA" instead.
  expect_true(all(c("ZQXWVPLM", "QQXWPLZZ") %in% out$embed_calls))
  expect_false("NA" %in% out$embed_calls)
})

test_that("eBISG keeps unique unmatched surnames distinct in the embedding step", {
  ## Pre-fix, two different non-hyphenated unmatched surnames both
  ## collapsed to lastname.match = "NA" and so produced one embedding
  ## input shared across all such rows.
  vf <- data.frame(
    surname = c("ZQXWVPLM", "QQXWPLZZ"),
    state   = "NJ",
    stringsAsFactors = FALSE
  )
  out <- with_mocked_ebisg(function() {
    predict_race(voter.file = vf, surname.only = TRUE, model = "eBISG")
  })
  expect_setequal(out$embed_calls, c("ZQXWVPLM", "QQXWPLZZ"))
})

test_that("eBISG produces finite, normalized predictions for unmatched non-hyphenated surnames", {
  vf <- data.frame(
    surname = c("SMITH", "FAKE-NAME", "ZQXWVPLM"),
    state   = "NJ",
    stringsAsFactors = FALSE
  )
  out <- with_mocked_ebisg(function() {
    predict_race(voter.file = vf, surname.only = TRUE, model = "eBISG")
  })
  pred_cols <- grep("^pred\\.", names(out$result), value = TRUE)
  expect_equal(length(pred_cols), 5L)
  expect_false(any(is.na(out$result[, pred_cols])))
  expect_false(any(is.nan(as.matrix(out$result[, pred_cols]))))
  expect_equal(
    unname(rowSums(out$result[, pred_cols])),
    rep(1, nrow(vf)),
    tolerance = 1e-6
  )
})

test_that("eBISG skips rows with NA / empty surname rather than overriding with NA", {
  vf <- data.frame(
    surname = c("SMITH", NA_character_, ""),
    state   = "NJ",
    stringsAsFactors = FALSE
  )
  ## merge_names rejects NA surnames at validation, so we restrict to the
  ## empty-string case here, which it accepts.
  vf <- vf[vf$surname %in% c("SMITH", "") | is.na(vf$surname), ]
  vf <- vf[!is.na(vf$surname), ]
  out <- with_mocked_ebisg(function() {
    predict_race(voter.file = vf, surname.only = TRUE, model = "eBISG")
  })
  expect_false("" %in% out$embed_calls)
  pred_cols <- grep("^pred\\.", names(out$result), value = TRUE)
  expect_false(any(is.na(out$result[, pred_cols])))
})

#' Surname probability merging function.
#'
#' \code{merge_names} merges names in a user-input dataset with corresponding
#'  race/ethnicity probabilities derived from both the U.S. Census Surname List
#'  and Spanish Surname List and voter files from states in the Southern U.S.
#'
#' This function allows users to match names in their dataset with database entries
#'  estimating P(name | ethnicity) for each of the five major racial groups for each
#'  name. The database probabilities are derived from both the U.S. Census Surname List
#'  and Spanish Surname List and voter files from states in the Southern U.S.
#'
#'  By default, the function matches names as follows:
#'  1) Search raw surnames in the database;
#'  2) Remove any punctuation and search again;
#'  3) Remove any spaces and search again;
#'  4) Remove suffixes (e.g., "Jr") and search again (last names only)
#'  5) Split double-barreled names into two parts and search first part of name;
#'  6) Split double-barreled names into two parts and search second part of name;
#'
#'  Each step only applies to names not matched in a previous step.
#'  Steps 2 through 6 are not applied if \code{clean.surname} is FALSE.
#'
#'  Note: Any name appearing only on the Spanish Surname List is assigned a
#'  probability of 1 for Hispanics/Latinos and 0 for all other racial groups.
#'
#' @param voter.file An object of class \code{data.frame}.  Must contain a row for each individual being predicted,
#' as well as a field named \code{\var{last}} containing each individual's surname.
#' If first name is also being used for prediction, the file must also contain a field
#' named \code{\var{first}}. If middle name is also being used for prediction, the field
#' must also contain a field named \code{\var{middle}}.
#' @param namesToUse A character vector identifying which names to use for the prediction.
#' The default value is \code{"last"}, indicating that only the last name will be used.
#' Other options are \code{"last, first"}, indicating that both last and first names will be
#' used, and \code{"last, first, middle"}, indicating that last, first, and middle names will all
#' be used.
#' @param census.surname A \code{TRUE}/\code{FALSE} object. If \code{TRUE},
#'  function will call \code{merge_surnames} to merge in Pr(Race | Surname)
#'  from U.S. Census Surname List (2000, 2010, or 2020) and Spanish Surname List.
#'  If \code{FALSE}, user must provide a \code{name.dictionary} (see below).
#'  Default is \code{TRUE}.
#' @param table.surnames An object of class \code{data.frame} provided by the
#' users as an alternative surname dictionary. It will consist of a list of
#' U.S. surnames, along with the associated probabilities P(name | ethnicity)
#' for ethnicities: white, Black, Hispanic, Asian, and other. Default is \code{NULL}.
#' (\code{\var{last_name}} for U.S. surnames, \code{\var{p_whi_last}} for White,
#' \code{\var{p_bla_last}} for Black, \code{\var{p_his_last}} for Hispanic,
#' \code{\var{p_asi_last}} for Asian, \code{\var{p_oth_last}} for other).
#' @param table.first See \code{\var{table.surnames}}.
#' @param table.middle See \code{\var{table.surnames}}.
#' @param impute.missing See \code{predict_race}.
#' @param model See \code{predict_race}.
#' @param clean.names A \code{TRUE}/\code{FALSE} object. If \code{TRUE},
#' any surnames in \code{\var{voter.file}} that cannot initially be matched
#' to the database will be cleaned, according to U.S. Census specifications,
#' in order to increase the chance of finding a match. Default is \code{TRUE}.
#' @return Output will be an object of class \code{data.frame}. It will
#'  consist of the original user-input data with additional columns that
#'  specify the part of the name matched with Census data (\code{\var{surname.match}}),
#'  and the probabilities Pr(Race | Surname) for each racial group
#'  (\code{\var{p_whi}} for White, \code{\var{p_bla}} for Black,
#'  \code{\var{p_his}} for Hispanic/Latino,
#'  \code{\var{p_asi}} for Asian and Pacific Islander, and
#'  \code{\var{p_oth}} for Other/Mixed).
#' @importFrom dplyr coalesce
#' @examples
#' data(voters)
#' \dontrun{try(merge_names(voters, namesToUse = "surname", census.surname = TRUE))}
#' @keywords internal
merge_names <- function(voter.file, namesToUse, census.surname, table.surnames = NULL, table.first = NULL, table.middle = NULL, clean.names = TRUE, impute.missing = FALSE, model = "BISG") {

  # check the names
  if (namesToUse == "surname") {
    if (!("surname" %in% names(voter.file))) {
      stop("Voter data frame needs to have a column named 'surname'.")
    }
  } else if (namesToUse == "surname, first") {
    if (!("surname" %in% names(voter.file)) || !("first" %in% names(voter.file))) {
      stop("Voter data frame needs to have a column named 'surname' and a column called 'first'.")
    }
  } else if (namesToUse == "surname, first, middle") {
    if (!("surname" %in% names(voter.file)) || !("first" %in% names(voter.file)) ||
      !("middle" %in% names(voter.file))) {
      stop("Voter data frame needs to have a column named 'surname', a column called 'first', and a column called 'middle'.")
    }
  }

  wru_data_preflight()

  path <- ifelse(getOption("wru_data_wd", default = FALSE), getwd(), tempdir())

  first_c <- readRDS(paste0(path, "/wru-data-first_c.rds"))
  mid_c <- readRDS(paste0(path, "/wru-data-mid_c.rds"))
  if(census.surname){
    last_c <- readRDS(paste0(path, "/wru-data-census_last_c.rds"))
  } else {
    last_c <- readRDS(paste0(path, "/wru-data-last_c.rds"))
  }
  
  p_eth <- c("c_whi", "c_bla", "c_his", "c_asi", "c_oth")
  if (is.null(table.surnames)) {
    lastNameDict <- last_c
  } else {
    lastNameDict <- table.surnames
    names(lastNameDict) <- names(last_c)
    lastNameDict[is.na(lastNameDict)] <- 0
  }
  if (is.null(table.first)) {
    firstNameDict <- first_c
  } else {
    firstNameDict <- table.first
    firstNameDict[is.na(firstNameDict)] <- 0
    names(firstNameDict) <- names(first_c)
  }
  if (is.null(table.middle)) {
    middleNameDict <- mid_c
  } else {
    middleNameDict <- table.middle
    middleNameDict[is.na(middleNameDict)] <- 0
    names(middleNameDict) <- names(mid_c)
  }

  nameDict <- list(
    "first" = firstNameDict,
    "middle" = middleNameDict,
    "last" = lastNameDict
  )

  ## Convert names in voter file to upper case
  df <- voter.file

  df$lastname.match <- df$lastname.upper <- toupper(as.character(df$surname))
  if (grepl("first", namesToUse)) {
    df$firstname.match <- df$firstname.upper <- toupper(as.character(df$first))
  }
  if (grepl("middle", namesToUse)) {
    df$middlename.match <- df$middlename.upper <- toupper(as.character(df$middle))
    df$middlename.match[is.na(df$middlename.match)] <- ""
  }

  ## Merge Surnames with Census List (No Cleaning Yet)
  df <- merge(df, lastNameDict, by.x = "lastname.match", by.y = "last_name", all.x = TRUE, sort = FALSE)
  if (grepl("first", namesToUse)) {
    df <- merge(df, firstNameDict, by.x = "firstname.match", by.y = "first_name", all.x = TRUE, sort = FALSE)
  }
  if (grepl("middle", namesToUse)) {
    df <- merge(df, middleNameDict, by.x = "middlename.match", by.y = "middle_name", all.x = TRUE, sort = FALSE)
  }

  if (namesToUse == "surname" && sum(!(df$lastname.upper %in% lastNameDict$last_name)) == 0) {
    return(df[, c(names(voter.file), "lastname.match", paste0(p_eth, "_last"))])
  }
  if (namesToUse == "surname, first" && sum(!(df$lastname.match %in% lastNameDict$last_name)) == 0 &&
    sum(!(df$firstname.upper %in% firstNameDict$first_name)) == 0) {
    return(df[, c(names(voter.file), "lastname.match", "firstname.match", paste0(p_eth, "_last"), paste0(p_eth, "_first"))])
  }
  if (namesToUse == "surname, first, middle" && sum(!(df$lastname.match %in% lastNameDict$last_name)) == 0 &&
    sum(!(df$firstname.upper %in% firstNameDict$first_name)) == 0 && sum(!(df$middlename.upper %in% middleNameDict$middle_name)) == 0) {
    return(df[, c(names(voter.file), "lastname.match", "firstname.match", "middlename.match", paste0(p_eth, "_last"), paste0(p_eth, "_first"), paste0(p_eth, "_middle"))])
  }

  ## Clean names (if specified by user)
  if (clean.names) {
    for (nameType in strsplit(namesToUse, ", ")[[1]]) {
      if(nameType=="surname"){
        nameType <- "last"
      }
      df1 <- df[!is.na(df[, paste("c_whi_", nameType, sep = "")]), ] # Matched names
      df2 <- df[is.na(df[, paste("c_whi_", nameType, sep = "")]), ] # Unmatched names

      ## Remove All Punctuation and Try Merge Again
      if (nrow(df2) > 0) {
        df2[, paste(nameType, "name.match", sep = "")] <- gsub("[^[:alnum:] ]", "", df2[, paste(nameType, "name.upper", sep = "")])

        df2 <- merge(df2[, !grepl(paste("_", nameType, sep = ""), names(df2))], nameDict[[nameType]],
          all.x = TRUE,
          by.x = paste(nameType, "name.match", sep = ""), by.y = paste(nameType, "name", sep = "_"),
          sort = FALSE
        )
        df2 <- df2[, names(df1)] # reorder the columns

        if (sum(!is.na(df2[, paste("c_whi_", nameType, sep = ""), ])) > 0) {
          df1 <- rbind(df1, df2[!is.na(df2[, paste("c_whi_", nameType, sep = ""), ]), ])
          df2 <- df2[is.na(df2[, paste("c_whi_", nameType, sep = "")]), ]
        }
      }

      ## Remove All Spaces and Try Merge Again
      if (nrow(df2) > 0) {
        df2[, paste(nameType, "name.match", sep = "")] <- gsub(" ", "", df2[, paste(nameType, "name.match", sep = "")])
        df2 <- merge(df2[, !grepl(paste("_", nameType, sep = ""), names(df2))], nameDict[[nameType]],
          all.x = TRUE,
          by.x = paste(nameType, "name.match", sep = ""), by.y = paste(nameType, "name", sep = "_"),
          sort = FALSE
        )
        df2 <- df2[, names(df1)] # reorder the columns

        if (sum(!is.na(df2[, paste("c_whi_", nameType, sep = ""), ])) > 0) {
          df1 <- rbind(df1, df2[!is.na(df2[, paste("c_whi_", nameType, sep = ""), ]), ])
          df2 <- df2[is.na(df2[, paste("c_whi_", nameType, sep = "")]), ]
        }
      }

      # Edits specific to common issues with last names
      if (nameType == "last" & nrow(df2) > 0) {

        ## Remove Jr/Sr/III Suffixes for last names
        suffix <- c("JUNIOR", "SENIOR", "THIRD", "III", "JR", " II", " J R", " S R", " IV")
        for (i in 1:length(suffix)) {
          df2$lastname.match <- ifelse(substr(df2$lastname.match, nchar(df2$lastname.match) - (nchar(suffix)[i] - 1), nchar(df2$lastname.match)) == suffix[i],
            substr(df2$lastname.match, 1, nchar(df2$lastname.match) - nchar(suffix)[i]),
            df2$lastname.match
          )
        }
        df2$lastname.match <- ifelse(nchar(df2$lastname.match) >= 7,
          ifelse(substr(df2$lastname.match, nchar(df2$lastname.match) - 1, nchar(df2$lastname.match)) == "SR",
            substr(df2$lastname.match, 1, nchar(df2$lastname.match) - 2),
            df2$lastname.match
          ),
          df2$lastname.match
        ) # Remove "SR" only if name has at least 7 characters

        df2 <- merge(
          df2[, !grepl(paste("_", nameType, sep = ""), names(df2))], 
          lastNameDict, by.x = "lastname.match", by.y = "last_name", 
          all.x = TRUE, sort = FALSE)
        df2 <- df2[, names(df1)] # reorder the columns

        if (sum(!is.na(df2[, paste("c_whi_", nameType, sep = ""), ])) > 0) {
          df1 <- rbind(df1, df2[!is.na(df2[, paste("c_whi_", nameType, sep = ""), ]), ])
          df2 <- df2[is.na(df2[, paste("c_whi_", nameType, sep = "")]), ]
        }
      }


      ## Names with Hyphens or Spaces, e.g. Double-Barreled Names
      if (nrow(df2) > 0) {
        df2$name2 <- df2$name1 <- NA
        df2$name1[grep("-", df2[, paste(nameType, "name.upper", sep = "")])] <- sapply(strsplit(grep("-", df2[, paste(nameType, "name.upper", sep = "")], value = T), "-"), "[", 1)
        df2$name2[grep("-", df2[, paste(nameType, "name.upper", sep = "")])] <- sapply(strsplit(grep("-", df2[, paste(nameType, "name.upper", sep = "")], value = T), "-"), "[", 2)
        df2$name1[grep(" ", df2[, paste(nameType, "name.upper", sep = "")])] <- sapply(strsplit(grep(" ", df2[, paste(nameType, "name.upper", sep = "")], value = T), " "), "[", 1)
        df2$name2[grep(" ", df2[, paste(nameType, "name.upper", sep = "")])] <- sapply(strsplit(grep(" ", df2[, paste(nameType, "name.upper", sep = "")], value = T), " "), "[", 2)

        ## Use first half of name to merge in priors
        df2[, paste(nameType, "name.match", sep = "")] <- as.character(df2$name1)
        df2 <- merge(df2[, !grepl(paste("_", nameType, sep = ""), names(df2))], nameDict[[nameType]],
          all.x = TRUE,
          by.x = paste(nameType, "name.match", sep = ""), by.y = paste(nameType, "name", sep = "_"),
          sort = FALSE
        )
        df2 <- df2[, c(names(df1), "name1", "name2")] # reorder the columns

        if (sum(!is.na(df2[, paste("c_whi_", nameType, sep = ""), ])) > 0) {
          df1 <- rbind(df1, df2[!is.na(df2[, paste("c_whi_", nameType, sep = "")]), !(names(df2) %in% c("name1", "name2"))])
          df2 <- df2[is.na(df2[, paste("c_whi_", nameType, sep = "")]), ]
        }
      }

      ## Use second half of name to merge in priors for rest
      if (nrow(df2) > 0) {
        df2[, paste(nameType, "name.match", sep = "")] <- as.character(df2$name2)
        df2 <- merge(df2[, !grepl(paste("_", nameType, sep = ""), names(df2))], nameDict[[nameType]],
          all.x = TRUE,
          by.x = paste(nameType, "name.match", sep = ""), by.y = paste(nameType, "name", sep = "_"),
          sort = FALSE
        )
        df2 <- df2[, c(names(df1), "name1", "name2")] # reorder the columns

        if (sum(!is.na(df2[, paste("c_whi_", nameType, sep = ""), ])) > 0) {
          df1 <- rbind(df1, df2[!is.na(df2[, paste("c_whi_", nameType, sep = "")]), !(names(df2) %in% c("name1", "name2"))])
          df2 <- df2[is.na(df2[, paste("c_whi_", nameType, sep = "")]), ]
        }
      }

      if (nrow(df2) > 0) {
        df <- rbind(df1, df2[, !(names(df2) %in% c("name1", "name2"))])
      } else {
        df <- df1
      }
    }
  }


  ## For unmatched names, just fill with an column mean if impute is true, or with constant if false
  c_miss_last <- mean(is.na(df$c_whi_last))
  if (c_miss_last > 0) {
    message(paste(paste(sum(is.na(df$c_whi_last)), " (", round(100 * mean(is.na(df$c_whi_last)), 1), "%) individuals' last names were not matched.", sep = "")))
  }
  if (grepl("first", namesToUse)) {
    c_miss_first <- mean(is.na(df$c_whi_first))
    if (c_miss_first > 0) {
      message(paste(paste(sum(is.na(df$c_whi_first)), " (", round(100 * mean(is.na(df$c_whi_first)), 1), "%) individuals' first names were not matched.", sep = "")))
    }
  }
  if (grepl("middle", namesToUse)) {
    c_miss_mid <- mean(is.na(df$c_whi_middle))
    if (c_miss_mid > 0) {
      message(paste(paste(sum(is.na(df$c_whi_middle)), " (", round(100 * mean(is.na(df$c_whi_middle)), 1), "%) individuals' middle names were not matched.", sep = "")))
    }
  }

  if (impute.missing) {
    impute.vec <- colMeans(df[, grep("c_", names(df), value = TRUE)], na.rm = TRUE)
    for (i in grep("c_", names(df), value = TRUE)) {
      df[, i] <- dplyr::coalesce(df[, i], impute.vec[i])
    }
  } else {
    for (i in grep("c_", names(df), value = TRUE)) {
      df[, i] <- dplyr::coalesce(df[, i], 1)
    }
  }

  # return the data
  if (namesToUse == "surname") {
    return(df[, c(names(voter.file), "lastname.match", paste(p_eth, "last", sep = "_"))])
  } else if (namesToUse == "surname, first") {
    return(df[, c(
      names(voter.file), "lastname.match", "firstname.match",
      paste(p_eth, "last", sep = "_"), paste(p_eth, "first", sep = "_")
    )])
  } else if (namesToUse == "surname, first, middle") {
    return(df[, c(
      names(voter.file), "lastname.match", "firstname.match", "middlename.match",
      paste(p_eth, "last", sep = "_"), paste(p_eth, "first", sep = "_"), paste(p_eth, "middle", sep = "_")
    )])
  }
}


#' Preflight for name data
#'
#' Checks if namedata is available in the current working directory, if not
#' downloads it from github using piggyback. By default, wru will download the
#' data to a temporary directory that lasts as long as your session does.
#' However, you may wish to set the \code{wru_data_wd} option to save the 
#' downloaded data to your current working directory for more permanence. 
#'
#' @importFrom piggyback pb_download
wru_data_preflight <- function() {
  dest <- ifelse(getOption("wru_data_wd", default = FALSE), getwd(), tempdir())
  tryCatch(
    # Oddity of conditions for .token. Ignores token if is ""
    piggyback::pb_download(repo = "kosukeimai/wru", dest = dest, .token = "", tag = "v2.0.0"), 
    error = function(e) message("There was an error retrieving data: ", e$message)
  )
}

#' Dataset with FIPS codes for US states
#'
#' Dataset including FIPS codes and postal abbreviations for each U.S. state,
#' district, and territory.
#'
#' @format
#' A tibble with 57 rows and 3 columns:
#' \describe{
#'   \item{`state`}{Two-letter postal abbreviation}
#'   \item{`state_code`}{Two-digit FIPS code}
#'   \item{`state_name`}{English name}
#' }
#' @source Derived from [tidycensus::fips_codes()]
"state_fips"

#' Convert between state names, postal abbreviations, and FIPS codes
#'
#' @param x A [numeric] or [character] vector of state names,
#'   postal abbreviations, or FIPS codes.
#'   Matches for state names and abbreviations are not case sensitive.
#'   FIPS codes may be matched from numeric or character vectors,
#'   with or without leading zeroes.
#'
#' @return
#' \describe{
#'   \item{`as_state_fips_code()`}{
#'     A [character] vector of two-digit FIPS codes.
#'     One-digit FIPS codes are prefixed with a leading zero,
#'     e.g., `"06"` for California.
#'   }
#'   \item{`as_state_abbreviation()`}{
#'     A [character] vector of two-letter postal abbreviations,
#'     e.g., `"CA"` for California.
#'   }
#' }
#'
#' @examples
#' as_fips_code("california")
#' as_state_abbreviation("california")
#' 
#' # Character vector matches ignore case
#' as_fips_code(c("DC", "Md", "va"))
#' as_state_abbreviation(c("district of columbia", "Maryland", "VIRGINIA"))
#' 
#' # Note that `3` and `7` are standardized to `NA`,
#' # because no state is assigned those FIPS codes
#' as_fips_code(1:10)
#' as_state_abbreviation(1:10)
#' 
#' # You can even mix methods in the same vector
#' as_fips_code(c("utah", "NM", 8, "04"))
#' as_state_abbreviation(c("utah", "NM", 8, "04"))
#'
#' @keywords internal
#' @export
as_fips_code <- function(x) {
  state_fips <- wru::state_fips
  state_fips$state_code[
    dplyr::coalesce(
      match(toupper(x), state_fips$state),
      match(tolower(x), tolower(state_fips$state_name)),
      match(suppressWarnings(as.numeric(x)), as.numeric(state_fips$state_code))
    )
  ]
}

#' @rdname as_fips_code
#' @export
as_state_abbreviation <- function(x) {
  state_fips <- wru::state_fips
  state_fips$state[
    dplyr::coalesce(
      match(toupper(x), state_fips$state),
      match(tolower(x), tolower(state_fips$state_name)),
      match(suppressWarnings(as.numeric(x)), as.numeric(state_fips$state_code))
    )
  ]
}

#' Census API function.
#'
#' \code{get_census_api} obtains U.S. Census data via the public API.
#'
#' This function obtains U.S. Census data via the public API. User
#' can specify the variables and region(s) for which to obtain data.
#'
#' @inheritParams get_census_data
#' @param data_url URL root of the API,
#'  e.g., \code{"https://api.census.gov/data/2020/dec/pl"}.
#' @param var.names A character vector of variables to get,
#'  e.g., \code{c("P2_005N", "P2_006N", "P2_007N", "P2_008N")}.
#'  If there are more than 50 variables, then function will automatically
#'  split variables into separate queries.
#' @param region Character object specifying which region to obtain data for.
#'  Must contain "for" and possibly "in",
#'  e.g., \code{"for=block:1213&in=state:47+county:015+tract:*"}.
#' @param retry The number of retries at the census website if network interruption occurs.
#' @return If successful, output will be an object of class \code{data.frame}.
#'  If unsuccessful, function prints the URL query that caused the error.
#'
#' @examples
#' \dontrun{
#' get_census_api(
#'   data_url = "https://api.census.gov/data/2020/dec/pl",
#'   var.names = c("P2_005N", "P2_006N", "P2_007N", "P2_008N"), region = "for=county:*&in=state:34"
#' )
#' }
#'
#' @references
#' Based on code authored by Nicholas Nagle, which is available
#' \href{https://rstudio-pubs-static.s3.amazonaws.com/19337_2e7f827190514c569ea136db788ce850.html}{here}.
#'
#' @keywords internal
get_census_api <- function(
    data_url,
    key = Sys.getenv("CENSUS_API_KEY"),
    var.names,
    region,
    retry = 0
) {
  if (length(var.names) > 50) {
    var.names <- vec_to_chunk(var.names) # Split variables into a list
    get <- lapply(var.names, function(x) paste(x, sep = "", collapse = ","))
    data <- lapply(
      var.names,
      function(x) get_census_api_2(data_url, key, x, region, retry)
    )
  } else {
    get <- paste(var.names, sep = "", collapse = ",")
    data <- list(get_census_api_2(data_url, key, get, region, retry))
  }

  ## Format output. If there were no errors, than paste the data together. If there is an error, just return the unformatted list.
  if (all(sapply(data, is.data.frame))) {
    colnames <- unlist(lapply(data, names))
    data <- do.call(cbind, data)
    names(data) <- colnames
    ## Prettify the output and remove any non-unique columns
    data <- data[, unique(colnames, fromLast = TRUE)]
    ## Reorder columns so that numeric fields follow non-numeric fields
    data <- data[, c(which(sapply(data, class) != "numeric"), which(sapply(data, class) == "numeric"))]
    return(data)
  } else {
    message("Unable to create single data.frame in get_census_api")
    return(data)
  }
}

#' Census API URL assembler.
#'
#' \code{get_census_api_2} assembles URL components for \code{get_census_api}.
#' 
#' This function assembles the URL components and sends the request to the Census server. 
#' It is used by the \code{get_census_api} function. The user should not need to call this 
#' function directly.
#'
#' @inheritParams get_census_data
#' @param data_url URL root of the API,
#'  e.g., \code{"https://api.census.gov/data/2020/dec/pl"}.
#' @param get A character vector of variables to get, 
#'  e.g., \code{c("P2_005N", "P2_006N", "P2_007N", "P2_008N")}.
#'  If there are more than 50 variables, then function will automatically 
#'  split variables into separate queries.
#' @param region Character object specifying which region to obtain data for.
#'  Must contain "for" and possibly "in", 
#'  e.g., \code{"for=block:1213&in=state:47+county:015+tract:*"}.
#' @param retry The number of retries at the census website if network interruption occurs.
#' @return If successful, output will be an object of class \code{data.frame}. 
#'  If unsuccessful, function prints the URL query that was constructed.
#'
#' @examples
#' \dontrun{try(get_census_api_2(data_url = "https://api.census.gov/data/2020/dec/pl",
#' get = c("P2_005N", "P2_006N", "P2_007N", "P2_008N"), region = "for=county:*&in=state:34"))}
#' 
#' @references
#' Based on code authored by Nicholas Nagle, which is available 
#' \href{https://rstudio-pubs-static.s3.amazonaws.com/19337_2e7f827190514c569ea136db788ce850.html}{here}.
#'
#' @keywords internal
get_census_api_2 <- function(
    data_url,
    key = Sys.getenv("CENSUS_API_KEY"),
    get,
    region,
    retry = 3
){
  if(length(get) > 1) {
    get <- paste(get, collapse=',', sep='')
  }
  
  api_call <- paste(data_url, 'key=', key, '&get=', get, '&', region, sep='')
  
  dat_raw <- try(readLines(api_call, warn="F"))

  while (inherits(dat_raw, "try-error") && (retry > 0)) {
    message(paste("Try census server again:", data_url))
    Sys.sleep(1)
    retry <- retry - 1
    dat_raw <- try(readLines(api_call, warn="F"))
  }

  if (inherits(dat_raw, "try-error")) {
    message("Data access failure at the census website, please try again by re-run the previous command")
    stop(message(api_call))
    return()
  }
  if (!inherits(dat_raw, "try-error") & "TRUE" %in% names(table(grepl("Invalid Key", dat_raw)))) {
    stop('Invalid Key: 
         A valid key must be included with each data API request. 
         You included a key with this request, however, it is not valid. 
         Please check your key and try again.'
         )
  }
  dat_df <- data.frame()

  ## Split the datastream into a list with each row as an element.
  tmp <- strsplit(gsub("[^[:alnum:], _]", '', dat_raw), "\\,")
  dat_df <- as.data.frame(do.call(rbind, tmp[-1]), stringsAsFactors=FALSE)
  names(dat_df) <- tmp[[1]]
  ## Convert to numeric
  value_cols <- grep("[0-9]", names(dat_df), value=TRUE)
  for(col in value_cols) dat_df[,col] <- as.numeric(as.character(dat_df[,col]))
  return(dat_df)
}

#' Internal model fitting functions
#'
#' These functions are intended for internal use only. Users should use the
#' [predict_race()] interface rather any of these functions directly.
#'
#' These functions fit different versions of WRU. \code{.predict_race_old} fits
#' the original WRU model, also known as BISG with census-based surname dictionary.
#' \code{.predict_race_new} fits a new version of BISG which uses a new, augmented
#' surname dictionary, and can also accommodate the use of first and middle
#' name information. Finally, \code{.predict_race_me} fits a fully Bayesian Improved
#' Surname Geocoding model (fBISG), which fits a model with measurement-error
#' correction of erroneous zeros in census tables, in addition to also accommodating
#' the augmented surname dictionary, and the first and middle name
#' dictionaries when making predictions.
#'
#' @inheritParams predict_race
#' @param voter.file See documentation in \code{race_predict}.
#' @param census.surname See documentation in \code{race_predict}.
#' @param surname.only See documentation in \code{race_predict}.
#' @param surname.year See documentation in \code{race_predict}.
#' @param census.geo See documentation in \code{race_predict}.
#' @param census.data See documentation in \code{race_predict}.
#' @param age See documentation in \code{race_predict}.
#' @param sex See documentation in \code{race_predict}.
#' @param year See documentation in \code{race_predict}.
#' @param party See documentation in \code{race_predict}.
#' @param retry See documentation in \code{race_predict}.
#' @param impute.missing See documentation in \code{race_predict}.
#' @param skip_bad_geos See documentation in \code{race_predict}.
#' @param names.to.use See documentation in \code{race_predict}.
#' @param race.init See documentation in \code{race_predict}.
#' @param name.dictionaries See documentation in \code{race_predict}.
#' @param ctrl See `control` in documentation for [predict_race()].
#' @param use.counties A logical, defaulting to FALSE. Should census data be filtered by counties available in \var{census.data}?
#'
#' @inherit predict_race return
#'
#' @name modfuns
NULL

#' @section .predict_race_old:
#' Original WRU race prediction function, implementing classical BISG with census-based
#' surname dictionary.
#' @importFrom stats rmultinom
#' @importFrom utils txtProgressBar setTxtProgressBar
#' @rdname modfuns
#' @keywords internal

.predict_race_old <- function(
    voter.file,
    census.surname = TRUE,
    surname.only = FALSE,
    surname.year = 2020,
    name.dictionaries = NULL,
    census.geo,
    census.key = Sys.getenv("CENSUS_API_KEY"),
    census.data = NULL,
    age = FALSE,
    sex = FALSE,
    year = "2020",
    party,
    retry = 3,
    impute.missing = TRUE,
    use.counties = FALSE
) {
  
  # warning: 2020 census data only support prediction when both age and sex are equal to FALSE
  if ((sex == TRUE || age == TRUE) && (year == "2020")) {
    stop("Warning: only predictions with both age and sex equal to FALSE are supported when using 2020 census data.")
  }
  
  if (!missing(census.geo) && (census.geo == "precinct")) {
    # geo <- "precinct"
    stop("Error: census_helper function does not currently support merging precinct-level data.")
  }
  
  vars.orig <- names(voter.file)
  
  if (surname.only == TRUE) {
    message("Proceeding with surname-only predictions...")
    if (!("surname" %in% names(voter.file))) {
      stop("Voter data frame needs to have a column named surname")
    }
  } else {
    if (missing(census.geo) || is.null(census.geo) || all(is.na(census.geo)) || census.geo %in% c("county", "tract", "block", "place") == FALSE) {
      stop("census.geo must be either 'county', 'tract', 'block', or 'place'")
    } else {
      message(paste("Proceeding with Census geographic data at", census.geo, "level..."))
    }
    if (missing(census.data) || is.null(census.data) || all(is.na(census.data))) {
      census.key <- validate_key(census.key)
      message("Downloading Census geographic data using provided API key...")
    } else {
      if (!("state" %in% names(voter.file))) {
        stop("voter.file object needs to have a column named state.")
      }
      if (sum(toupper(unique(as.character(voter.file$state))) %in% toupper(names(census.data)) == FALSE) > 0) {
        message("census.data object does not include all states in voter.file object.")
        census.key <- validate_key(census.key)
        message("Downloading Census geographic data for states not included in census.data object...")
      } else {
        message("Using Census geographic data from provided census.data object...")
      }
    }
  }
  
  eth <- c("whi", "bla", "his", "asi", "oth")
  
  ## Merge in Pr(Race | Surname) if necessary
  if (census.surname) {
    if (!(surname.year %in% c(2000, 2010, 2020))) {
      stop(paste(surname.year, "is not a valid surname.year. It should be 2000, 2010 or 2020 (default)."))
    }
    voter.file <- merge_surnames(voter.file, surname.year = surname.year, name.data = NULL, impute.missing = impute.missing)
  } else {
    # Check if voter.file has the necessary data
    if (is.null(name.dictionaries) | !("surname" %in% names(name.dictionaries))) {
      stop("User must provide a 'name.dictionaries', with named element 'surname'.")
    }
    for (k in 1:length(eth)) {
      if ((paste("c", eth[k], sep = "_") %in% names(name.dictionaries[["surname"]])) == FALSE) {
        stop(paste("name.dictionaries element 'surname' needs to have columns named ", paste(paste("c", eth, sep = "_"), collapse = " and "), ".", sep = ""))
      }
    }
    name.dictionaries[["surname"]] <- apply(name.dictionaries[["surname"]], 1, function(x) x / sum(x, na.rm = TRUE))
    name.dictionaries[["surname"]][is.na(name.dictionaries[["surname"]])] <- 0
    voter.file <- merge_surnames(voter.file, surname.year = surname.year, name.data = name.dictionaries[["surname"]], impute.missing = impute.missing)
  }
  
  ## Surname-Only Predictions
  if (surname.only) {
    for (k in 1:length(eth)) {
      voter.file[paste("pred", eth[k], sep = ".")] <- voter.file[paste("p", eth[k], sep = "_")] / apply(voter.file[paste("p", eth, sep = "_")], 1, sum)
    }
    pred <- paste("pred", eth, sep = ".")
    return(voter.file[c(vars.orig, pred)])
  }
  
  ## Merge in Pr(Party | Race) if necessary
  if (missing(party) == FALSE) {
    voter.file$PID <- voter.file[, party]
    voter.file <- merge(voter.file, get("pid")[names(get("pid")) %in% "party" == F], by = "PID", all.x = TRUE)
  }
  
  if (census.geo == "place") {
    if (!("place" %in% names(voter.file))) {
      stop("voter.file object needs to have a column named place.")
    }
    voter.file <- census_helper(
      key = census.key,
      voter.file = voter.file,
      states = "all",
      geo = "place",
      age = age,
      sex = sex,
      year = year,
      census.data = census.data,
      retry = retry
    )
  }
  
  if (census.geo == "block_group") {
    if (!("block_group" %in% names(voter.file)) || !("county" %in% names(voter.file)) || !("tract" %in% names(voter.file))) {
      stop("voter.file object needs to have columns named block, tract, and county.")
    }
    voter.file <- census_helper(
      key = census.key,
      voter.file = voter.file,
      states = "all",
      geo = "block_group",
      age = age,
      sex = sex,
      year = year,
      census.data = census.data,
      retry = retry,
      use.counties = use.counties
    )
  }
  
  if (census.geo == "block") {
    if (!("tract" %in% names(voter.file)) || !("county" %in% names(voter.file)) || !("block" %in% names(voter.file))) {
      stop("voter.file object needs to have columns named block, tract, and county.")
    }
    voter.file <- census_helper(
      key = census.key,
      voter.file = voter.file,
      states = "all",
      geo = "block",
      age = age,
      sex = sex,
      year = year,
      census.data = census.data,
      retry = retry,
      use.counties = use.counties
    )
  }
  
  if (census.geo == "precinct") {
    geo <- "precinct"
    stop("Error: census_helper function does not currently support precinct-level data.")
  }
  
  if (census.geo == "tract") {
    if (!("tract" %in% names(voter.file)) || !("county" %in% names(voter.file))) {
      stop("voter.file object needs to have columns named tract and county.")
    }
    voter.file <- census_helper(
      key = census.key,
      voter.file = voter.file,
      states = "all",
      geo = "tract",
      age = age,
      sex = sex,
      year = year,
      census.data = census.data, 
      retry = retry,
      use.counties = use.counties
    )
  }
  
  if (census.geo == "county") {
    if (!("county" %in% names(voter.file))) {
      stop("voter.file object needs to have a column named county.")
    }
    voter.file <- census_helper(
      key = census.key,
      voter.file = voter.file,
      states = "all",
      geo = "county",
      age = age,
      sex = sex,
      year = year,
      census.data = census.data, 
      retry = retry
    )
  }
  
  ## Pr(Race | Surname, Geolocation)
  if (missing(party)) {
    for (k in 1:length(eth)) {
      voter.file[paste("u", eth[k], sep = "_")] <- voter.file[paste("p", eth[k], sep = "_")] * voter.file[paste("r", eth[k], sep = "_")]
    }
    voter.file$u_tot <- apply(voter.file[paste("u", eth, sep = "_")], 1, sum, na.rm = TRUE)
    for (k in 1:length(eth)) {
      voter.file[paste("q", eth[k], sep = "_")] <- voter.file[paste("u", eth[k], sep = "_")] / voter.file$u_tot
    }
  }
  
  ## Pr(Race | Surname, Geolocation, Party)
  if (missing(party) == FALSE) {
    for (k in 1:length(eth)) {
      voter.file[paste("u", eth[k], sep = "_")] <- voter.file[paste("p", eth[k], sep = "_")] * voter.file[paste("r", eth[k], sep = "_")] * voter.file[paste("r_pid", eth[k], sep = "_")]
    }
    voter.file$u_tot <- apply(voter.file[paste("u", eth, sep = "_")], 1, sum, na.rm = TRUE)
    for (k in 1:length(eth)) {
      voter.file[paste("q", eth[k], sep = "_")] <- voter.file[paste("u", eth[k], sep = "_")] / voter.file$u_tot
    }
  }
  
  for (k in 1:length(eth)) {
    voter.file[paste("pred", eth[k], sep = ".")] <- voter.file[paste("q", eth[k], sep = "_")]
  }
  pred <- paste("pred", eth, sep = ".")
  
  return(voter.file[c(vars.orig, pred)])
}

#' @section .predict_race_new :
#' New race prediction function, implementing classical BISG with augmented
#' surname dictionary, as well as first and middle name information.
#' @rdname modfuns

predict_race_new <- function(
    voter.file,
    names.to.use,
    year = "2020",
    age = FALSE,
    sex = FALSE,
    census.geo = c("tract", "block", "block_group", "county", "place", "zcta"),
    census.key = Sys.getenv("CENSUS_API_KEY"),
    name.dictionaries,
    surname.only=FALSE,
    census.data = NULL,
    retry = 0,
    impute.missing = TRUE,
    skip_bad_geos = FALSE,
    census.surname = FALSE,
    use.counties = FALSE
) {
  
  # Check years
  if (!(year %in% c("2000", "2010", "2020"))){
    stop("Year should be one of 2000, 2010, or 2020 (default).")
  }
  # Define 2020 race marginal
  race.margin <- c(r_whi=0.5783619, r_bla=0.1205021, r_his=0.1872988,
                   r_asi=0.06106737, r_oth=0.05276981)
  
  census.geo <- tolower(census.geo)
  census.geo <- rlang::arg_match(census.geo)
  
  vars.orig <- names(voter.file)
  
  # check the names
  if (names.to.use == "surname") {
    message("Proceeding with last name predictions...")
    if (!("surname" %in% names(voter.file))) {
      stop("Voter data frame needs to have a column named 'surname'.")
    }
  } else if (names.to.use == "surname, first") {
    message("Proceeding with first and last name-only predictions...")
    if (!("surname" %in% names(voter.file)) || !("first" %in% names(voter.file))) {
      stop("Voter data frame needs to have a column named 'surname' and a column called 'first'.")
    }
  } else if (names.to.use == "surname, first, middle") {
    message("Proceeding with first, last, and middle name predictions...")
    if (!("surname" %in% names(voter.file)) || !("first" %in% names(voter.file)) ||
        !("middle" %in% names(voter.file))) {
      stop("Voter data frame needs to have a column named 'surname', a column called 'first', and a column called 'middle'.")
    }
  }
  
  ## Preliminary Data quality checks
  wru_data_preflight()

  path <- ifelse(getOption("wru_data_wd", default = FALSE), getwd(), tempdir())

  first_c <- readRDS(paste0(path, "/wru-data-first_c.rds"))
  mid_c <- readRDS(paste0(path, "/wru-data-mid_c.rds"))
  if(census.surname){
    last_c <- readRDS(paste0(path, "/wru-data-census_last_c.rds"))
  } else {
    last_c <- readRDS(paste0(path, "/wru-data-last_c.rds"))
  }
  if (any(!is.null(name.dictionaries))) {
    if (!is.null(name.dictionaries[["surname"]])) {
      stopifnot(identical(names(name.dictionaries[["surname"]]), names(last_c)))
    }
    if (!is.null(name.dictionaries[["first"]])) {
      stopifnot(identical(names(name.dictionaries[["first"]]), names(first_c)))
    }
    if (!is.null(name.dictionaries[["middle"]])) {
      stopifnot(identical(names(name.dictionaries[["middle"]]), names(mid_c)))
    }
  }
  
  # check the geographies
  if (surname.only == FALSE) {
    message("Proceeding with Census geographic data at ", census.geo, " level...")

    if (is.null(census.data)) {
      census.key <- validate_key(census.key)
      message("Downloading Census geographic data using provided API key...")
    } else {
      if (!("state" %in% names(voter.file))) {
        stop("voter.file object needs to have a column named state.")
      }
      census_data_preflight(census.data, census.geo, year)
      if (sum(toupper(unique(as.character(voter.file$state))) %in% toupper(names(census.data)) == FALSE) > 0) {
        message("census.data object does not include all states in voter.file object.")
        census.key <- validate_key(census.key)
        message("Downloading Census geographic data for states not included in census.data object...")
      } else {
        message("Using Census geographic data from provided census.data object...")
      }
    }
    
    geo_id_names <- determine_geo_id_names(census.geo)
    
    if (!all(geo_id_names %in% names(voter.file))) {
      stop(message("To use",census.geo,"as census.geo, voter.file needs to include the following column(s):",
               paste(geo_id_names, collapse=", ")))
    }
    
    voter.file <- census_helper_new(
      key = census.key,
      voter.file = voter.file,
      states = "all",
      geo = census.geo,
      age = age, 
      sex = sex,
      year = year,
      census.data = census.data, 
      retry = retry,
      use.counties = use.counties,
      skip_bad_geos = skip_bad_geos
    )
  }
  
  eth <- c("whi", "bla", "his", "asi", "oth")
  
  ## Merge in Pr(Name | Race)
  voter.file <- merge_names(voter.file = voter.file,
                            namesToUse = names.to.use,
                            census.surname = census.surname, 
                            table.surnames=name.dictionaries[["surname"]], 
                            table.first=name.dictionaries[["first"]],
                            table.middle=name.dictionaries[["middle"]],
                            clean.names = TRUE,
                            impute.missing = impute.missing,
                            model = 'BISG')
  
  if (surname.only == TRUE) {
    # Pr(Race | Surname)
    preds <- voter.file[, grep("_last$", names(voter.file))] * 
      matrix(race.margin, nrow=nrow(voter.file), ncol=length(race.margin), byrow = TRUE)
  } else {
    # Pr(Race | Surname, Geolocation)
    preds <-  voter.file[, grep("_last$", names(voter.file))]  * voter.file[, grep("^r_", names(voter.file))]
    if (grepl("first", names.to.use)) {
      preds <- preds * voter.file[, grep("_first$", names(voter.file))]
    }
    if (grepl("middle", names.to.use)) {
      preds <- preds * voter.file[, grep("_middle$", names(voter.file))]
    }
  }
  
  if(impute.missing){
   for(i in ncol(preds)){
     preds[, i] <- dplyr::coalesce(preds[, i], race.margin[i])
   }
  }
  ## Normalize (recycle marginal)
  preds <- preds/rowSums(preds)
  ## Revert to national Pr(Race) for missing predictions
  colnames(preds) <- paste("pred", eth, sep = ".")
  
  return(data.frame(cbind(voter.file[c(vars.orig)], preds)))
}


#' @section .predict_race_me:
#' New race prediction function, implementing fBISG (i.e. measurement
#' error correction, fully Bayesian model) with augmented
#' surname dictionary, as well as first and middle name information.
#' @importFrom dplyr pull
#' @rdname modfuns

predict_race_me <- function(
    voter.file,
    names.to.use,
    year = "2020",
    age = FALSE,
    sex = FALSE, 
    census.geo = c("tract", "block", "block_group", "county", "place", "zcta"),
    census.key = Sys.getenv("CENSUS_API_KEY"),
    name.dictionaries,
    surname.only = FALSE,
    census.data = NULL,
    retry = 0,
    impute.missing = TRUE,
    census.surname = FALSE,
    use.counties = FALSE,
    race.init,
    ctrl
) {
  census.geo <- tolower(census.geo)
  census.geo <- rlang::arg_match(census.geo)
  
  if(!is.null(census.data)) {
    census_data_preflight(census.data, census.geo, year)
  }
  
  n_race <- 5
  if (!(names.to.use %in% c(c("surname"), c("surname, first"), c("surname, first, middle")))) {
    stop("'names.to.use' must be one of 'surname', 'surname, first', or 'surname, first, middle'")
  }
  name_types <- gsub(" ", "", strsplit(names.to.use, ",")[[1]])
  if (!all(name_types %in% names(voter.file))) {
    stop("When used, 'surname', 'first', and 'middle' must be variable names in voter.file.")
  }
  
  ## Preliminary Data quality checks
  wru_data_preflight()
  path <- ifelse(getOption("wru_data_wd", default = FALSE), getwd(), tempdir())
  
  if(census.surname){
    last_c <- readRDS(paste0(path, "/wru-data-census_last_c.rds"))
  } else {
    last_c <- readRDS(paste0(path, "/wru-data-last_c.rds"))
  }
  if (!is.null(name.dictionaries[["surname"]])) {
    stopifnot(identical(names(name.dictionaries[["surname"]]), names(last_c)))
    last_c <- name.dictionaries[["surname"]]
  } 
  if("first" %in% name_types){
    first_c <- readRDS(paste0(path, "/wru-data-first_c.rds"))
    if (!is.null(name.dictionaries[["first"]])){
      stopifnot(identical(names(name.dictionaries[["first"]]), names(first_c)))
      first_c <- name.dictionaries[["first"]]
    }
  } 
  if("middle" %in% name_types){
    mid_c <- readRDS(paste0(path, "/wru-data-mid_c.rds"))
    if (!is.null(name.dictionaries[["middle"]])){
      stopifnot(identical(names(name.dictionaries[["middle"]]), names(mid_c)))
      mid_c <- name.dictionaries[["middle"]]
    }
  } 
  
  ## Other quick checks...
  stopifnot(
    all(!is.na(voter.file$surname))
  )
  
  orig.names <- names(voter.file)
  orig.state <- voter.file$state
  voter.file$state <- toupper(voter.file$state)
  voter.file$rec_id_ <- 1:nrow(voter.file)
  
  
  ## Set RNG seed
  set.seed(ctrl$seed)
  if(!(ctrl$usr_seed) & (ctrl$verbose)){
    message("fBISG relies on MCMC; for reproducibility, I am setting RNG seed and returning it as attribute 'RNGseed'.\n",
            "To silence this message, you can set a seed explicitly by defining the 'seed' element in the control list.")
  }
  
  ## Initial race
  race_pred_args <- list(
    census.data = NULL,
    names.to.use = names.to.use,
    census.geo = census.geo,
    census.key = NULL,
    model = "BISG",
    name.dictionaries = name.dictionaries,
    retry = 0
  )
  
  ## level of geo estimation
  geo_id_names <- c("state", determine_geo_id_names(census.geo))
  
  #race_pred_args[names(args_usr)] <- args_usr
  all_states <- unique(voter.file$state)
  census.data <- census.data[all_states]
  race.suff <- c("whi", "bla", "his", "asi", "oth")

  geo_id <- do.call(paste, voter.file[, geo_id_names])
  
  if (ctrl$verbose) {
    message("Forming Pr(race | location) tables from census data...\n")
  }
  
  vars_ <- census_geo_api_names(year = year)
  
  N_rg <- purrr::map(
    census.data,
    function(x) {
      all_names <- names(x[[census.geo]])
      
      if (any(c("P2_005N", "P005003") %in% all_names)) {
        vars_ <- census_geo_api_names_legacy(year = year)
      }
      
      totals <- x[[census.geo]][, match(c(geo_id_names, unlist(vars_)), all_names)]
      
      totals$r_whi <- rowSums(totals[, vars_[["r_whi"]], drop = FALSE]) # White population
      totals$r_bla <- rowSums(totals[, vars_[["r_bla"]], drop = FALSE]) # Black population
      totals$r_his <- rowSums(totals[, vars_[["r_his"]], drop = FALSE]) # Latino population
      totals$r_asi <- rowSums(totals[, vars_[["r_asi"]], drop = FALSE]) # Asian + NH/PI population
      totals$r_oth <- rowSums(totals[, vars_[["r_oth"]], drop = FALSE]) # AI/AN + Other + Mixed population

      totals <- totals[, -match(unlist(vars_), names(totals))]
      totals
    }
  )
  N_rg <- dplyr::bind_rows(N_rg)
  N_rg_geo <- do.call(paste, N_rg[, geo_id_names])
  ## Subset to geo's in vf
  N_rg <- N_rg[N_rg_geo %in% geo_id, ]
  rm(race_pred_args)
  if (nrow(N_rg) != length(unique(geo_id))) {
    stop("Some records in voter.file have geographic locations that I wasn't able to find in the census.data.\n
          Records may have mis-matched geographic units that do not exist in the census.")
  }
  
  
  ## Split data by geographic cluster
  voter.file$state_cluster <- geo_id
  N_rg$state_cluster <- do.call(paste, N_rg[, geo_id_names])
  N_rg <- split(N_rg, N_rg$state_cluster)
  geo_id <- split(geo_id, voter.file$state_cluster)
  r_g_t <- mapply(function(tot_, gid_, g_n_) {
    Nrg_geo_new <- do.call(paste, tot_[, g_n_])
    geo_ <- match(gid_, Nrg_geo_new)
    tot_ <- t(tot_[, grep("^r_", colnames(tot_))]) ## Races in rows
    return(list(
      geo_ = geo_,
      #alpha_ = as.matrix(tab_),
      N_rg_ = tot_
    ))
  }, N_rg, geo_id,
  MoreArgs = list(g_n_ = geo_id_names),
  SIMPLIFY = FALSE
  )
  
  orig_ord <- split(voter.file$rec_id_, voter.file$state_cluster)
  n_groups <- length(orig_ord)
  ## Create name indeces
  name_data <- vector("list", 3)
  names(name_data) <- c("surname", "first", "middle")
  if (ctrl$verbose) {
    message("Pre-processing names...\n")
  }
  for (ntype in c("surname", "first", "middle")) {
    if (ntype %in% name_types) {
      ntab <- switch(ntype,
                     surname = last_c,
                     first = first_c,
                     middle = mid_c)
      kw_names <- toupper(dplyr::pull(ntab, 1))
      proc_names_vf <- .name_preproc(voter.file[[ntype]], c(kw_names))
      u_vf_names <- unique(proc_names_vf)
      kw_in_vf <- kw_names %in% proc_names_vf
      u_kw <- kw_names[kw_in_vf]
      n_u_kw <- length(u_kw)
      reord <- order(match(u_vf_names, u_kw))
      u_vf_names <- u_vf_names[reord]
      w_names <- match(proc_names_vf, u_vf_names)
      w_names <- split(w_names, voter.file$state_cluster)
      pi_ <- as.matrix(ntab[which(kw_in_vf), -1])
      #pi_ <- apply(M_, 2, function(x) x / sum(x, na.rm = TRUE))
      if (impute.missing) {
        pi_miss <- colMeans(pi_, na.rm = TRUE)
      } else {
        pi_miss <- rep(1, n_race)
      }
      pi_[is.na(pi_)] <- 0
      name_data[[ntype]] <- list(
        record_name_id = w_names,
        pi_ = t(pi_),
        pi_miss = pi_miss
      )
    } else {
      name_data[[ntype]] <- list(
        record_name_id = replicate(n_groups, vector("integer"), simplify = FALSE),
        pi_ = matrix(NA, 0, 0),
        pi_miss = array(NA, 5)
      )
    }
  }
  
  ## Build missing distribution
  pi.miss <- do.call(cbind, lapply(name_data, function(x) {
    x$pi_miss
  }))
  
  ## Split inits by cluster
  race.init <- split(race.init, voter.file$state_cluster)
  
  
  ## Name selector
  which.names <- switch(names.to.use,
                        "surname" = 0L,
                        "surname, first" = 1L,
                        "surname, first, middle" = 2L
  )
  
  ## Run Gibbs sampler
  if (ctrl$verbose) {
    message("Sampling races...\n")
    pb <- txtProgressBar(min = 0, max = n_groups, style = 3)
  }
  race_samples <- lapply(seq.int(n_groups),
                         function(cluster) {
                           tmp <- sample_me(
                             name_data[["surname"]]$record_name_id[[cluster]] - 1L,
                             name_data[["first"]]$record_name_id[[cluster]] - 1L,
                             name_data[["middle"]]$record_name_id[[cluster]] - 1L,
                             r_g_t[[cluster]]$geo_ - 1L,
                             r_g_t[[cluster]]$N_rg_,
                             name_data[["surname"]]$pi_,
                             name_data[["first"]]$pi_,
                             name_data[["middle"]]$pi_,
                             pi.miss,
                             which.names,
                             ctrl$iter,
                             ctrl$burnin,
                             race.init[[cluster]] - 1L,
                             0
                           )
                           if (ctrl$verbose) {
                             setTxtProgressBar(pb, cluster)
                           }
                           return(cbind(orig_ord[[cluster]], tmp))
                         }
  )
  if (ctrl$verbose) {
    close(pb)
  }
  if (ctrl$verbose) {
    message("Post-processing results and wrapping up.\n")
  }
  ## Get posterior race probabilities and append to voter.file
  race_samples <- do.call(rbind, race_samples)
  race_samples <- race_samples[order(race_samples[, 1]), -1]
  race_probs <- proportions(race_samples, 1)
  colnames(race_probs) <- paste0("pred.", race.suff)
  voter.file <- cbind(voter.file[, orig.names], race_probs)
  voter.file$state <- orig.state
  attr(voter.file, "RNGseed") <- ctrl$seed
  
  ## Return expanded voter.file with RNG see attribute
  return(voter.file)
}


#' @section .predict_race_embedding:
#' eBISG race prediction function, which uses pre-trained text embeddings
#' (E5-Large) to predict race probabilities for names not found in Census
#' surname lists, rather than falling back to generic population-level priors.
#' @rdname modfuns
#' @keywords internal

predict_race_embedding <- function(
    voter.file,
    names.to.use,
    year = "2020",
    age = FALSE,
    sex = FALSE,
    census.geo = c("tract", "block", "block_group", "county", "place", "zcta"),
    census.key = Sys.getenv("CENSUS_API_KEY"),
    name.dictionaries,
    surname.only = FALSE,
    census.data = NULL,
    retry = 0,
    impute.missing = TRUE,
    skip_bad_geos = FALSE,
    census.surname = FALSE,
    use.counties = FALSE,
    ebisg.model = "intfloat/multilingual-e5-large"
) {

  # Resolve the embedding model configuration
  cfg <- resolve_ebisg_model(ebisg.model)

  # eBISG always provides per-name predictions for unmatched surnames via the
  # MLP, so the caller's impute.missing setting is moot for the surname step.
  # Warn if the user explicitly asked for FALSE so they understand what's
  # happening.
  if (!isTRUE(impute.missing)) {
    warning(
      "impute.missing = FALSE is overridden under model = 'eBISG': ",
      "the embedding model provides per-name predictions for unmatched ",
      "surnames, so no rows are left with NA name probabilities."
    )
  }

  # Check years
  if (!(year %in% c("2000", "2010", "2020"))) {
    stop("Year should be one of 2000, 2010, or 2020 (default).")
  }
  # Define 2020 race marginal
  race.margin <- c(r_whi = 0.5783619, r_bla = 0.1205021, r_his = 0.1872988,
                   r_asi = 0.06106737, r_oth = 0.05276981)

  census.geo <- tolower(census.geo)
  census.geo <- rlang::arg_match(census.geo)

  vars.orig <- names(voter.file)

  # Check names
  if (names.to.use == "surname") {
    message("Proceeding with eBISG last name predictions...")
    if (!("surname" %in% names(voter.file))) {
      stop("Voter data frame needs to have a column named 'surname'.")
    }
  } else if (names.to.use == "surname, first") {
    message("Proceeding with eBISG first and last name predictions...")
    if (!("surname" %in% names(voter.file)) || !("first" %in% names(voter.file))) {
      stop("Voter data frame needs to have a column named 'surname' and a column called 'first'.")
    }
  } else if (names.to.use == "surname, first, middle") {
    message("Proceeding with eBISG first, last, and middle name predictions...")
    if (!("surname" %in% names(voter.file)) || !("first" %in% names(voter.file)) ||
        !("middle" %in% names(voter.file))) {
      stop("Voter data frame needs to have columns named 'surname', 'first', and 'middle'.")
    }
  }

  ## Preliminary data quality checks
  wru_data_preflight()

  path <- ifelse(getOption("wru_data_wd", default = FALSE), getwd(), tempdir())

  first_c <- readRDS(file.path(path, "wru-data-first_c.rds"))
  mid_c <- readRDS(file.path(path, "wru-data-mid_c.rds"))
  if (census.surname) {
    last_c <- readRDS(file.path(path, "wru-data-census_last_c.rds"))
  } else {
    last_c <- readRDS(file.path(path, "wru-data-last_c.rds"))
  }
  if (any(!is.null(name.dictionaries))) {
    if (!is.null(name.dictionaries[["surname"]])) {
      stopifnot(identical(names(name.dictionaries[["surname"]]), names(last_c)))
    }
    if (!is.null(name.dictionaries[["first"]])) {
      stopifnot(identical(names(name.dictionaries[["first"]]), names(first_c)))
    }
    if (!is.null(name.dictionaries[["middle"]])) {
      stopifnot(identical(names(name.dictionaries[["middle"]]), names(mid_c)))
    }
  }

  # Check geographies
  if (surname.only == FALSE) {
    message("Proceeding with Census geographic data at ", census.geo, " level...")

    if (is.null(census.data)) {
      census.key <- validate_key(census.key)
      message("Downloading Census geographic data using provided API key...")
    } else {
      if (!("state" %in% names(voter.file))) {
        stop("voter.file object needs to have a column named state.")
      }
      census_data_preflight(census.data, census.geo, year)
      if (sum(toupper(unique(as.character(voter.file$state))) %in%
              toupper(names(census.data)) == FALSE) > 0) {
        message("census.data object does not include all states in voter.file object.")
        census.key <- validate_key(census.key)
        message("Downloading Census geographic data for states not included in census.data object...")
      } else {
        message("Using Census geographic data from provided census.data object...")
      }
    }

    geo_id_names <- determine_geo_id_names(census.geo)

    if (!all(geo_id_names %in% names(voter.file))) {
      stop(message(
        "To use ", census.geo, " as census.geo, voter.file needs to include the following column(s): ",
        paste(geo_id_names, collapse = ", ")
      ))
    }

    voter.file <- census_helper_new(
      key = census.key,
      voter.file = voter.file,
      states = "all",
      geo = census.geo,
      age = age,
      sex = sex,
      year = year,
      census.data = census.data,
      retry = retry,
      use.counties = use.counties,
      skip_bad_geos = skip_bad_geos
    )
  }

  eth <- c("whi", "bla", "his", "asi", "oth")

  ## Merge in Pr(Name | Race) using standard merge_names
  voter.file <- merge_names(
    voter.file = voter.file,
    namesToUse = names.to.use,
    census.surname = census.surname,
    table.surnames = name.dictionaries[["surname"]],
    table.first = name.dictionaries[["first"]],
    table.middle = name.dictionaries[["middle"]],
    clean.names = TRUE,
    impute.missing = TRUE,
    model = "BISG"
  )

  ## Load the surname dictionary to identify which names were actually matched
  if (census.surname) {
    last_dict <- readRDS(file.path(path, "wru-data-census_last_c.rds"))
  } else {
    last_dict <- readRDS(file.path(path, "wru-data-last_c.rds"))
  }
  known_surnames <- toupper(last_dict[[1]])  # first column is the name

  ## Initialized lazily on first use; both the surname and first-name
  ## blocks below check is.null(model_dir) to avoid double-downloading.
  model_dir <- NULL

  ## Identify unmatched surnames via merge_names' final scratchpad. The
  ## scratchpad is only used for the matched/unmatched decision; the text
  ## handed to the embedding model must come from the original surname
  ## column, since merge_names overwrites lastname.match during cleaning
  ## (a non-hyphenated unmatched name ends up as the literal string "NA").
  cleaned_surnames  <- toupper(voter.file$lastname.match)
  original_surnames <- toupper(as.character(voter.file$surname))
  unmatched_mask    <- !(cleaned_surnames %in% known_surnames)
  embed_mask        <- unmatched_mask &
                       !is.na(original_surnames) &
                       original_surnames != ""
  n_unmatched <- sum(embed_mask)

  if (n_unmatched > 0) {
    ## Check Python dependencies and download model weights
    ensure_ebisg_python()
    model_dir <- ebisg_data_preflight(cfg)

    message(
      "eBISG: ", n_unmatched, " of ", nrow(voter.file),
      " records have surnames not in Census list. ",
      "Using ", cfg$transformer, " embeddings for these names."
    )

    ## Embed the original surname text (not the cascade scratchpad) so each
    ## row gets a name-specific prediction.
    unmatched_surnames <- unique(original_surnames[embed_mask])

    message("Embedding ", length(unmatched_surnames), " unique unmatched surnames...")
    embeddings <- ebisg_embed_names(unmatched_surnames, transformer = cfg$transformer)
    mlp_path   <- if (is.null(cfg$tag)) cfg$surname_mlp else file.path(model_dir, cfg$surname_mlp)
    probs_6    <- ebisg_predict_mlp(embeddings, mlp_path)
    probs_5    <- map_6class_to_5class(probs_6)

    sn_lookup <- as.data.frame(probs_5)
    rownames(sn_lookup) <- unmatched_surnames

    ## Override imputed values for embeddable unmatched rows. Rows excluded
    ## from embed_mask (NA / empty original surname) keep merge_names'
    ## column-mean fallback rather than getting NA written over them.
    for (k in seq_along(eth)) {
      col_name <- paste0("c_", eth[k], "_last")
      voter.file[[col_name]][embed_mask] <-
        sn_lookup[original_surnames[embed_mask], paste0("c_", eth[k])]
    }
  } else {
    message("eBISG: All surnames matched Census list.")
  }

  ## Also handle unmatched first names with embedding if requested. Same
  ## logic as the surname block: the cleaning scratchpad decides which
  ## rows are unmatched, but the text we embed comes from the original
  ## first-name column.
  if (grepl("first", names.to.use) && !is.null(cfg$firstname_mlp)) {
    first_dict <- readRDS(file.path(path, "wru-data-first_c.rds"))
    known_firstnames    <- toupper(first_dict[[1]])
    cleaned_firstnames  <- toupper(voter.file$firstname.match)
    original_firstnames <- toupper(as.character(voter.file$first))
    unmatched_first_mask <- !(cleaned_firstnames %in% known_firstnames)
    embed_first_mask     <- unmatched_first_mask &
                            !is.na(original_firstnames) &
                            original_firstnames != ""
    n_unmatched_first <- sum(embed_first_mask)

    if (n_unmatched_first > 0) {
      if (is.null(model_dir)) {
        ensure_ebisg_python()
        model_dir <- ebisg_data_preflight(cfg)
      }

      message(
        "eBISG: ", n_unmatched_first,
        " records have first names not in dictionary. ",
        "Using ", cfg$transformer, " embeddings."
      )

      unmatched_firstnames <- unique(original_firstnames[embed_first_mask])

      embeddings_fn <- ebisg_embed_names(unmatched_firstnames, transformer = cfg$transformer)
      fn_mlp_path   <- if (is.null(cfg$tag)) cfg$firstname_mlp else file.path(model_dir, cfg$firstname_mlp)
      probs_6_fn    <- ebisg_predict_mlp(embeddings_fn, fn_mlp_path)
      probs_5_fn    <- map_6class_to_5class(probs_6_fn)

      fn_lookup <- as.data.frame(probs_5_fn)
      rownames(fn_lookup) <- unmatched_firstnames

      for (k in seq_along(eth)) {
        col_name <- paste0("c_", eth[k], "_first")
        voter.file[[col_name]][embed_first_mask] <-
          fn_lookup[original_firstnames[embed_first_mask], paste0("c_", eth[k])]
      }
    }
  }

  ## Compute predictions (same Bayesian combination as predict_race_new)
  if (surname.only == TRUE) {
    # Pr(Race | Surname)
    preds <- voter.file[, grep("_last$", names(voter.file))] *
      matrix(race.margin, nrow = nrow(voter.file), ncol = length(race.margin), byrow = TRUE)
  } else {
    # Pr(Race | Surname, Geolocation)
    preds <- voter.file[, grep("_last$", names(voter.file))] *
      voter.file[, grep("^r_", names(voter.file))]
    if (grepl("first", names.to.use)) {
      preds <- preds * voter.file[, grep("_first$", names(voter.file))]
    }
    if (grepl("middle", names.to.use)) {
      preds <- preds * voter.file[, grep("_middle$", names(voter.file))]
    }
  }

  ## Normalize
  preds <- preds / rowSums(preds)
  colnames(preds) <- paste("pred", eth, sep = ".")

  return(data.frame(cbind(voter.file[c(vars.orig)], preds)))
}

# Note: must provide a valid U.S. Census API key for test cases that use U.S. Census statistics
# > usethis::edit_r_profile
# Sys.setenv("CENSUS_API_KEY" = "yourkey")
# For testing package coverage use: Sys.setenv("NOT_CRAN" = "TRUE")
options("piggyback.verbose" = FALSE)
options("wru_data_wd" = TRUE)

skip_if_not(nzchar(Sys.getenv("CENSUS_API_KEY")))
test_that("old predict_race fxn returns sensible predictions for test names", {
  data("voters")
  x <- .predict_race_old(
    voter.file = voters,
    census.surname = TRUE,
    surname.only = TRUE,
    surname.year = 2010,
    census.geo = "tract",
    census.key = Sys.getenv("CENSUS_API_KEY"),
    age = FALSE,
    sex = FALSE,
    year = "2010",
    retry = 3,
    impute.missing = TRUE,
    use.counties = FALSE
  )
  expect_equal(x[x$surname == "Lopez", "pred.whi"], 0.0486000, tolerance = .000001)
  expect_equal(x[x$surname == "Khanna", "pred.whi"], 0.0676000, tolerance = .000001)
  expect_equal(x[x$surname == "Lopez", "pred.bla"], 0.00570000, tolerance = .000001)
  expect_equal(x[x$surname == "Khanna", "pred.bla"], 0.00430000, tolerance = .000001)
  expect_equal(x[x$surname == "Lopez", "pred.his"], 0.92920000, tolerance = .000001) #assumed to be high Hispanic score
  expect_equal(x[x$surname == "Zhou", "pred.asi"], 0.98200000, tolerance = .000001) #assumed to be high Asian score
  
})

#' Preflight census data
#' 
#' @inheritParams predict_race
#' @keywords internal

census_data_preflight <- function(census.data, census.geo, year) {
  vars_ <- unlist(census_geo_api_names(year = year))
  legacy_vars <- unlist(census_geo_api_names_legacy(year = year))
  
  test <- lapply(census.data, function(x) {
    nms_to_test <- names(x[[census.geo]])
    all(vars_ %in% nms_to_test) || all(legacy_vars %in% nms_to_test)
  }) 
  missings <- names(test)[!unlist(test)]
  
  if(any(!unlist(test))) {
    stop(
      paste0(
        "Missing ", 
        paste0(vars_, collapse = ", "), 
        " from census.data object. Please update your census.data by",
        " running `get_census_data` again."
      )
    )
  }
}

#' Variable vector into chunks.
#'
#' \code{vec_to_chunk} takes a list of variables and collects them into 50-variable chunks.
#' 
#' This function takes a list of variable names and collects them into chunks with no more than 
#' 50 variables each. This helps to get around requests with more than 50 variables,because the 
#' API only allows queries of 50 variables at a time. 
#' The user should not need to call this function directly.
#'
#' @param x Character vector of variable names.
#' @return Object of class \code{list}.
#'
#' @examples
#' \dontrun{
#' vec_to_chunk(x = c(paste("P012F0", seq(10:49), sep = ""),
#'               paste("P012I0", seq(10, 49), sep = "")))
#' }
#'
#' @references
#' Based on code authored by Nicholas Nagle, which is available 
#' \href{https://rstudio-pubs-static.s3.amazonaws.com/19337_2e7f827190514c569ea136db788ce850.html}{here}.
#'
#' @keywords internal

vec_to_chunk <- function(x){
  s <- seq_along(x)
  x1 <- split(x, ceiling(s/50))
  return(x1)
}

#' Multilevel Census data download function.
#'
#' \code{get_census_data} returns county-, tract-, and block-level Census data 
#' for specified state(s). Using this function to download Census data in advance 
#' can save considerable time when running \code{predict_race} and \code{census_helper}.
#' 
#' @param key A character string containing a valid Census API key, 
#'   which can be requested from the
#'   [U.S. Census API key signup page](https://api.census.gov/data/key_signup.html).
#'   
#'   By default, attempts to find a census key stored in an
#'   [environment variable][Sys.getenv] named `CENSUS_API_KEY`.
#'   
#' @param states which states to extract Census data for, e.g., \code{c("NJ", "NY")}.
#' @param age A \code{TRUE}/\code{FALSE} object indicating whether to condition on 
#'  age or not. If \code{FALSE} (default), function will return Pr(Geolocation | Race).
#'  If \code{TRUE}, function will return Pr(Geolocation, Age | Race). 
#'  If \code{\var{sex}} is also \code{TRUE}, function will return Pr(Geolocation, Age, Sex | Race).
#' @param sex A \code{TRUE}/\code{FALSE} object indicating whether to condition on 
#'  sex or not. If \code{FALSE} (default), function will return Pr(Geolocation | Race). 
#'  If \code{TRUE}, function will return Pr(Geolocation, Sex | Race). 
#'  If \code{\var{age}} is also \code{TRUE}, function will return Pr(Geolocation, Age, Sex | Race).
#' @param year A character object specifying the year of U.S. Census data to be downloaded.
#'  Use \code{"2010"}, or \code{"2020"}. Default is \code{"2020"}.
#'  Warning: 2020 U.S. Census data is downloaded only when \code{\var{age}} and 
#'  \code{\var{sex}} are both \code{FALSE}.
#' @param census.geo An optional character vector specifying what level of 
#' geography to use to merge in U.S. Census 2010 geographic data. Currently
#' \code{"county"}, \code{"tract"}, \code{"block"}, and \code{"place"} are supported.
#' @param retry The number of retries at the census website if network interruption occurs.
#' @param county.list A named list of character vectors of counties present in your \var{voter.file}, per state. 
#' @return Output will be an object of class \code{list} indexed by state. 
#' Output will contain a subset of the following elements: 
#' \code{state}, \code{age}, \code{sex}, 
#' \code{county}, \code{tract}, \code{block_group}, \code{block}, and \code{place}.
#' 
#' @export
#'
#' @examples
#' \dontrun{get_census_data(states = c("NJ", "NY"), age = TRUE, sex = FALSE)}
#' \dontrun{get_census_data(states = "MN", age = FALSE, sex = FALSE, year = "2020")}
get_census_data <- function(
    key = Sys.getenv("CENSUS_API_KEY"),
    states,
    age = FALSE,
    sex = FALSE,
    year = "2020",
    census.geo = c("tract", "block", "block_group", "county", "place", "zcta"),
    retry = 3,
    county.list = NULL
) {
  key <- validate_key(key)
  
  census.geo <- tolower(census.geo)
  census.geo <- rlang::arg_match(census.geo)
  
  states <- toupper(states)
  
  message("Collecting ", year, " Census data...")
  
  CensusObj <- NULL
  for (s in states) {
    CensusObj[[s]] <- list(state = s, age = age, sex = sex, year = year)
    if (census.geo == "place") {
      place <- census_geo_api(key, s, geo = "place", age, sex, year, retry)
      CensusObj[[s]]$place <- place
    }
    if (census.geo == "block") {
      block <- census_geo_api(key, s, geo = "block", age, sex, year, retry, counties = county.list[[s]])
      CensusObj[[s]]$block <- block
    }
    
    if (census.geo == "block_group") {
      block_group <- census_geo_api(key, s, geo = "block_group", age, sex, year, retry, counties = county.list[[s]])
      CensusObj[[s]]$block_group <- block_group
    }
    
    if ((census.geo == "block") || (census.geo == "tract") || (census.geo == "block_group")) {
      tract <- census_geo_api(key, s, geo = "tract", age, sex, year, retry, counties = county.list[[s]])
      CensusObj[[s]]$tract <- tract
    }

    if ((census.geo == "block") || (census.geo == "tract") || (census.geo == "county") || (census.geo == "block_group")) {
      county <- census_geo_api(key, s, geo = "county", age, sex, year, retry)
      CensusObj[[s]]$county <- county
    }
    
    if (census.geo == "zcta") {
      if (!is.null(county.list)) {
        cli::cli_abort(c(
          "The {.arg county.list} argument must be set to {.code NULL}
          when {.arg census_geo} is {.val zcta},
          because the Census Bureau does release data that divides ZCTAs by county."
        ))
      }
      
      CensusObj[[s]]$zcta <- census_geo_api(key, s, geo = "zcta", age, sex, year, retry)
    }
  }
  return(CensusObj)
}

if(Sys.getenv("CENSUS_API_KEY") != "") {
  test_that("Census block download works", {
    # takes a long time!
    future::plan(future::multisession)
    r <- suppressMessages(get_census_data(
      key = NULL, 
      state = c("DC"), 
      census.geo = "block" 
    ))
    
    expect_named(r$DC, c("state", "age", "sex", "year", "block", "tract", "county"))
    expect_true(all(r$DC$block$state == "DC"))
  })
  
  test_that("Census block_group download works", {
    future::plan(future::multisession)
    r <- suppressMessages(get_census_data(
      key = NULL,
      state = "RI",
      census.geo = "block_group"
    ))
    expect_named(r$RI, c("state", "age", "sex", "year", "block_group", "tract", "county"))
    expect_true(all(r$RI$place$state == "RI"))
  })
  
  test_that("Census tract download works", {
    future::plan(future::multisession)
    r <- suppressMessages(get_census_data(
      key = NULL,
      state = c("NY"),
      census.geo = "tract",
      county.list =  list(NY = "061")
    ))
    expect_named(r$NY, c("state", "age", "sex", "year", "tract", "county"))
    expect_true(all(r$NY$tract$state == "NY"))
    expect_true(all(r$NY$county$state == "NY"))
  })
  
  test_that("Census county download works", {
    future::plan(future::multisession)
    r <- suppressMessages(get_census_data(
      key = NULL, 
      state = "NJ",
      census.geo = "county",
      county.list = list(NJ = "021")
    ))
    expect_named(r$NJ, c("state", "age", "sex", "year", "county"))
    expect_true(all(r$NJ$county$state == "NJ"))
  })
  
  test_that("Census place download works", {
    future::plan(future::multisession)
    r <- suppressMessages(get_census_data(
      key = NULL,
      state = "RI",
      census.geo = "place"
    ))
    expect_named(r$RI, c("state", "age", "sex", "year", "place"))
    expect_true(all(r$RI$place$state == "RI"))
  })
  
  test_that("Census ZCTA download works", {
    r <- suppressMessages(get_census_data(
      key = NULL,
      state = "DC",
      census.geo = "zcta"
    ))
    expect_named(r$DC, c("state", "age", "sex", "year", "zcta"))
    expect_true(all(r$DC$zcta$state == "DC"))
    
    r <- suppressMessages(get_census_data(
      key = NULL,
      state = "DC",
      census.geo = "zcta",
      year = "2010"
    ))
    expect_named(r$DC, c("state", "age", "sex", "year", "zcta"))
    expect_true(all(r$DC$zcta$state == "DC"))
  })
}

if(Sys.getenv("CENSUS_API_KEY") != "") {
  test_that('Roll-ups of proportions between levels for race behave as expected', {
    skip_on_cran()
    future::plan(future::multisession)
    r <- suppressMessages(get_census_data(
      key = NULL,
      state = c("DE"),
      census.geo = "block",
    ))
    
    r_tract_from_block <- r$DE$block |>
      dplyr::group_by(tract) |>
      dplyr::summarize(
        sum_whi = sum(r_whi),
        sum_bla = sum(r_bla),
        sum_his = sum(r_his),
        sum_asi = sum(r_asi),
        sum_oth = sum(r_oth)
      )
    
    r_tract_level <- r$DE$tract |>
      dplyr::select(tract, r_whi:r_oth) |>
      dplyr::group_by(tract) |>
      dplyr::summarize(
        sum_whi = sum(r_whi),
        sum_bla = sum(r_bla),
        sum_his = sum(r_his),
        sum_asi = sum(r_asi),
        sum_oth = sum(r_oth)
      )
    
    r_county_from_tract <- r$DE$tract |>
      dplyr::select(county, r_whi:r_oth) |>
      dplyr::group_by(county) |>
      dplyr::summarize(
        sum_whi = sum(r_whi),
        sum_bla = sum(r_bla),
        sum_his = sum(r_his),
        sum_asi = sum(r_asi),
        sum_oth = sum(r_oth)
      )
    
    r_county_level <- r$DE$county |>
      dplyr::select(county, r_whi:r_oth) |>
      dplyr::group_by(county) |>
      dplyr::summarize(
        sum_whi = sum(r_whi),
        sum_bla = sum(r_bla),
        sum_his = sum(r_his),
        sum_asi = sum(r_asi),
        sum_oth = sum(r_oth)
      )
    
    expect_true(
      all.equal(
        colSums(r$DE$block[, c('r_whi', 'r_bla', 'r_his', 'r_asi', 'r_oth')]),
        c('r_whi' = 1, 'r_bla' = 1, 'r_his' = 1, 'r_asi' = 1, 'r_oth' = 1)
      )
    )
    expect_equal(r_county_from_tract, r_county_level, tolerance = 1e-7)
    expect_equal(r_tract_level, r_tract_from_block, tolerance = 1e-7)
  })
  
  test_that('Roll-ups of population sums between levels for race behave as expected', {
    future::plan(future::multisession)
    r <- suppressMessages(get_census_data(
      key = NULL,
      state = c("WY"),
      census.geo = "block",
    ))
    
    r_sum_from_block <- r$WY$block |>
      dplyr::select(-state, -county, -tract, -block, -dplyr::starts_with("r_")) |>
      colSums()
    
    r_sum_from_tract <- r$WY$tract |>
      dplyr::select(-state, -county, -tract, -dplyr::starts_with("r_")) |>
      colSums()
    
    r_sum_from_county <- r$WY$county  |>
      dplyr::select(-state, -county, -dplyr::starts_with("r_")) |>
      colSums()
    
    expect_equal(r_sum_from_block, r_sum_from_tract)
    expect_equal(r_sum_from_block, r_sum_from_county)
    
    r_zcta_level <- suppressMessages(get_census_data(
      key = NULL,
      state = "WY",
      census.geo = "zcta"
    ))
    
    r_sum_from_zcta <- r_zcta_level$WY$zcta |>
      dplyr::select(-state, -zcta, -dplyr::starts_with("r_")) |>
      colSums()
    
    expect_equal(r_sum_from_block, r_sum_from_zcta, tolerance = 0.001)
  })
  
  test_that('Roll-ups for ZIP and county race sum to same for state', {
    r_zcta_level <- suppressMessages(get_census_data(
      key = NULL,
      state = "RI",
      census.geo = "zcta"
    ))
    
    r_county_level <- suppressMessages(get_census_data(
      key = NULL,
      state = "RI",
      census.geo = "county"
    ))
    
    expect_equal(
      r_zcta_level$RI$zcta |>
        dplyr::select(-state, -zcta, -dplyr::starts_with("r_")) |>
        colSums(),
      r_county_level$RI$county |>
        dplyr::select(-state, -county, -dplyr::starts_with("r_")) |>
        colSums(),
      tolerance = 0.001
    )
  })
  
  test_that('Roll-ups of population sums between levels for race/sex/age behave as expected', {
    skip_on_cran()
    future::plan(future::multisession)
    r <- suppressMessages(get_census_data(
      key = NULL,
      state = c("AK"),
      census.geo = "block",
      sex = TRUE,
      age = TRUE
    ))
    
    r_sum_from_block <- r$AK$block |>
      dplyr::select(dplyr::starts_with("r_")) |>
      colSums()
    
    r_sum_from_tract <- r$AK$tract |>
      dplyr::select(dplyr::starts_with("r_")) |>
      colSums()
    
    r_sum_from_county <- r$AK$county |>
      dplyr::select(dplyr::starts_with("r_")) |>
      colSums()
    
    expect_equal(r_sum_from_block, r_sum_from_tract)
    expect_equal(r_sum_from_block, r_sum_from_county)
    
    r_zcta_level <- suppressMessages(get_census_data(
      key = NULL,
      state = "AK",
      census.geo = "zcta",
      sex = TRUE,
      age = TRUE
    ))
    
    r_sum_from_zcta <- r_zcta_level$AK$zcta |>
      dplyr::select(dplyr::starts_with("r_")) |>
      colSums()
    
    expect_equal(r_sum_from_block, r_sum_from_zcta, tolerance = 0.001)
  })
}

#' Census Data download function.
#'
#' \code{census_geo_api} retrieves U.S. Census geographic data for a given state.
#'
#' This function allows users to download U.S. Census geographic data (2010 or 2020), 
#' at either the county, tract, block, or place level, for a particular state. 
#'
#' @inheritParams get_census_data
#' @param state A required character object specifying which state to extract Census data for, 
#'  e.g., \code{"NJ"}.
#' @param geo A character object specifying what aggregation level to use. 
#'  Use `"block"`, `"block_group"`, `"county"`, `"place"`, `"tract"`, or `"zcta"`. 
#'  Default is \code{"tract"}. Warning: extracting block-level data takes very long.
#' @param age A \code{TRUE}/\code{FALSE} object indicating whether to condition on 
#'  age or not. If \code{FALSE} (default), function will return Pr(Geolocation | Race).
#'  If \code{TRUE}, function will return Pr(Geolocation, Age | Race). 
#'  If \code{\var{sex}} is also \code{TRUE}, function will return Pr(Geolocation, Age, Sex | Race).
#' @param sex A \code{TRUE}/\code{FALSE} object indicating whether to condition on 
#'  sex or not. If \code{FALSE} (default), function will return Pr(Geolocation | Race). 
#'  If \code{TRUE}, function will return Pr(Geolocation, Sex | Race). 
#'  If \code{\var{age}} is also \code{TRUE}, function will return Pr(Geolocation, Age, Sex | Race).
#' @param year A character object specifying the year of U.S. Census data to be downloaded.
#'  Use \code{"2010"}, or \code{"2020"}. Default is \code{"2020"}.
#'  Warning: 2020 U.S. Census data is downloaded only when \code{\var{age}} and 
#'  \code{\var{sex}} are both \code{FALSE}.
#' @param retry The number of retries at the census website if network interruption occurs.
#' @param save_temp File indicating where to save the temporary outputs. 
#'  Defaults to NULL. If specified, the function will look for an .RData file
#'  with the same format as the expected output. 
#' @param counties A vector of counties contained in your data. If \code{NULL}, all counties are pulled.
#' Useful for smaller predictions where only a few counties are considered. Must be zero padded.
#' @return Output will be an object of class \code{list}, indexed by state names. It will 
#'  consist of the original user-input data with additional columns of Census geographic data.
#'
#' @examples
#' \dontshow{data(voters)}
#' \dontrun{census_geo_api(states = c("NJ", "DE"), geo = "block")}
#' \dontrun{census_geo_api(states = "FL", geo = "tract", age = TRUE, sex = TRUE)}
#' \dontrun{census_geo_api(states = "MA", geo = "place", age = FALSE, sex = FALSE,
#'  year = "2020")}
#'
#' @references
#' Relies on `get_census_api()`, `get_census_api_2()`, and `vec_to_chunk()` functions authored by Nicholas Nagle, 
#' available [here](https://rstudio-pubs-static.s3.amazonaws.com/19337_2e7f827190514c569ea136db788ce850.html).
#' 
#' @importFrom furrr future_map_dfr
#' @importFrom purrr map_dfr
#' @keywords internal

census_geo_api <- function(
    key = Sys.getenv("CENSUS_API_KEY"),
    state,
    geo = c("tract", "block", "block_group", "county", "place", "zcta"),
    age = FALSE,
    sex = FALSE,
    year = c("2020", "2010"),
    retry = 3,
    save_temp = NULL,
    counties = NULL
) {
  key <- validate_key(key)
  
  geo <- tolower(geo)
  geo <- rlang::arg_match(geo)
  
  year <- as.character(year)
  year <- rlang::arg_match(year)
  
  census <- NULL
  state <- as_state_abbreviation(state)
  
  df.out <- NULL
  state.fips <- as_fips_code(state)
  
  vars <- census_geo_api_names(year = year, age = age, sex = sex)
  census_data_url <- census_geo_api_url(year = year)
  
  if (geo == "place") {
    region <- paste("for=place:*&in=state:", state.fips, sep = "")
    census <- get_census_api(census_data_url, key = key, var.names = unlist(vars), region = region, retry)
  }
  
  if (geo == "county") {
    if (is.null(counties)) {
      region <- paste("for=county:*&in=state:", state.fips, sep = "")
    } else {
      counties_paste <- paste0(counties, collapse = ",")
      region <- paste("for=county:",counties_paste,"&in=state:", state.fips, sep = "")
    }

    census <- get_census_api(census_data_url, key = key, var.names = unlist(vars), region = region, retry)
  }
  
  if (geo == "tract") {
    if (is.null(counties)) {
      region_county <- paste("for=county:*&in=state:", state.fips, sep = "")
    } else {
      counties_paste <- paste0(counties, collapse = ",")
      region_county <- paste("for=county:",counties_paste,"&in=state:", state.fips, sep = "")
    }
    
    county_df <- get_census_api(census_data_url, key = key, var.names = unlist(vars), region = region_county, retry)
    
    if(is.null(counties)) {
      county_list <- county_df$county
    } else {
      county_list <- intersect(counties, county_df$county)
    }
    
    if(length(county_list) > 0) {
      census <- furrr::future_map_dfr(seq_along(county_list), function(county) {
        message(paste("County ", county, " of ", length(county_list), ": ", county_list[county], sep = ""))
        region_county <- paste("for=tract:*&in=state:", state.fips, "+county:", county_list[county], sep = "")
        get_census_api(data_url = census_data_url, key = key, var.names = unlist(vars), region = region_county, retry)
      }, .progress = TRUE)
    } else {
      message('There were no intersecting counties in your voter.file data (tract)')
    } 
  }
  
  if (geo == "block_group") {
    if (is.null(counties)) {
      region_county <- paste("for=county:*&in=state:", state.fips, sep = "")
    } else {
      counties_paste <- paste0(counties, collapse = ",")
      region_county <- paste("for=county:",counties_paste,"&in=state:", state.fips, sep = "")
    }
    
    county_df <- get_census_api(census_data_url, key = key, var.names = unlist(vars), region = region_county, retry)
    
    if(is.null(counties)) {
      county_list <- county_df$county
    } else {
      county_list <- intersect(counties, county_df$county)
    }
    
    if(length(county_list) > 0) {
      message('Running block_group by county...')
      
      census <- purrr::map_dfr(
        1:length(county_list), 
        function(county) {
          # too verbose, commenting out
          message(paste("County ", county, " of ", length(county_list), ": ", county_list[county], sep = ""))
          
          blockgroup <- paste("for=block+group:*&in=state:", state.fips, "+county:", county_list[county], sep = "")
  
          # message(region_tract)
          blockgroup_df <- get_census_api(census_data_url, key = key, var.names = unlist(vars), region = blockgroup, retry)
          names(blockgroup_df)[4] <- "block_group" # Fix name, it comes in with a space from api. 
          blockgroup_df
        }
      )
      message("\n") # new line for progress bar
    } else {
      message('There were no intersecting counties in your voter.file data (block)')
    }
  }
  
  if (geo == "block") {
    if (is.null(counties)) {
      region_county <- paste("for=county:*&in=state:", state.fips, sep = "")
    } else {
      counties_paste <- paste0(counties, collapse = ",")
      region_county <- paste("for=county:",counties_paste,"&in=state:", state.fips, sep = "")
    }
    
    county_df <- get_census_api(census_data_url, key = key, var.names = unlist(vars), region = region_county, retry)
    
    if(is.null(counties)) {
      county_list <- county_df$county
    } else {
      county_list <- intersect(counties, county_df$county)
    }
    
    if(length(county_list) > 0) {
      message('Running block by county...')
      
      census <- purrr::map_dfr(
        1:length(county_list), 
        function(county) {
          # too verbose, commenting out
          message(paste("County ", county, " of ", length(county_list), ": ", county_list[county], sep = ""))
          
          region_tract <- paste("for=tract:*&in=state:", state.fips, "+county:", county_list[county], sep = "")
          # message(region_tract)
          tract_df <- get_census_api(census_data_url, key = key, var.names = unlist(vars), region = region_tract, retry)
          tract_list <- tract_df$tract
          
          furrr::future_map_dfr(1:length(tract_list), function(tract) {
            message(paste("Tract ", tract, " of ", length(tract_list), ": ", tract_list[tract], sep = ""))
            
            region_block <- paste("for=block:*&in=state:", state.fips, "+county:", county_list[county], "+tract:", tract_list[tract], sep = "")
            get_census_api(census_data_url, key = key, var.names = unlist(vars), region = region_block, retry)
          }, .progress = TRUE)
        }
      )
      message("\n") # new line for progress bar
    } else {
      message('There were no intersecting counties in your voter.file data (block)')
    } 
  }
  
  if (geo == "zcta") {
    census <- census_geo_api_zcta(
      census_data_url = census_data_url,
      key = key,
      vars = vars,
      state = state,
      counties = counties,
      retry = retry
    )
  }
  
  census <- dplyr::mutate(census, state = as_state_abbreviation(state))
  
  r_columns <- purrr::map(vars, function(vars) rowSums(census[vars]))
  
  census <- dplyr::bind_cols(census, r_columns)
  census <- dplyr::group_by(census, dplyr::across(dplyr::any_of("state")))
  census <- dplyr::mutate(
    census,
    dplyr::across(
      # Divide all r_columns by the total population of the corresponding race
      dplyr::all_of(names(r_columns)),
      function(x) {
        x / sum(
          dplyr::pick(
            sub("^.+_(.{3})$", "r_\\1", dplyr::cur_column(), perl = TRUE)
          )
        )
      }
    )
  )
  census <- dplyr::ungroup(census)
  
  census
}

skip_if_not(nzchar(Sys.getenv("CENSUS_API_KEY")))

test_that("snapshot", {
  # TODO: Test that sub-geographies sum to match pooled geographies (e.g. blocks sum to block groups, sum to tracts, sum to counties)
  
  # These snapshots were generated using the calculations in v2.0.0
  # and verified that the calculations resulted in the same numbers for PR #120.
  expect_snapshot_value(
    census_geo_api(state = "DE", geo = "county", year = "2020"),
    style = "deparse"
  )
  expect_snapshot_value(
    census_geo_api(state = "DE", geo = "county", year = "2010"),
    style = "deparse"
  )
  expect_snapshot_value(
    census_geo_api(state = "DE", geo = "county", year = "2010", sex = TRUE),
    style = "deparse"
  )
  expect_snapshot_value(
    census_geo_api(state = "DE", geo = "county", year = "2010", age = TRUE),
    style = "deparse"
  )
})

expect_subset_sums_equal_overall_total <- function(data) {
  `%>%` <- dplyr::`%>%`
  
  sums <- data %>%
    dplyr::select(-dplyr::starts_with("r_")) %>%
    tidyr::pivot_longer(dplyr::starts_with("P")) %>%
    dplyr::mutate(
      subset = dplyr::case_when(
        grepl("001", name) ~ "overall",
        .default = "subset"
      ),
      name = sub("_.+", "", name)
    ) %>%
    dplyr::summarise(
      value = sum(value),
      .by = -value
    ) %>%
    dplyr::summarize(
      are_equal = length(unique(value)) <= 1,
      .by = c(-subset, -value)
    )
  
  expect_true(all(sums$are_equal))
}

test_that("sums", {
  expect_subset_sums_equal_overall_total(
    census_geo_api(state = "DE", geo = "county", year = "2020", sex = TRUE)
  )
  expect_subset_sums_equal_overall_total(
    census_geo_api(state = "DE", geo = "county", year = "2020", age = TRUE)
  )
  expect_subset_sums_equal_overall_total(
    census_geo_api(state = "DE", geo = "county", year = "2020", age = TRUE, sex = TRUE)
  )
  expect_subset_sums_equal_overall_total(
    census_geo_api(state = "DE", geo = "county", year = "2010", sex = TRUE)
  )
  expect_subset_sums_equal_overall_total(
    census_geo_api(state = "DE", geo = "county", year = "2010", age = TRUE)
  )
  expect_subset_sums_equal_overall_total(
    census_geo_api(state = "DE", geo = "county", year = "2010", age = TRUE, sex = TRUE)
  )
})

# TODO: Test that all variables sum to total population of geography

#' Legacy data formatting function.
#'
#' \code{format_legacy_data} formats legacy data from the U.S. census to allow
#' for Bayesian name geocoding. 
#'
#' This function allows users to construct datasets for analysis using the census legacy data format. 
#' These data are available for the 2020 census at 
#' https://www2.census.gov/programs-surveys/decennial/2020/data/01-Redistricting_File--PL_94-171/. 
#' This function returns data structured analogously to data from the Census API, which is not yet 
#' available for the 2020 Census as of September 2021. 
#'
#' @param legacyFilePath A character vector giving the location of a legacy census data folder,
#' sourced from https://www2.census.gov/programs-surveys/decennial/2020/data/01-Redistricting_File--PL_94-171/. 
#' These file names should end in ".pl".
#' @param state The two letter state postal code.
#' @param outFile Optional character vector determining whether the formatted RData object should be saved. The
#' filepath should end in ".RData". 
#'
#' @import PL94171
#' 
#' @examples
#' \dontrun{
#' gaCensusData <- format_legacy_data(PL94171::pl_url('GA', 2020))
#' predict_race_new(ga.voter.file, namesToUse = 'last, first, mid', census.geo = 'block',
#'      census.data = gaCensusData)
#'}

#'
#' @export
format_legacy_data <- function(legacyFilePath, state, outFile = NULL) {
  
  # aggregation levels to convert (county, tract, block group, and block)
  summaryLevels <- c('050', '140', '150', '750')
  
  # read in the data
  pl <- pl_read(legacyFilePath)
  pl <- pl_select_standard(pl) 
  
  # iterate through the levels 
  censusData.2020 <- lapply(summaryLevels, FUN = function(level) {
    levelData <- pl[pl$summary_level == level,]
    
    # construct the base data frame
    df <- data.frame(state = toupper(state), 
                     county = levelData$county,
                     P005003 = levelData$pop_white, 
                     P005004 = levelData$pop_black, 
                     P005010 = levelData$pop_hisp,
                     P005006 = levelData$pop_asian,
                     P005007 = levelData$pop_nhpi,
                     P005005 = levelData$pop_aian, 
                     P005008 = levelData$pop_other, 
                     P005009 = levelData$pop_two)
    
    # add geographic levels
    if(level != '050') {
      df$tract <- substr(levelData$GEOID, nchar(levelData$GEOID) - 5, nchar(levelData$GEOID))
      if(level != '140') {
        df$blockGroup <- substr(levelData$GEOID, nchar(levelData$GEOID), nchar(levelData$GEOID))
        if(level != '150') 
          df$block <- substr(levelData$GEOID, nchar(levelData$GEOID)-2, nchar(levelData$GEOID))
      }
    }
    
    df
  })
  
  # format and optionally save the file 
  names(censusData.2020) <- c('county', 'tract', 'blockGroup', 'block')
  if(!is.null(outFile))
    save(censusData.2020, file = outFile)
  
  # return the object
  return(censusData.2020)
}

options("piggyback.verbose" = FALSE)
options("wru_data_wd" = TRUE)

test_that("legacy data returns expected geo groups",{
  skip_on_cran()
  de <- format_legacy_data(PL94171::pl_url('DE', 2020), state = "DE")
  
  expect_named(de, c("county", "tract", "blockGroup", "block"))
})

#' Example voter file.
#'
#' An example dataset containing voter file information.
#'
#' @format A data frame with 10 rows and 12 variables:
#' \describe{
#'   \item{VoterID}{Voter identifier (numeric)}
#'   \item{surname}{Surname}
#'   \item{state}{State of residence}
#'   \item{CD}{Congressional district}
#'   \item{county}{Census county (three-digit code)}
#'   \item{first}{First name}
#'   \item{last}{Last name or surname}
#'   \item{tract}{Census tract (six-digit code)}
#'   \item{block}{Census block (four-digit code)}
#'   \item{precinct}{Voting precinct}
#'   \item{place}{Voting place}
#'   \item{age}{Age in years}
#'   \item{sex}{0=male, 1=female}
#'   \item{party}{Party registration (character)}
#'   \item{PID}{Party registration (numeric)}
#'   #' }
#' @docType data
#' @keywords datasets
#' @name voters
#' @examples
#' data(voters)
"voters"

#' Pre-process vector of names to match census style. Internal function
#'
#' @param voter_names Character vector to be pre-processed.
#' @param target_names Character vector of census names to be matched.
#'
#' @return A character vector of pre-processed named
#' @keywords internal
#'

.name_preproc <- function(voter_names, target_names){
  post_names <- as.character(voter_names)
  post_names <- toupper(post_names)
  ## 1) Raw match
  match_tmp <- post_names %in% target_names
  ## 2) remove punctuation
  post_names[!match_tmp] <- gsub("[^[:alnum:] ]", "",
                                 post_names[!match_tmp])
  match_tmp <- post_names %in% target_names
  ## 3) Remove spaces
  post_names[!match_tmp] <- gsub(" ", "",
                                 post_names[!match_tmp])
  match_tmp <- post_names %in% target_names
  ## 4) Remove suffixes
  suffix <- c("JUNIOR", "SENIOR", "THIRD", "III", "JR", " II", " J R", " S R", " IV")
  for (i in 1:length(suffix)) {
    post_names[!match_tmp] <- ifelse(substr(post_names[!match_tmp],
                                       nchar(post_names[!match_tmp]) - (nchar(suffix)[i] - 1),
                                       nchar(post_names[!match_tmp])) == suffix[i], 
                                substr(post_names[!match_tmp], 1, nchar(post_names[!match_tmp]) - nchar(suffix)[i]), 
                                post_names[!match_tmp])
  }
  post_names[!match_tmp] <- ifelse(nchar(post_names[!match_tmp]) >= 7, 
                                   ifelse(substr(post_names[!match_tmp], 
                                                 nchar(post_names[!match_tmp]) - 1,
                                                 nchar(post_names[!match_tmp])) == "SR", 
                                          substr(post_names[!match_tmp], 1, nchar(post_names[!match_tmp]) - 2), 
                                          post_names[!match_tmp]), 
                                   post_names[!match_tmp]) #Remove "SR" only if name has at least 7 characters
  match_tmp <- post_names %in% target_names
  ## 5) Split double-barreled names, match on first part
  post_names_tmp <- post_names
  post_names_tmp[(!match_tmp) & grep("[-, ]", post_names[(!match_tmp)])] <- sapply(strsplit(grep("[-, ]",
                                                                                                 post_names[(!match_tmp)],
                                                                                                 value = TRUE),
                                                                                            "[-, ]"),
                                                                                   "[", 1) 
  match_tmp_2 <- post_names_tmp %in% target_names
  post_names[(match_tmp_2 == TRUE) & (match_tmp == FALSE)] <- post_names_tmp[(match_tmp_2 == TRUE) & (match_tmp == FALSE)]
  match_tmp <- post_names %in% target_names
  ## 6) Split double-barreled names, match on second part
  post_names_tmp <- post_names
  post_names_tmp[(!match_tmp) & grep("[-, ]", post_names[(!match_tmp)])] <- sapply(strsplit(grep("[-, ]",
                                                                                                 post_names[(!match_tmp)],
                                                                                                 value = TRUE),
                                                                                            "[-, ]"),
                                                                                   "[", 2) 
  match_tmp_2 <- post_names_tmp %in% target_names
  post_names[(match_tmp_2 == TRUE) & (match_tmp == FALSE)] <- post_names_tmp[(match_tmp_2 == TRUE) & (match_tmp == FALSE)]
  return(unlist(post_names))
}

assert_boolean <- function(
    x,
    argument_name = rlang::caller_arg(x),
    call = rlang::caller_call()
) {
  if (length(x) != 1) {
    cli::cli_abort(
      c(
        "{.arg {argument_name}} must be a {.code TRUE} or {.code FALSE} value of length {.val {1}}.",
        x = "{.arg {argument_name}} has a length of {.val {length(x)}}."
      ),
      call = call
    )
  }
  
  if (!inherits(x, "logical")) {
    cli::cli_abort(
      c(
        "{.arg {argument_name}} must be a {.class logical} {.code TRUE} or {.code FALSE} value.",
        x = "{.arg {argument_name}} is an object of class {.cls {class(x)}}."
      ),
      call = call
    )
  }
  
  if (!x %in% c(TRUE, FALSE)) {
    cli::cli_abort(
      c(
        "{.arg {argument_name}} must be {.code TRUE} or {.code FALSE}.",
        x = "{.arg {argument_name}} is {.val {x}}."
      ),
      call = call
    )
  }
  
  x
}

#' Surname probability merging function.
#'
#' \code{merge_surnames} merges surnames in user-input dataset with corresponding 
#'  race/ethnicity probabilities from U.S. Census Surname List and Spanish Surname List.
#'
#' This function allows users to match surnames in their dataset with the U.S. 
#'  Census Surname List (from 2000 or 2010) and Spanish Surname List to obtain 
#'  Pr(Race | Surname) for each of the five major racial groups.
#'  
#'  By default, the function matches surnames to the Census list as follows: 
#'  1) Search raw surnames in Census surname list; 
#'  2) Remove any punctuation and search again; 
#'  3) Remove any spaces and search again; 
#'  4) Remove suffixes (e.g., Jr) and search again; 
#'  5) Split double-barreled surnames into two parts and search first part of name; 
#'  6) Split double-barreled surnames into two parts and search second part of name; 
#'  7) For any remaining names, impute probabilities using distribution 
#'  for all names not appearing on Census list.
#'  
#'  Each step only applies to surnames not matched in a previous ste. 
#'  Steps 2 through 7 are not applied if \code{clean.surname} is FALSE.
#'  
#'  Note: Any name appearing only on the Spanish Surname List is assigned a 
#'  probability of 1 for Hispanics/Latinos and 0 for all other racial groups.
#'
#' @param voter.file An object of class \code{data.frame}. Must contain a field 
#'  named 'surname' containing list of surnames to be merged with Census lists.
#' @param surname.year An object of class \code{numeric} indicating which year 
#'  Census Surname List is from. Accepted values are \code{2010} and \code{2000}. 
#'  Default is \code{2020}.
#' @param name.data An object of class \code{data.frame}. Must contain a leading 
#' column of surnames, and 5 subsequent columns, with Pr(Race | Surname) for each 
#' of the five major racial categories.      
#' @param clean.surname A \code{TRUE}/\code{FALSE} object. If \code{TRUE}, 
#' any surnames in \code{\var{voter.file}} that cannot initially be matched 
#' to surname lists will be cleaned, according to U.S. Census specifications, 
#' in order to increase the chance of finding a match. Default is \code{TRUE}.
#' @param impute.missing A \code{TRUE}/\code{FALSE} object. If \code{TRUE}, 
#' race/ethnicity probabilities will be imputed for unmatched names using  
#' race/ethnicity distribution for all other names (i.e., not on Census List).
#' Default is \code{TRUE}.
#' @return Output will be an object of class \code{data.frame}. It will 
#'  consist of the original user-input data with additional columns that 
#'  specify the part of the name matched with Census data (\code{\var{surname.match}}), 
#'  and the probabilities Pr(Race | Surname) for each racial group 
#'  (\code{\var{p_whi}} for White, \code{\var{p_bla}} for Black, 
#'  \code{\var{p_his}} for Hispanic/Latino, 
#'  \code{\var{p_asi}} for Asian and Pacific Islander, and 
#'  \code{\var{p_oth}} for Other/Mixed).
#'#'
#' @examples
#' data(voters)
#' \dontrun{try(merge_surnames(voters))}
#'
#' @keywords internal

merge_surnames <- function(voter.file, surname.year = 2020, name.data, clean.surname = TRUE, impute.missing = TRUE) {

  if ("surname" %in% names(voter.file) == FALSE) {
    stop('Data does not contain surname field.')
  }
  
  ## Census Surname List
  if (surname.year == 2000) {
    surnames2000$surname <- as.character(surnames2000$surname)
    surnames <- surnames2000
  } else if (surname.year == 2010) {
    surnames2010$surname <- as.character(surnames2010$surname)
    surnames <- surnames2010
  } else {
    surnames <- name.data
    colnames(surnames) <- colnames(surnames2010)
    surnames$surname <- as.character(surnames$surname)
  }
  
  p_eth <- c("p_whi", "p_bla", "p_his", "p_asi", "p_oth")
  
  ## Convert Surnames in Voter File to Upper Case 
  df <- voter.file
  df$surname.match <- df$surname.upper <- toupper(as.character(df$surname))

  ## Merge Surnames with Census List (No Cleaning Yet)
  df <- merge(df[names(df) %in% p_eth == FALSE], surnames[c("surname", p_eth)], by.x = "surname.match", by.y = "surname", all.x = TRUE)

  if (nrow(df[df$surname.upper %in% surnames$surname == F, ]) == 0) {
    return(df[, c(names(voter.file), "surname.match", p_eth)])
  }

  df[df$surname.upper %in% surnames$surname == F, ]$surname.match <- ""
  
  df1 <- df[df$surname.upper %in% surnames$surname, ] #Matched surnames
  df2 <- df[df$surname.upper %in% surnames$surname == F, ] #Unmatched surnames
  
  ## Clean Surnames (if Specified by User)
  if (clean.surname) {
    
    ## Remove All Punctuation and Try Merge Again
    df2$surname.match <- gsub("[^[:alnum:] ]", "", df2$surname.upper)
    df2 <- merge(df2[names(df2) %in% p_eth == FALSE], surnames[c("surname", p_eth)], by.x = "surname.match", by.y = "surname", all.x = TRUE)
    if (nrow(df2[df2$surname.match %in% surnames$surname, ]) > 0) {
      df1 <- rbind(df1, df2[df2$surname.match %in% surnames$surname, ])
      df2 <- df2[df2$surname.match %in% surnames$surname == F, ]
      if (nrow(df2[df2$surname.match %in% surnames$surname, ]) > 0) {df2$surname.match <- ""}
    }

    ## Remove All Spaces and Try Merge Again
    df2$surname.match <- gsub(" ", "", df2$surname.match)
    df2 <- merge(df2[names(df2) %in% p_eth == FALSE], surnames[c("surname", p_eth)], by.x = "surname.match", by.y = "surname", all.x = TRUE)
    if (nrow(df2[df2$surname.match %in% surnames$surname, ]) > 0) {
      df1 <- rbind(df1, df2[df2$surname.match %in% surnames$surname, ])
      df2 <- df2[df2$surname.match %in% surnames$surname == FALSE, ]
      if (nrow(df2[df2$surname.match %in% surnames$surname, ]) > 0) {df2$surname.match <- ""}
    }

    ## Remove Jr/Sr/III Suffixes
    suffix <- c("JUNIOR", "SENIOR", "THIRD", "III", "JR", " II", " J R", " S R", " IV")
    for (i in 1:length(suffix)) {
      df2$surname.match <- ifelse(substr(df2$surname.match, nchar(df2$surname.match) - (nchar(suffix)[i] - 1), nchar(df2$surname.match)) == suffix[i], 
                                  substr(df2$surname.match, 1, nchar(df2$surname.match) - nchar(suffix)[i]), 
                                  df2$surname.match)
    }
    df2$surname.match <- ifelse(nchar(df2$surname.match) >= 7, 
                                ifelse(substr(df2$surname.match, nchar(df2$surname.match) - 1, nchar(df2$surname.match)) == "SR", 
                                       substr(df2$surname.match, 1, nchar(df2$surname.match) - 2), 
                                       df2$surname.match), 
                                df2$surname.match) #Remove "SR" only if name has at least 7 characters
    df2 <- merge(df2[names(df2) %in% p_eth == FALSE], surnames[c("surname", p_eth)], by.x = "surname.match", by.y = "surname", all.x = TRUE)
    if (nrow(df2[df2$surname.match %in% surnames$surname, ]) > 0) {
      df1 <- rbind(df1, df2[df2$surname.match %in% surnames$surname, ])
      df2 <- df2[df2$surname.match %in% surnames$surname == FALSE, ]
      if (nrow(df2[df2$surname.match %in% surnames$surname, ]) > 0) {df2$surname.match <- ""}
    }

    ## Names with Hyphens or Spaces, e.g. Double-Barreled Names
    df2$surname2 <- df2$surname1 <- NA
    df2$surname1[grep("-", df2$surname.upper)] <- sapply(strsplit(grep("-", df2$surname.upper, value = TRUE), "-"), "[", 1)
    df2$surname2[grep("-", df2$surname.upper)] <- sapply(strsplit(grep("-", df2$surname.upper, value = TRUE), "-"), "[", 2)  
    df2$surname1[grep(" ", df2$surname.upper)] <- sapply(strsplit(grep(" ", df2$surname.upper, value = TRUE), " "), "[", 1)
    df2$surname2[grep(" ", df2$surname.upper)] <- sapply(strsplit(grep(" ", df2$surname.upper, value = TRUE), " "), "[", 2)
    
    ## Use first half of name to merge in priors
    df2$surname.match <- as.character(df2$surname1)
    df2 <- merge(df2[names(df2) %in% c(p_eth) == F], surnames[c("surname", p_eth)], by.x = "surname.match", by.y = "surname", all.x = TRUE)[names(df2)]
    if (nrow(df2[df2$surname.match %in% surnames$surname, ]) > 0) {
      df1 <- rbind(df1, df2[df2$surname.match %in% surnames$surname, names(df2) %in% names(df1)])
      df2 <- df2[df2$surname.match %in% surnames$surname == F, ]
      if (nrow(df2[df2$surname.match %in% surnames$surname, ]) > 0) {df2$surname.match <- ""}
    }
    
    ## Use second half of name to merge in priors for rest
    df2$surname.match <- as.character(df2$surname2)
    df2 <- merge(df2[names(df2) %in% c(p_eth, "surname1", "surname2") == F], surnames[c("surname", p_eth)], by.x = "surname.match", by.y = "surname", all.x = TRUE)[names(df2) %in% c("surname1", "surname2") == F]
    if (nrow(df2[df2$surname.match %in% surnames$surname, ]) > 0) {
      df1 <- rbind(df1, df2[df2$surname.match %in% surnames$surname, names(df2) %in% names(df1)])
      df2 <- df2[df2$surname.match %in% surnames$surname == F, ]
      if (nrow(df2[df2$surname.match %in% surnames$surname, ]) > 0) {df2$surname.match <- ""}
    }
  }

  ## Impute priors for names not on Census lists
  if (impute.missing) {
    if (nrow(df2) > 0) {
      df2$surname.match <- ""
      df2$p_whi <- .6665; df2$p_bla <- .0853; df2$p_his <- .1367; df2$p_asi <- .0797; df2$p_oth <- .0318
      message(paste("Probabilities were imputed for", nrow(df2), ifelse(nrow(df2) == 1, "surname", "surnames"), "that could not be matched to Census list."))
    }
  } else message(paste(nrow(df2), ifelse(nrow(df2) == 1, "surname was", "surnames were"), "not matched."))
  
  df <- rbind(df1, df2)
  return(df[, c(names(voter.file), "surname.match", p_eth)])
}

#' Census helper function.
#'
#' \code{census_helper} links user-input dataset with Census geographic data.
#'
#' This function allows users to link their geocoded dataset (e.g., voter file)
#' with U.S. Census data (2010 or 2020). The function extracts Census Summary File data
#' at the county, tract, block, or place level. Census data calculated are
#' Pr(Geolocation | Race) where geolocation is county, tract, block, or place.
#'
#' @inheritParams get_census_data
#' @param voter.file An object of class \code{data.frame}. Must contain field(s) named
#'  \code{\var{county}}, \code{\var{tract}}, \code{\var{block}}, and/or \code{\var{place}}
#'  specifying geolocation. These should be character variables that match up with
#'  U.S. Census categories. County should be three characters (e.g., "031" not "31"),
#'  tract should be six characters, and block should be four characters.
#'  Place should be five characters if it is included.
#' @param states A character vector specifying which states to extract
#'  Census data for, e.g. \code{c("NJ", "NY")}. Default is \code{"all"}, which extracts
#'  Census data for all states contained in user-input data.
#' @param geo A character object specifying what aggregation level to use.
#'  Use \code{"county"}, \code{"tract"}, \code{"block"} or \code{"place"}. Default is \code{"tract"}.
#'  Warning: extracting block-level data takes very long.
#' @param age A \code{TRUE}/\code{FALSE} object indicating whether to condition on
#'  age or not. If \code{FALSE} (default), function will return Pr(Geolocation | Race).
#'  If \code{TRUE}, function will return Pr(Geolocation, Age | Race).
#'  If \code{\var{sex}} is also \code{TRUE}, function will return Pr(Geolocation, Age, Sex | Race).
#' @param sex A \code{TRUE}/\code{FALSE} object indicating whether to condition on
#'  sex or not. If \code{FALSE} (default), function will return Pr(Geolocation | Race).
#'  If \code{TRUE}, function will return Pr(Geolocation, Sex | Race).
#'  If \code{\var{age}} is also \code{TRUE}, function will return Pr(Geolocation, Age, Sex | Race).
#' @param year A character object specifying the year of U.S. Census data to be downloaded.
#'  Use \code{"2010"}, or \code{"2020"}. Default is \code{"2020"}.
#'  Warning: 2020 U.S. Census data is downloaded only when \code{\var{age}} and
#'  \code{\var{sex}} are both \code{FALSE}.
#' @param census.data A optional census object of class \code{list} containing
#' pre-saved Census geographic data. Can be created using \code{get_census_data} function.
#' If \code{\var{census.data}} is provided, the \code{\var{age}} element must have the same value
#' as the \code{\var{age}} option specified in this function (i.e., \code{TRUE} in both or
#' \code{FALSE} in both). Similarly, the \code{\var{sex}} element in the object provided in
#' \code{\var{census.data}} must have the same value as the \code{\var{sex}} option here.
#' Moreover, the \code{\var{year}} element in the object provided in \code{\var{census.data}}
#' must have the same value as the \code{\var{year}} option in the function (i.e., \code{"2010"}
#' in both or \code{"2020"} in both).
#' If \code{\var{census.data}} is missing, Census geographic data will be obtained via Census API.
#' @param retry The number of retries at the census website if network interruption occurs.
#' @param use.counties A logical, defaulting to FALSE. Should census data be filtered by counties available in \var{census.data}?
#' @return Output will be an object of class \code{data.frame}. It will
#'  consist of the original user-input data with additional columns of
#'  Census data.
#'
#' @examples
#' \dontshow{
#' data(voters)
#' }
#' \dontrun{
#' census_helper(voter.file = voters, states = "nj", geo = "block")
#' }
#' \dontrun{
#' census_helper(
#'   voter.file = voters, states = "all", geo = "tract",
#'   age = TRUE, sex = TRUE
#' )
#' }
#' \dontrun{
#' census_helper(
#'   voter.file = voters, states = "all", geo = "county",
#'   age = FALSE, sex = FALSE, year = "2020"
#' )
#' }
#'
#' @keywords internal

census_helper <- function(
    key = Sys.getenv("CENSUS_API_KEY"),
    voter.file,
    states = "all",
    geo = "tract",
    age = FALSE,
    sex = FALSE,
    year = "2020",
    census.data = NULL,
    retry = 3,
    use.counties = FALSE
) {
  if (is.null(census.data) || (typeof(census.data) != "list")) {
    toDownload <- TRUE
  } else {
    toDownload <- FALSE
  }

  if (toDownload) {
    key <- validate_key(key)
  }

  states <- toupper(states)
  if (states == "ALL") {
    states <- toupper(as.character(unique(voter.file$state)))
  }

  df.out <- NULL

  for (s in 1:length(states)) {
    message(paste("State ", s, " of ", length(states), ": ", states[s], sep = ""))
    state <- toupper(states[s])

    if (geo == "place") {
      geo.merge <- c("place")
      if ((toDownload) || (is.null(census.data[[state]])) || (census.data[[state]]$age != age) || (census.data[[state]]$sex != sex) || (census.data[[state]]$year != year)) {
        census <- census_geo_api(key, state, geo = "place", age, sex, year, retry)
      } else {
        census <- census.data[[toupper(state)]]$place
      }
    }

    if (geo == "county") {
      geo.merge <- c("county")
      if ((toDownload) || (is.null(census.data[[state]])) || (census.data[[state]]$age != age) || (census.data[[state]]$sex != sex) || (census.data[[state]]$year != year)) {
        census <- census_geo_api(key, state, geo = "county", age, sex, year, retry)
      } else {
        census <- census.data[[toupper(state)]]$county
      }
    }

    if (geo == "tract") {
      geo.merge <- c("county", "tract")
      if ((toDownload) || (is.null(census.data[[state]])) || (census.data[[state]]$age != age) || (census.data[[state]]$sex != sex) || (census.data[[state]]$year != year)) {
        if (use.counties) {
          census <- census_geo_api(key, state, geo = "tract", age, sex, year, retry, 
                                   # Only those counties within the target state
                                   counties = unique(voter.file$county[voter.file$state == state]))
        } else {
          census <- census_geo_api(key, state, geo = "tract", age, sex, year, retry)
        }
      } else {
        census <- census.data[[toupper(state)]]$tract
      }
    }

    if (geo == "block") {
      geo.merge <- c("county", "tract", "block")
      if ((toDownload) || (is.null(census.data[[state]])) || (census.data[[state]]$age != age) || (census.data[[state]]$sex != sex) || (census.data[[state]]$year != year)) {
        if (use.counties) {
          census <- census_geo_api(key, state, geo = "block", age, sex, year, retry, 
                                   # Only those counties within the target state
                                   counties = unique(voter.file$county[voter.file$state == state]))
        } else {
          census <- census_geo_api(key, state, geo = "block", age, sex, year, retry)
        }
      } else {
        census <- census.data[[toupper(state)]]$block
      }
    }

    if (is.null(census) & use.counties) {
      message("No intersecting counties in counties supplied")
      return(NULL)
    }

    census$state <- state

    if (age == T) {
      ## Add Census Age Categories
      voter.file$agecat <- NA
      voter.file$agecat <- ifelse(voter.file$age <= 4, 1, voter.file$agecat)
      voter.file$agecat <- ifelse(voter.file$age >= 5 & voter.file$age <= 9, 2, voter.file$agecat)
      voter.file$agecat <- ifelse(voter.file$age >= 10 & voter.file$age <= 14, 3, voter.file$agecat)
      voter.file$agecat <- ifelse(voter.file$age >= 15 & voter.file$age <= 17, 4, voter.file$agecat)
      voter.file$agecat <- ifelse(voter.file$age >= 18 & voter.file$age <= 19, 5, voter.file$agecat)
      voter.file$agecat <- ifelse(voter.file$age == 20, 6, voter.file$agecat)
      voter.file$agecat <- ifelse(voter.file$age == 21, 7, voter.file$agecat)
      voter.file$agecat <- ifelse(voter.file$age >= 22 & voter.file$age <= 24, 8, voter.file$agecat)
      voter.file$agecat <- ifelse(voter.file$age >= 25 & voter.file$age <= 29, 9, voter.file$agecat)
      voter.file$agecat <- ifelse(voter.file$age >= 30 & voter.file$age <= 34, 10, voter.file$agecat)
      voter.file$agecat <- ifelse(voter.file$age >= 35 & voter.file$age <= 39, 11, voter.file$agecat)
      voter.file$agecat <- ifelse(voter.file$age >= 40 & voter.file$age <= 44, 12, voter.file$agecat)
      voter.file$agecat <- ifelse(voter.file$age >= 45 & voter.file$age <= 49, 13, voter.file$agecat)
      voter.file$agecat <- ifelse(voter.file$age >= 50 & voter.file$age <= 54, 14, voter.file$agecat)
      voter.file$agecat <- ifelse(voter.file$age >= 55 & voter.file$age <= 59, 15, voter.file$agecat)
      voter.file$agecat <- ifelse(voter.file$age >= 60 & voter.file$age <= 61, 16, voter.file$agecat)
      voter.file$agecat <- ifelse(voter.file$age >= 62 & voter.file$age <= 64, 17, voter.file$agecat)
      voter.file$agecat <- ifelse(voter.file$age >= 65 & voter.file$age <= 66, 18, voter.file$agecat)
      voter.file$agecat <- ifelse(voter.file$age >= 67 & voter.file$age <= 69, 19, voter.file$agecat)
      voter.file$agecat <- ifelse(voter.file$age >= 70 & voter.file$age <= 74, 20, voter.file$agecat)
      voter.file$agecat <- ifelse(voter.file$age >= 75 & voter.file$age <= 79, 21, voter.file$agecat)
      voter.file$agecat <- ifelse(voter.file$age >= 80 & voter.file$age <= 84, 22, voter.file$agecat)
      voter.file$agecat <- ifelse(voter.file$age >= 85, 23, voter.file$agecat)
    }

    if (age == F & sex == F) {

      ## Calculate Pr(Geolocation | Race)
      if (year == "2010") {
        geoPopulations <- rowSums(census[, grepl("P00", names(census))])
        vars <- c(
          pop_white = "P005003", pop_black = "P005004",
          pop_aian = "P005005", pop_asian = "P005006",
          pop_nhpi = "P005007", pop_other = "P005008",
          pop_two = "P005009", pop_hisp = "P005010"
        )
        drop <- c(grep("state", names(census)), grep("P005", names(census)))
      } else if (year == "2020") {
        geoPopulations <- rowSums(census[, grepl("P2_", names(census))])
        vars <- c(
          pop_white = "P2_005N", pop_black = "P2_006N",
          pop_aian = "P2_007N", pop_asian = "P2_008N",
          pop_nhpi = "P2_009N", pop_other = "P2_010N",
          pop_two = "P2_011N", pop_hisp = "P2_002N"
        )
        drop <- c(grep("state", names(census)), grep("P2_", names(census)))
      }

      census$r_whi <- census[, vars["pop_white"]] / sum(census[, vars["pop_white"]]) # Pr(Geo | White)
      census$r_bla <- census[, vars["pop_black"]] / sum(census[, vars["pop_black"]]) # Pr(Geo | Black)
      census$r_his <- census[, vars["pop_hisp"]] / sum(census[, vars["pop_hisp"]]) # Pr(Geo | Latino)
      census$r_asi <- (census[, vars["pop_asian"]] + census[, vars["pop_nhpi"]]) / (sum(census[, vars["pop_asian"]]) + sum(census[, vars["pop_nhpi"]])) # Pr(Geo | Asian or NH/PI)
      census$r_oth <- (census[, vars["pop_aian"]] + census[, vars["pop_other"]] + census[, vars["pop_two"]]) / (sum(census[, vars["pop_aian"]]) + sum(census[, vars["pop_other"]]) + sum(census[, vars["pop_two"]])) # Pr(Geo | AI/AN, Other, or Mixed)

      voters.census <- merge(voter.file[toupper(voter.file$state) == toupper(states[s]), ], census[, -drop], by = geo.merge, all.x = T)
    }

    if (age == F & sex == T) {

      ## Calculate Pr(Geolocation, Sex | Race)
      eth.cen <- c("whi", "bla", "his", "asi", "oth")
      eth.let <- c("I", "B", "H", "D", "F")

      for (i in 1:length(eth.cen)) {
        if (i != 4 & i != 5) {
          census[paste("r_mal", eth.cen[i], sep = "_")] <- census[paste("P012", eth.let[i], "002", sep = "")] / sum(census[paste("P012", eth.let[i], "001", sep = "")])
          census[paste("r_fem", eth.cen[i], sep = "_")] <- census[paste("P012", eth.let[i], "026", sep = "")] / sum(census[paste("P012", eth.let[i], "001", sep = "")])
        }
        if (i == 4) {
          ## Combine Asian and Native Hawaiian/Pacific Islander
          census[paste("r_mal", eth.cen[i], sep = "_")] <- (census$P012D002 + census$P012E002) / sum(census$P012D001 + census$P012E001)
          census[paste("r_fem", eth.cen[i], sep = "_")] <- (census$P012D026 + census$P012E026) / sum(census$P012D001 + census$P012E001)
        }
        if (i == 5) {
          ## Combine American India/Alaska Native and Other
          census[paste("r_mal", eth.cen[i], sep = "_")] <- (census$P012C002 + census$P012F002) / sum(census$P012C001 + census$P012F001)
          census[paste("r_fem", eth.cen[i], sep = "_")] <- (census$P012C026 + census$P012F026) / sum(census$P012C001 + census$P012F001)
        }
      }

      voters.census <- merge(voter.file[toupper(voter.file$state) == toupper(states[s]), ], census[names(census) != "state"], by = geo.merge, all.x = T)
      for (i in 1:length(eth.cen)) {
        voters.census[voters.census$sex == 0, paste("r", eth.cen[i], sep = "_")] <-
          voters.census[voters.census$sex == 0, paste("r_mal", eth.cen[i], sep = "_")]
        voters.census[voters.census$sex == 1, paste("r", eth.cen[i], sep = "_")] <-
          voters.census[voters.census$sex == 1, paste("r_fem", eth.cen[i], sep = "_")]
      }
    }

    if (age == T & sex == F) {

      ## Calculate Pr(Geolocation, Age Category | Race)
      eth.cen <- c("whi", "bla", "his", "asi", "oth")
      eth.let <- c("I", "B", "H", "D", "F")
      age.cat <- c(seq(1, 23), seq(1, 23))
      age.cen <- as.character(c(c("03", "04", "05", "06", "07", "08", "09"), seq(10, 25), seq(27, 49)))

      for (i in 1:length(eth.cen)) {
        for (j in 1:23) {
          if (i != 4 & i != 5) {
            census[paste("r", age.cat[j], eth.cen[i], sep = "_")] <- (census[paste("P012", eth.let[i], "0", age.cen[j], sep = "")] + census[paste("P012", eth.let[i], "0", age.cen[j + 23], sep = "")]) / sum(census[paste("P012", eth.let[i], "001", sep = "")])
          }
          if (i == 4) {
            ## Combine Asian and Native Hawaiian/Pacific Islander
            census[paste("r", age.cat[j], eth.cen[i], sep = "_")] <- (census[paste("P012D0", age.cen[j], sep = "")] + census[paste("P012D0", age.cen[j + 23], sep = "")] + census[paste("P012E0", age.cen[j], sep = "")] + census[paste("P012E0", age.cen[j + 23], sep = "")]) / sum(census$P012D001 + census$P012E001)
          }
          if (i == 5) {
            ## Combine American India/Alaska Native and Other
            census[paste("r", age.cat[j], eth.cen[i], sep = "_")] <- (census[paste("P012C0", age.cen[j], sep = "")] + census[paste("P012C0", age.cen[j + 23], sep = "")] + census[paste("P012F0", age.cen[j], sep = "")] + census[paste("P012F0", age.cen[j + 23], sep = "")]) / sum(census$P012C001 + census$P012F001)
          }
        }
      }

      voters.census <- merge(voter.file[toupper(voter.file$state) == toupper(states[s]), ], census[names(census) != "state"], by = geo.merge, all.x = T)
      for (i in 1:length(eth.cen)) {
        for (j in 1:23) {
          voters.census[voters.census$agecat == j, paste("r", eth.cen[i], sep = "_")] <-
            voters.census[voters.census$agecat == j, paste("r", j, eth.cen[i], sep = "_")]
        }
      }
    }

    if (age == T & sex == T) {

      ## Calculate Pr(Tract, Sex, Age Category | Race)
      eth.cen <- c("whi", "bla", "his", "asi", "oth")
      eth.let <- c("I", "B", "H", "D", "F")
      sex.let <- c("mal", "fem")
      age.cat <- c(seq(1, 23), seq(1, 23))
      age.cen <- as.character(c(c("03", "04", "05", "06", "07", "08", "09"), seq(10, 25), seq(27, 49)))

      for (i in 1:length(eth.cen)) {
        for (k in 1:length(sex.let)) {
          for (j in 1:23) {
            if (k == 2) {
              j <- j + 23
            }
            if (i != 4 & i != 5) {
              census[paste("r", sex.let[k], age.cat[j], eth.cen[i], sep = "_")] <- census[paste("P012", eth.let[i], "0", age.cen[j], sep = "")] / sum(census[paste("P012", eth.let[i], "001", sep = "")])
            }
            if (i == 4) {
              ## Combine Asian and Native Hawaiian/Pacific Islander
              census[paste("r", sex.let[k], age.cat[j], eth.cen[i], sep = "_")] <- (census[paste("P012D0", age.cen[j], sep = "")] + census[paste("P012E0", age.cen[j], sep = "")]) / sum(census$P012D001 + census$P012E001)
            }
            if (i == 5) {
              ## Combine American India/Alaska Native and Other
              census[paste("r", sex.let[k], age.cat[j], eth.cen[i], sep = "_")] <- (census[paste("P012C0", age.cen[j], sep = "")] + census[paste("P012F0", age.cen[j], sep = "")]) / sum(census$P012C001 + census$P012F001)
            }
          }
        }
      }

      voters.census <- merge(voter.file[toupper(voter.file$state) == toupper(states[s]), ], census[names(census) != "state"], by = geo.merge, all.x = T)
      for (i in 1:length(eth.cen)) {
        for (j in 1:23) {
          voters.census[
            voters.census$sex == 0 & voters.census$agecat == j,
            paste("r", eth.cen[i], sep = "_")
          ] <-
            voters.census[
              voters.census$sex == 0 & voters.census$agecat == j,
              paste("r_mal", j, eth.cen[i], sep = "_")
            ]
          voters.census[
            voters.census$sex == 1 & voters.census$agecat == j,
            paste("r", eth.cen[i], sep = "_")
          ] <-
            voters.census[
              voters.census$sex == 1 & voters.census$agecat == j,
              paste("r_fem", j, eth.cen[i], sep = "_")
            ]
        }
      }
    }

    keep.vars <- c(
      names(voter.file)[names(voter.file) != "agecat"],
      paste("r", c("whi", "bla", "his", "asi", "oth"), sep = "_")
    )
    df.out <- as.data.frame(rbind(df.out, voters.census[keep.vars]))
  }

  return(df.out)
}

# Note: must provide a valid U.S. Census API key for test cases that use U.S. Census statistics
# > usethis::edit_r_profile
# Sys.setenv("CENSUS_API_KEY" = "yourkey")
# For testing package coverage use: Sys.setenv("NOT_CRAN" = "TRUE")
options("piggyback.verbose" = FALSE)
options("wru_data_wd" = TRUE)

skip_if_not(nzchar(Sys.getenv("CENSUS_API_KEY")))
test_that("census_helper old still returns predictions", {
  data(voters)
  census <- readRDS(test_path("data/census_test_nj_block_2020.rds"))
  x <- census_helper(
    voter.file = voters, 
    states = "NJ",
    year = "2020",
    census.data = census
    )
  expect_named(x, c('VoterID', 'surname', 'state', 'CD', 'county',
                    'tract', 'block', 'precinct', 'age', 'sex', 'party', 
                    'PID', 'place', 'last', 'first', 'r_whi', 'r_bla', 'r_his', 
                    'r_asi', 'r_oth'))
})

#' Utilities for eBISG (embedding-supplemented BISG)
#'
#' These functions support the eBISG model, which uses pre-trained text
#' embeddings to predict race probabilities for surnames not found in
#' Census surname lists.
#'
#' @name ebisg
NULL

# Package-level environment for caching loaded Python objects within a single
# R session. This caches the *loaded* SentenceTransformer / NameMLP objects in
# RAM to avoid re-deserializing them on every predict_race() call.
#
# The underlying weight files persist on disk:
#   - MLP .pt files: downloaded once via piggyback to tempdir() (or to getwd()
#     if options(wru_data_wd = TRUE)); re-downloads only if absent.
#   - E5-Large tensors: cached by sentence_transformers under
#     ~/.cache/huggingface/hub/.
#
# So no network round-trips happen between sessions. To force a reload of an
# updated .pt within a session, restart R or `rm(list=ls(.ebisg_env))`.
.ebisg_env <- new.env(parent = emptyenv())

#' Registry of built-in embedding models for eBISG
#'
#' Each entry specifies the HuggingFace transformer model, its embedding
#' dimension, and the corresponding MLP checkpoint filenames distributed
#' via piggyback. Users wanting to plug in a different sentence-transformer
#' (e.g., one of the newer top entries on
#' \href{https://huggingface.co/spaces/mteb/leaderboard}{MTEB}) can bypass
#' this registry by passing a list directly to \code{ebisg.model} with their
#' own \code{transformer}, \code{dim}, and \code{surname_mlp} (path to a local
#' .pt checkpoint). See \code{resolve_ebisg_model()} for the schema.
#' @keywords internal
ebisg_model_registry <- list(
  "intfloat/multilingual-e5-large" = list(
    transformer = "intfloat/multilingual-e5-large",
    dim = 1024L,
    surname_mlp = "ebisg-surname-mlp.pt",
    firstname_mlp = "ebisg-firstname-mlp.pt",
    fullname_mlp = "ebisg-fullname-mlp.pt",
    tag = "ebisg-v1.0"
  )
)


#' Resolve an ebisg.model argument into a config list
#'
#' @param ebisg.model A character string giving the HuggingFace model ID
#'   of a built-in model (e.g., \code{"intfloat/multilingual-e5-large"}),
#'   or a named list with elements \code{transformer}, \code{dim},
#'   \code{surname_mlp}, and optionally \code{firstname_mlp}.
#' @return A list with elements \code{transformer}, \code{dim},
#'   \code{surname_mlp}, \code{firstname_mlp}, \code{fullname_mlp}, \code{tag}.
#' @keywords internal
resolve_ebisg_model <- function(ebisg.model) {
  if (is.character(ebisg.model)) {
    # Don't lowercase: HuggingFace model IDs are case-sensitive in the org part.
    if (!(ebisg.model %in% names(ebisg_model_registry))) {
      stop(
        "Unknown ebisg.model '", ebisg.model, "'. ",
        "Available models: ",
        paste(names(ebisg_model_registry), collapse = ", "), ". ",
        "Or pass a custom list with 'transformer', 'dim', and 'surname_mlp'."
      )
    }
    return(ebisg_model_registry[[ebisg.model]])
  } else if (is.list(ebisg.model)) {
    required <- c("transformer", "dim", "surname_mlp")
    missing <- setdiff(required, names(ebisg.model))
    if (length(missing) > 0) {
      stop(
        "Custom ebisg.model list must include: ",
        paste(missing, collapse = ", ")
      )
    }
    # Defaults for optional fields
    if (is.null(ebisg.model$firstname_mlp)) ebisg.model$firstname_mlp <- NULL
    if (is.null(ebisg.model$fullname_mlp)) ebisg.model$fullname_mlp <- NULL
    if (is.null(ebisg.model$tag)) ebisg.model$tag <- NULL
    return(ebisg.model)
  } else {
    stop("ebisg.model must be a character string or a named list.")
  }
}


#' Set up Python environment for eBISG predictions
#'
#' Installs the required Python packages (sentence-transformers and torch)
#' and downloads pre-trained MLP model weights. Call this once before using
#' \code{predict_race(..., model = "eBISG")}.
#'
#' @param envname Name of the Python virtual environment to use. Defaults to
#'   \code{"r-ebisg"}.
#' @param ebisg.model Which embedding model to set up. Defaults to
#'   \code{"intfloat/multilingual-e5-large"}.
#'
#' @export
setup_ebisg <- function(envname = "r-ebisg",
                        ebisg.model = "intfloat/multilingual-e5-large") {
  if (!requireNamespace("reticulate", quietly = TRUE)) {
    stop(
      "The 'reticulate' package is required for eBISG. ",
      "Install with: install.packages('reticulate')"
    )
  }
  cfg <- resolve_ebisg_model(ebisg.model)
  message("Installing Python packages into virtualenv '", envname, "'...")
  reticulate::py_install(
    c("sentence-transformers", "torch"),
    pip = TRUE,
    envname = envname
  )
  # Wire the freshly-set-up env into this R session so subsequent
  # py_module_available() / py$ lookups hit the right interpreter.
  reticulate::use_virtualenv(envname, required = TRUE)
  message("Downloading eBISG model weights...")
  ebisg_data_preflight(cfg)
  message(
    "eBISG setup complete. ",
    "You can now use predict_race(..., model = 'eBISG')"
  )
}


#' Check that Python dependencies for eBISG are available
#' @keywords internal
ensure_ebisg_python <- function() {
  if (!requireNamespace("reticulate", quietly = TRUE)) {
    stop(
      "The 'reticulate' package is required for model='eBISG'. ",
      "Install with: install.packages('reticulate')"
    )
  }
  if (!reticulate::py_module_available("sentence_transformers")) {
    stop(
      "Python package 'sentence-transformers' is required for model='eBISG'.\n",
      "Run setup_ebisg() or install manually with:\n",
      "  reticulate::py_install('sentence-transformers')\n",
      "  reticulate::py_install('torch')"
    )
  }
  if (!reticulate::py_module_available("torch")) {
    stop(
      "Python package 'torch' is required for model='eBISG'.\n",
      "Run setup_ebisg() or install manually with:\n",
      "  reticulate::py_install('torch')"
    )
  }
}


#' Download eBISG model weights via piggyback
#'
#' @param cfg A model config list from \code{resolve_ebisg_model()}.
#' @return The destination directory path.
#' @keywords internal
ebisg_data_preflight <- function(cfg) {
  dest <- ifelse(
    getOption("wru_data_wd", default = FALSE), getwd(), tempdir()
  )

  # For custom models with local paths, skip download
  if (is.null(cfg$tag)) {
    # Custom model: check that files exist at the given paths
    if (!file.exists(cfg$surname_mlp)) {
      stop("Custom surname MLP not found: ", cfg$surname_mlp)
    }
    return(dirname(cfg$surname_mlp))
  }

  files <- c(cfg$surname_mlp)
  if (!is.null(cfg$firstname_mlp)) files <- c(files, cfg$firstname_mlp)
  if (!is.null(cfg$fullname_mlp)) files <- c(files, cfg$fullname_mlp)

  # Only download if files are not already present
  missing <- !file.exists(file.path(dest, files))
  if (any(missing)) {
    tryCatch(
      piggyback::pb_download(
        file = files[missing],
        repo = "kosukeimai/wru",
        dest = dest,
        .token = "",
        tag = cfg$tag
      ),
      error = function(e) {
        stop(
          "Failed to download eBISG model weights: ", e$message, "\n",
          "Run setup_ebisg() to configure the eBISG environment."
        )
      }
    )
  }
  dest
}


#' Load the Python eBISG helper module
#'
#' Imports `inst/python/ebisg_helper.py` as a private module via
#' `reticulate::import_from_path()` and caches the module object on
#' `.ebisg_env`. Using `import_from_path` keeps the helper's symbols
#' encapsulated rather than dumping them into Python's global namespace
#' as `source_python` would.
#'
#' @keywords internal
get_ebisg_module <- function() {
  if (is.null(.ebisg_env$module)) {
    helper_path <- system.file("python", package = "wru")
    if (helper_path == "") {
      stop("Cannot find inst/python/ebisg_helper.py in the wru package.")
    }
    .ebisg_env$module <- reticulate::import_from_path(
      "ebisg_helper", path = helper_path
    )
  }
  .ebisg_env$module
}


#' Embed names using a sentence-transformer model via Python
#'
#' @param names Character vector of names to embed.
#' @param transformer HuggingFace model ID (e.g.,
#'   \code{"intfloat/multilingual-e5-large"}).
#' @param cache_dir Optional cache directory for model weights.
#' @return A matrix of dimensions (length(names), embedding_dim).
#' @keywords internal
ebisg_embed_names <- function(names, transformer = "intfloat/multilingual-e5-large",
                              cache_dir = NULL) {
  mod <- get_ebisg_module()
  embeddings <- mod$embed_names(
    as.list(names),
    model_name = transformer,
    cache_dir = cache_dir
  )
  as.matrix(embeddings)
}


#' Run MLP inference on embeddings
#'
#' @param embeddings Numeric matrix (N x dim) of name embeddings.
#' @param model_path Path to the .pt checkpoint file.
#' @return A matrix of dimensions (N, 6) with columns
#'   (whi, bla, his, asi, aian, oth).
#' @keywords internal
ebisg_predict_mlp <- function(embeddings, model_path) {
  mod <- get_ebisg_module()

  # Cache loaded models by path
  if (is.null(.ebisg_env$models)) {
    .ebisg_env$models <- list()
  }
  if (is.null(.ebisg_env$models[[model_path]])) {
    result <- mod$load_mlp(model_path)
    .ebisg_env$models[[model_path]] <- result[[1]]
  }

  model <- .ebisg_env$models[[model_path]]
  probs <- as.matrix(mod$predict_mlp(embeddings, model))
  # Source of truth for column ordering lives in Python (EBISG_RACE_COLS).
  colnames(probs) <- as.character(mod$EBISG_RACE_COLS)
  probs
}


#' Map 6-class eBISG output to wru's 5-class system
#'
#' The MLP outputs probabilities for 6 categories:
#' (whi, bla, his, asi, aian, oth) -- see \code{EBISG_RACE_COLS} in the
#' Python helper. wru uses 5 categories: (whi, bla, his, asi, oth).
#' This function sums \code{aian} and \code{oth} into \code{oth}.
#'
#' @param probs_6 A matrix with 6 columns named whi, bla, his, asi, aian, oth.
#' @return A matrix with 5 columns named c_whi, c_bla, c_his, c_asi, c_oth.
#' @keywords internal
map_6class_to_5class <- function(probs_6) {
  out <- cbind(
    probs_6[, "whi"],
    probs_6[, "bla"],
    probs_6[, "his"],
    probs_6[, "asi"],
    probs_6[, "aian"] + probs_6[, "oth"]
  )
  colnames(out) <- c("c_whi", "c_bla", "c_his", "c_asi", "c_oth")
  out
}

test_that("predict_race rejects unknown model strings", {
  data(voters)
  # Confirms the validation block in predict_race() actually fires; the
  # eBISG-accepting path is exercised by the end-to-end test below (when
  # Python is available).
  expect_error(
    predict_race(voter.file = voters, surname.only = TRUE, model = "not_a_model"),
    "model"
  )
})

test_that("map_6class_to_5class produces correct output", {
  probs_6 <- matrix(c(0.5, 0.1, 0.2, 0.1, 0.05, 0.05), nrow = 1,
                    dimnames = list(NULL, c("whi", "bla", "his", "asi", "aian", "oth")))
  result <- map_6class_to_5class(probs_6)
  expect_equal(ncol(result), 5)
  expect_equal(as.numeric(rowSums(result)), 1, tolerance = 1e-6)
  expect_equal(as.numeric(result[1, "c_whi"]), 0.5)
  expect_equal(as.numeric(result[1, "c_oth"]), 0.10)  # aian + oth = 0.05 + 0.05
})

test_that("map_6class_to_5class handles multiple rows", {
  probs_6 <- matrix(1 / 6, nrow = 5, ncol = 6,
                    dimnames = list(NULL, c("whi", "bla", "his", "asi", "aian", "oth")))
  result <- map_6class_to_5class(probs_6)
  expect_equal(nrow(result), 5)
  expect_equal(ncol(result), 5)
  expect_equal(as.numeric(rowSums(result)), rep(1, 5), tolerance = 1e-6)
})

test_that("ensure_ebisg_python gives clear error without reticulate", {
  skip_on_cran()
  # Only meaningful when reticulate isn't installed; skip otherwise so
  # testthat doesn't flag this as an empty test on CI (where it is installed).
  skip_if(requireNamespace("reticulate", quietly = TRUE),
          "reticulate is installed; this test is for the missing-deps path")
  expect_error(ensure_ebisg_python(), "reticulate")
})

test_that("eBISG predictions work end-to-end", {
  skip_on_cran()
  skip_if_not_installed("reticulate")
  skip_if_not(reticulate::py_module_available("sentence_transformers"),
              "sentence-transformers not available")
  skip_if_not(reticulate::py_module_available("torch"),
              "torch not available")

  data(voters)
  result <- predict_race(voter.file = voters, surname.only = TRUE, model = "eBISG")

  pred_cols <- grep("^pred\\.", names(result), value = TRUE)
  expect_equal(length(pred_cols), 5)

  # Probabilities should be valid
  for (col in pred_cols) {
    expect_true(all(result[[col]] >= 0 & result[[col]] <= 1))
  }

  # Row sums should equal 1
  row_sums <- rowSums(result[, pred_cols])
  expect_equal(row_sums, rep(1, nrow(result)), tolerance = 1e-6)
})

#' @importFrom rlang %||%
validate_key <- function(
    key,
    argument_name = rlang::caller_arg(key),
    call = rlang::caller_call()
) {
  key <- key %||% Sys.getenv("CENSUS_API_KEY")
  
  if (length(key) != 1) {
    cli::cli_abort(
      c(
        "{.arg {argument_name}} must be a {.cls character} string of length {.val {1}}.",
        x = "{.arg {argument_name}} has a length of {.val {length(key)}}."
      ),
      call = call
    )
  }
  
  if (!inherits(key, "character")) {
    cli::cli_abort(
      c(
        "{.arg {argument_name}} must be a {.cls character} string of length {.val {1}}.",
        x = "{.arg {argument_name}} is an object of class {.cls {class(key)}}."
      ),
      call = call
    )
  }
  
  if (!nzchar(key)) {
    cli::cli_abort(
      c(
        "{.arg {argument_name}} must not be an empty string.",
        "*" = "Have you set the {.envvar CENSUS_API_KEY} environment variable?
        See {.help wru::get_census_data} for more information."
      ),
      call = call
    )
  }
  
  if (is.na(key)) {
    cli::cli_abort(
      "{.arg {argument_name}} must not be {.val {NA_character_}}.",
      call = call
    )
  }
  
  key
}

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

#' Collapsed Gibbs sampler for hWRU. Internal function
#'
#' @param last_name Integer vector of last name identifiers for each record (zero indexed; as all that follow). Must match columns numbers in M_rs.  
#' @param first_name See last_name
#' @param mid_name See last_name
#' @param geo Integer vector of geographic units for each record. Must match column number in N_rg
#' @param N_rg Integer matrix of race | geography counts in census (geograpgies in columns).
#' @param pi_s Numeric matrix of race | surname prior probabilities.
#' @param pi_f Same as `pi_s`, but for first names.
#' @param pi_m Same as `pi_s`, but for middle names.
#' @param pi_nr Matrix of marginal probability distribution over missing names; non-keyword names default to this distribution.
#' @param which_names Integer; 0=surname only. 1=surname + first name. 2= surname, first, and middle names.
#' @param samples Integer number of samples to take after (in total)
#' @param burnin Integer number of samples to discard as burn-in of Markov chain
#' @param race_init Integer vector of initial race assignments
#' @param verbose Boolean; should informative messages be printed?
#'
#' @keywords internal
sample_me <- function(last_name, first_name, mid_name, geo, N_rg, pi_s, pi_f, pi_m, pi_nr, which_names, samples, burnin, race_init, verbose) {
    .Call(`_wru_sample_me`, last_name, first_name, mid_name, geo, N_rg, pi_s, pi_f, pi_m, pi_nr, which_names, samples, burnin, race_init, verbose)
}

determine_geo_id_names <- function(census.geo) {
  switch(
    census.geo,
    "tract" = c("county", "tract"),
    "block_group" = c("county", "tract", "block_group"),
    "block" = c("county", "tract", "block"),
    # Return `census.geo` unchanged for county, place, and zcta
    census.geo
  )
}

.onAttach <- function(libname, pkgname) {
  packageStartupMessage(
    "\n",
    "Please cite as:", "\n\n",
    format(utils::citation("wru"), style = "text"), "\n\n",
    "Note that wru 2.0.0 uses 2020 census data by default.", "\n",
    'Use the argument `year = "2010"`, to replicate analyses produced with earlier package versions.',
    "\n"
  )
}

census_geo_api_zcta <- function(
    census_data_url,
    key,
    vars,
    state,
    counties,
    retry
) {
  if (!is.null(counties)) {
    cli::cli_abort(
      '{.arg counties} must be {.code NULL} when {.code geo = "zcta"},
        because ZCTA-level census data split by county is not available.'
    )
  }
  
  region <- paste0(
    "for=zip%20code%20tabulation%20area%20(or%20part):*&in=state:",
    paste(as_fips_code(state), collapse = ",")
  )
  
  census <- get_census_api(
    census_data_url,
    key = key,
    var.names = unlist(vars),
    region = region,
    retry
  )
  
  names(census)[[2]] <- "zcta"
  
  census
}

#' Census Surname List (2000).
#'
#' Census Surname List from 2000 with race/ethnicity probabilities by surname.
#'
#' @format A data frame with 157,728 rows and 6 variables:
#' \describe{
#'   \item{surname}{Surname}
#'   \item{p_whi}{Pr(White | Surname)}
#'   \item{p_bla}{Pr(Black | Surname)}
#'   \item{p_his}{Pr(Hispanic/Latino | Surname)}
#'   \item{p_asi}{Pr(Asian/Pacific Islander | Surname)}
#'   \item{p_oth}{Pr(Other | Surname)}
#'   #' }
#'
#' @docType data
#' @keywords datasets
#' @name surnames2000
#' @examples
#' data(surnames2000)
"surnames2000"

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merge_surnames.R
\name{merge_surnames}
\alias{merge_surnames}
\title{Surname probability merging function.}
\usage{
merge_surnames(
  voter.file,
  surname.year = 2020,
  name.data,
  clean.surname = TRUE,
  impute.missing = TRUE
)
}
\arguments{
\item{voter.file}{An object of class \code{data.frame}. Must contain a field
named 'surname' containing list of surnames to be merged with Census lists.}

\item{surname.year}{An object of class \code{numeric} indicating which year
Census Surname List is from. Accepted values are \code{2010} and \code{2000}.
Default is \code{2020}.}

\item{name.data}{An object of class \code{data.frame}. Must contain a leading
column of surnames, and 5 subsequent columns, with Pr(Race | Surname) for each
of the five major racial categories.}

\item{clean.surname}{A \code{TRUE}/\code{FALSE} object. If \code{TRUE},
any surnames in \code{\var{voter.file}} that cannot initially be matched
to surname lists will be cleaned, according to U.S. Census specifications,
in order to increase the chance of finding a match. Default is \code{TRUE}.}

\item{impute.missing}{A \code{TRUE}/\code{FALSE} object. If \code{TRUE},
race/ethnicity probabilities will be imputed for unmatched names using
race/ethnicity distribution for all other names (i.e., not on Census List).
Default is \code{TRUE}.}
}
\value{
Output will be an object of class \code{data.frame}. It will
consist of the original user-input data with additional columns that
specify the part of the name matched with Census data (\code{\var{surname.match}}),
and the probabilities Pr(Race | Surname) for each racial group
(\code{\var{p_whi}} for White, \code{\var{p_bla}} for Black,
\code{\var{p_his}} for Hispanic/Latino,
\code{\var{p_asi}} for Asian and Pacific Islander, and
\code{\var{p_oth}} for Other/Mixed).
#'
}
\description{
\code{merge_surnames} merges surnames in user-input dataset with corresponding
race/ethnicity probabilities from U.S. Census Surname List and Spanish Surname List.
}
\details{
This function allows users to match surnames in their dataset with the U.S.
Census Surname List (from 2000 or 2010) and Spanish Surname List to obtain
Pr(Race | Surname) for each of the five major racial groups.

By default, the function matches surnames to the Census list as follows:
\enumerate{
\item Search raw surnames in Census surname list;
\item Remove any punctuation and search again;
\item Remove any spaces and search again;
\item Remove suffixes (e.g., Jr) and search again;
\item Split double-barreled surnames into two parts and search first part of name;
\item Split double-barreled surnames into two parts and search second part of name;
\item For any remaining names, impute probabilities using distribution
for all names not appearing on Census list.
}

Each step only applies to surnames not matched in a previous ste.
Steps 2 through 7 are not applied if \code{clean.surname} is FALSE.

Note: Any name appearing only on the Spanish Surname List is assigned a
probability of 1 for Hispanics/Latinos and 0 for all other racial groups.
}
\examples{
data(voters)
\dontrun{try(merge_surnames(voters))}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ebisg_utils.R
\name{ebisg_data_preflight}
\alias{ebisg_data_preflight}
\title{Download eBISG model weights via piggyback}
\usage{
ebisg_data_preflight(cfg)
}
\arguments{
\item{cfg}{A model config list from \code{resolve_ebisg_model()}.}
}
\value{
The destination directory path.
}
\description{
Download eBISG model weights via piggyback
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/race_prediction_funs.R
\name{modfuns}
\alias{modfuns}
\alias{.predict_race_old}
\alias{predict_race_new}
\alias{predict_race_me}
\alias{predict_race_embedding}
\title{Internal model fitting functions}
\usage{
.predict_race_old(
  voter.file,
  census.surname = TRUE,
  surname.only = FALSE,
  surname.year = 2020,
  name.dictionaries = NULL,
  census.geo,
  census.key = Sys.getenv("CENSUS_API_KEY"),
  census.data = NULL,
  age = FALSE,
  sex = FALSE,
  year = "2020",
  party,
  retry = 3,
  impute.missing = TRUE,
  use.counties = FALSE
)

predict_race_new(
  voter.file,
  names.to.use,
  year = "2020",
  age = FALSE,
  sex = FALSE,
  census.geo = c("tract", "block", "block_group", "county", "place", "zcta"),
  census.key = Sys.getenv("CENSUS_API_KEY"),
  name.dictionaries,
  surname.only = FALSE,
  census.data = NULL,
  retry = 0,
  impute.missing = TRUE,
  skip_bad_geos = FALSE,
  census.surname = FALSE,
  use.counties = FALSE
)

predict_race_me(
  voter.file,
  names.to.use,
  year = "2020",
  age = FALSE,
  sex = FALSE,
  census.geo = c("tract", "block", "block_group", "county", "place", "zcta"),
  census.key = Sys.getenv("CENSUS_API_KEY"),
  name.dictionaries,
  surname.only = FALSE,
  census.data = NULL,
  retry = 0,
  impute.missing = TRUE,
  census.surname = FALSE,
  use.counties = FALSE,
  race.init,
  ctrl
)

predict_race_embedding(
  voter.file,
  names.to.use,
  year = "2020",
  age = FALSE,
  sex = FALSE,
  census.geo = c("tract", "block", "block_group", "county", "place", "zcta"),
  census.key = Sys.getenv("CENSUS_API_KEY"),
  name.dictionaries,
  surname.only = FALSE,
  census.data = NULL,
  retry = 0,
  impute.missing = TRUE,
  skip_bad_geos = FALSE,
  census.surname = FALSE,
  use.counties = FALSE,
  ebisg.model = "intfloat/multilingual-e5-large"
)
}
\arguments{
\item{voter.file}{See documentation in \code{race_predict}.}

\item{census.surname}{See documentation in \code{race_predict}.}

\item{surname.only}{See documentation in \code{race_predict}.}

\item{surname.year}{See documentation in \code{race_predict}.}

\item{name.dictionaries}{See documentation in \code{race_predict}.}

\item{census.geo}{See documentation in \code{race_predict}.}

\item{census.key}{A character object specifying user's Census API key.
Required if \code{census.geo} is specified, because a valid Census API key is
required to download Census geographic data.

If \code{\link{NULL}}, the default, attempts to find a census key stored in an
\link[=Sys.getenv]{environment variable} named \code{CENSUS_API_KEY}.}

\item{census.data}{See documentation in \code{race_predict}.}

\item{age}{See documentation in \code{race_predict}.}

\item{sex}{See documentation in \code{race_predict}.}

\item{year}{See documentation in \code{race_predict}.}

\item{party}{See documentation in \code{race_predict}.}

\item{retry}{See documentation in \code{race_predict}.}

\item{impute.missing}{See documentation in \code{race_predict}.}

\item{use.counties}{A logical, defaulting to FALSE. Should census data be filtered by counties available in \var{census.data}?}

\item{names.to.use}{See documentation in \code{race_predict}.}

\item{skip_bad_geos}{See documentation in \code{race_predict}.}

\item{race.init}{See documentation in \code{race_predict}.}

\item{ctrl}{See \code{control} in documentation for \code{\link[=predict_race]{predict_race()}}.}

\item{ebisg.model}{Character string (HuggingFace model ID) or named list
specifying which embedding model to use when \code{model = "eBISG"}.
The only built-in option is \code{"intfloat/multilingual-e5-large"}
(default; 1024-dim), which is keyed by its full HuggingFace ID for
consistency with the Python side. To use a different sentence-transformer,
pass a named list with elements \code{transformer} (HuggingFace model ID),
\code{dim} (embedding dimension), and \code{surname_mlp},
\code{firstname_mlp} (paths to custom .pt checkpoints).}
}
\value{
Output will be an object of class \code{data.frame}. It will
consist of the original user-input \code{voter.file} with additional columns with
predicted probabilities for each of the five major racial categories:
\code{\var{pred.whi}} for White,
\code{\var{pred.bla}} for Black,
\code{\var{pred.his}} for Hispanic/Latino,
\code{\var{pred.asi}} for Asian/Pacific Islander, and
\code{\var{pred.oth}} for Other/Mixed.
}
\description{
These functions are intended for internal use only. Users should use the
\code{\link[=predict_race]{predict_race()}} interface rather any of these functions directly.
}
\details{
These functions fit different versions of WRU. \code{.predict_race_old} fits
the original WRU model, also known as BISG with census-based surname dictionary.
\code{.predict_race_new} fits a new version of BISG which uses a new, augmented
surname dictionary, and can also accommodate the use of first and middle
name information. Finally, \code{.predict_race_me} fits a fully Bayesian Improved
Surname Geocoding model (fBISG), which fits a model with measurement-error
correction of erroneous zeros in census tables, in addition to also accommodating
the augmented surname dictionary, and the first and middle name
dictionaries when making predictions.
}
\section{.predict_race_old}{

Original WRU race prediction function, implementing classical BISG with census-based
surname dictionary.
}

\section{.predict_race_new }{

New race prediction function, implementing classical BISG with augmented
surname dictionary, as well as first and middle name information.
}

\section{.predict_race_me}{

New race prediction function, implementing fBISG (i.e. measurement
error correction, fully Bayesian model) with augmented
surname dictionary, as well as first and middle name information.
}

\section{.predict_race_embedding}{

eBISG race prediction function, which uses pre-trained text embeddings
(E5-Large) to predict race probabilities for names not found in Census
surname lists, rather than falling back to generic population-level priors.
}

\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ebisg_utils.R
\name{setup_ebisg}
\alias{setup_ebisg}
\title{Set up Python environment for eBISG predictions}
\usage{
setup_ebisg(
  envname = "r-ebisg",
  ebisg.model = "intfloat/multilingual-e5-large"
)
}
\arguments{
\item{envname}{Name of the Python virtual environment to use. Defaults to
\code{"r-ebisg"}.}

\item{ebisg.model}{Which embedding model to set up. Defaults to
\code{"intfloat/multilingual-e5-large"}.}
}
\description{
Installs the required Python packages (sentence-transformers and torch)
and downloads pre-trained MLP model weights. Call this once before using
\code{predict_race(..., model = "eBISG")}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vec_to_chunk.R
\name{vec_to_chunk}
\alias{vec_to_chunk}
\title{Variable vector into chunks.}
\usage{
vec_to_chunk(x)
}
\arguments{
\item{x}{Character vector of variable names.}
}
\value{
Object of class \code{list}.
}
\description{
\code{vec_to_chunk} takes a list of variables and collects them into 50-variable chunks.
}
\details{
This function takes a list of variable names and collects them into chunks with no more than
50 variables each. This helps to get around requests with more than 50 variables,because the
API only allows queries of 50 variables at a time.
The user should not need to call this function directly.
}
\examples{
\dontrun{
vec_to_chunk(x = c(paste("P012F0", seq(10:49), sep = ""),
              paste("P012I0", seq(10, 49), sep = "")))
}

}
\references{
Based on code authored by Nicholas Nagle, which is available
\href{https://rstudio-pubs-static.s3.amazonaws.com/19337_2e7f827190514c569ea136db788ce850.html}{here}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ebisg_utils.R
\name{ebisg_embed_names}
\alias{ebisg_embed_names}
\title{Embed names using a sentence-transformer model via Python}
\usage{
ebisg_embed_names(
  names,
  transformer = "intfloat/multilingual-e5-large",
  cache_dir = NULL
)
}
\arguments{
\item{names}{Character vector of names to embed.}

\item{transformer}{HuggingFace model ID (e.g.,
\code{"intfloat/multilingual-e5-large"}).}

\item{cache_dir}{Optional cache directory for model weights.}
}
\value{
A matrix of dimensions (length(names), embedding_dim).
}
\description{
Embed names using a sentence-transformer model via Python
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{sample_me}
\alias{sample_me}
\title{Collapsed Gibbs sampler for hWRU. Internal function}
\usage{
sample_me(
  last_name,
  first_name,
  mid_name,
  geo,
  N_rg,
  pi_s,
  pi_f,
  pi_m,
  pi_nr,
  which_names,
  samples,
  burnin,
  race_init,
  verbose
)
}
\arguments{
\item{last_name}{Integer vector of last name identifiers for each record (zero indexed; as all that follow). Must match columns numbers in M_rs.}

\item{first_name}{See last_name}

\item{mid_name}{See last_name}

\item{geo}{Integer vector of geographic units for each record. Must match column number in N_rg}

\item{N_rg}{Integer matrix of race | geography counts in census (geograpgies in columns).}

\item{pi_s}{Numeric matrix of race | surname prior probabilities.}

\item{pi_f}{Same as \code{pi_s}, but for first names.}

\item{pi_m}{Same as \code{pi_s}, but for middle names.}

\item{pi_nr}{Matrix of marginal probability distribution over missing names; non-keyword names default to this distribution.}

\item{which_names}{Integer; 0=surname only. 1=surname + first name. 2= surname, first, and middle names.}

\item{samples}{Integer number of samples to take after (in total)}

\item{burnin}{Integer number of samples to discard as burn-in of Markov chain}

\item{race_init}{Integer vector of initial race assignments}

\item{verbose}{Boolean; should informative messages be printed?}
}
\description{
Collapsed Gibbs sampler for hWRU. Internal function
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ebisg_utils.R
\name{ebisg}
\alias{ebisg}
\title{Utilities for eBISG (embedding-supplemented BISG)}
\description{
These functions support the eBISG model, which uses pre-trained text
embeddings to predict race probabilities for surnames not found in
Census surname lists.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ebisg_utils.R
\name{ebisg_predict_mlp}
\alias{ebisg_predict_mlp}
\title{Run MLP inference on embeddings}
\usage{
ebisg_predict_mlp(embeddings, model_path)
}
\arguments{
\item{embeddings}{Numeric matrix (N x dim) of name embeddings.}

\item{model_path}{Path to the .pt checkpoint file.}
}
\value{
A matrix of dimensions (N, 6) with columns
(whi, bla, his, asi, aian, oth).
}
\description{
Run MLP inference on embeddings
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_state_fips.R
\docType{data}
\name{state_fips}
\alias{state_fips}
\title{Dataset with FIPS codes for US states}
\format{
A tibble with 57 rows and 3 columns:
\describe{
\item{\code{state}}{Two-letter postal abbreviation}
\item{\code{state_code}}{Two-digit FIPS code}
\item{\code{state_name}}{English name}
}
}
\source{
Derived from \code{\link[tidycensus:fips_codes]{tidycensus::fips_codes()}}
}
\usage{
state_fips
}
\description{
Dataset including FIPS codes and postal abbreviations for each U.S. state,
district, and territory.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merge_names.R
\name{merge_names}
\alias{merge_names}
\title{Surname probability merging function.}
\usage{
merge_names(
  voter.file,
  namesToUse,
  census.surname,
  table.surnames = NULL,
  table.first = NULL,
  table.middle = NULL,
  clean.names = TRUE,
  impute.missing = FALSE,
  model = "BISG"
)
}
\arguments{
\item{voter.file}{An object of class \code{data.frame}.  Must contain a row for each individual being predicted,
as well as a field named \code{\var{last}} containing each individual's surname.
If first name is also being used for prediction, the file must also contain a field
named \code{\var{first}}. If middle name is also being used for prediction, the field
must also contain a field named \code{\var{middle}}.}

\item{namesToUse}{A character vector identifying which names to use for the prediction.
The default value is \code{"last"}, indicating that only the last name will be used.
Other options are \code{"last, first"}, indicating that both last and first names will be
used, and \code{"last, first, middle"}, indicating that last, first, and middle names will all
be used.}

\item{census.surname}{A \code{TRUE}/\code{FALSE} object. If \code{TRUE},
function will call \code{merge_surnames} to merge in Pr(Race | Surname)
from U.S. Census Surname List (2000, 2010, or 2020) and Spanish Surname List.
If \code{FALSE}, user must provide a \code{name.dictionary} (see below).
Default is \code{TRUE}.}

\item{table.surnames}{An object of class \code{data.frame} provided by the
users as an alternative surname dictionary. It will consist of a list of
U.S. surnames, along with the associated probabilities P(name | ethnicity)
for ethnicities: white, Black, Hispanic, Asian, and other. Default is \code{NULL}.
(\code{\var{last_name}} for U.S. surnames, \code{\var{p_whi_last}} for White,
\code{\var{p_bla_last}} for Black, \code{\var{p_his_last}} for Hispanic,
\code{\var{p_asi_last}} for Asian, \code{\var{p_oth_last}} for other).}

\item{table.first}{See \code{\var{table.surnames}}.}

\item{table.middle}{See \code{\var{table.surnames}}.}

\item{clean.names}{A \code{TRUE}/\code{FALSE} object. If \code{TRUE},
any surnames in \code{\var{voter.file}} that cannot initially be matched
to the database will be cleaned, according to U.S. Census specifications,
in order to increase the chance of finding a match. Default is \code{TRUE}.}

\item{impute.missing}{See \code{predict_race}.}

\item{model}{See \code{predict_race}.}
}
\value{
Output will be an object of class \code{data.frame}. It will
consist of the original user-input data with additional columns that
specify the part of the name matched with Census data (\code{\var{surname.match}}),
and the probabilities Pr(Race | Surname) for each racial group
(\code{\var{p_whi}} for White, \code{\var{p_bla}} for Black,
\code{\var{p_his}} for Hispanic/Latino,
\code{\var{p_asi}} for Asian and Pacific Islander, and
\code{\var{p_oth}} for Other/Mixed).
}
\description{
\code{merge_names} merges names in a user-input dataset with corresponding
race/ethnicity probabilities derived from both the U.S. Census Surname List
and Spanish Surname List and voter files from states in the Southern U.S.
}
\details{
This function allows users to match names in their dataset with database entries
estimating P(name | ethnicity) for each of the five major racial groups for each
name. The database probabilities are derived from both the U.S. Census Surname List
and Spanish Surname List and voter files from states in the Southern U.S.

By default, the function matches names as follows:
\enumerate{
\item Search raw surnames in the database;
\item Remove any punctuation and search again;
\item Remove any spaces and search again;
\item Remove suffixes (e.g., "Jr") and search again (last names only)
\item Split double-barreled names into two parts and search first part of name;
\item Split double-barreled names into two parts and search second part of name;
}

Each step only applies to names not matched in a previous step.
Steps 2 through 6 are not applied if \code{clean.surname} is FALSE.

Note: Any name appearing only on the Spanish Surname List is assigned a
probability of 1 for Hispanics/Latinos and 0 for all other racial groups.
}
\examples{
data(voters)
\dontrun{try(merge_names(voters, namesToUse = "surname", census.surname = TRUE))}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ebisg_utils.R
\docType{data}
\name{ebisg_model_registry}
\alias{ebisg_model_registry}
\title{Registry of built-in embedding models for eBISG}
\format{
An object of class \code{list} of length 1.
}
\usage{
ebisg_model_registry
}
\description{
Each entry specifies the HuggingFace transformer model, its embedding
dimension, and the corresponding MLP checkpoint filenames distributed
via piggyback. Users wanting to plug in a different sentence-transformer
(e.g., one of the newer top entries on
\href{https://huggingface.co/spaces/mteb/leaderboard}{MTEB}) can bypass
this registry by passing a list directly to \code{ebisg.model} with their
own \code{transformer}, \code{dim}, and \code{surname_mlp} (path to a local
.pt checkpoint). See \code{resolve_ebisg_model()} for the schema.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/census_data_preflight.R
\name{census_data_preflight}
\alias{census_data_preflight}
\title{Preflight census data}
\usage{
census_data_preflight(census.data, census.geo, year)
}
\arguments{
\item{census.data}{A list indexed by two-letter state abbreviations,
which contains pre-saved Census geographic data.
Can be generated using \code{get_census_data} function.}

\item{census.geo}{An optional character vector specifying what level of
geography to use to merge in U.S. Census geographic data. Currently
\code{"county"}, \code{"tract"}, \code{"block_group"}, \code{"block"}, and \code{"place"}
are supported.
Note: sufficient information must be in user-defined \code{\var{voter.file}} object.
If \code{\var{census.geo} = "county"}, then \code{\var{voter.file}}
must have column named \code{county}.
If \code{\var{census.geo} = "tract"}, then \code{\var{voter.file}}
must have columns named \code{county} and \code{tract}.
And if \code{\var{census.geo} = "block"}, then \code{\var{voter.file}}
must have columns named \code{county}, \code{tract}, and \code{block}.
If \code{\var{census.geo} = "place"}, then \code{\var{voter.file}}
must have column named \code{place}.
If \code{census.geo = "zcta"}, then \code{voter.file} must have column named \code{zcta}.
Specifying \code{\var{census.geo}} will call \code{census_helper} function
to merge Census geographic data at specified level of geography.}

\item{year}{An optional character vector specifying the year of U.S. Census geographic
data to be downloaded. Use \code{"2010"}, or \code{"2020"}. Default is \code{"2020"}.}
}
\description{
Preflight census data
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/surnames2000.R
\docType{data}
\name{surnames2000}
\alias{surnames2000}
\title{Census Surname List (2000).}
\format{
A data frame with 157,728 rows and 6 variables:
\describe{
\item{surname}{Surname}
\item{p_whi}{Pr(White | Surname)}
\item{p_bla}{Pr(Black | Surname)}
\item{p_his}{Pr(Hispanic/Latino | Surname)}
\item{p_asi}{Pr(Asian/Pacific Islander | Surname)}
\item{p_oth}{Pr(Other | Surname)}
#' }
}
\usage{
surnames2000
}
\description{
Census Surname List from 2000 with race/ethnicity probabilities by surname.
}
\examples{
data(surnames2000)
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/name_preproc.R
\name{.name_preproc}
\alias{.name_preproc}
\title{Pre-process vector of names to match census style. Internal function}
\usage{
.name_preproc(voter_names, target_names)
}
\arguments{
\item{voter_names}{Character vector to be pre-processed.}

\item{target_names}{Character vector of census names to be matched.}
}
\value{
A character vector of pre-processed named
}
\description{
Pre-process vector of names to match census style. Internal function
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merge_names.R
\name{wru_data_preflight}
\alias{wru_data_preflight}
\title{Preflight for name data}
\usage{
wru_data_preflight()
}
\description{
Checks if namedata is available in the current working directory, if not
downloads it from github using piggyback. By default, wru will download the
data to a temporary directory that lasts as long as your session does.
However, you may wish to set the \code{wru_data_wd} option to save the
downloaded data to your current working directory for more permanence.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_state_fips.R
\name{as_fips_code}
\alias{as_fips_code}
\alias{as_state_abbreviation}
\title{Convert between state names, postal abbreviations, and FIPS codes}
\usage{
as_fips_code(x)

as_state_abbreviation(x)
}
\arguments{
\item{x}{A \link{numeric} or \link{character} vector of state names,
postal abbreviations, or FIPS codes.
Matches for state names and abbreviations are not case sensitive.
FIPS codes may be matched from numeric or character vectors,
with or without leading zeroes.}
}
\value{
\describe{
\item{\code{as_state_fips_code()}}{
A \link{character} vector of two-digit FIPS codes.
One-digit FIPS codes are prefixed with a leading zero,
e.g., \code{"06"} for California.
}
\item{\code{as_state_abbreviation()}}{
A \link{character} vector of two-letter postal abbreviations,
e.g., \code{"CA"} for California.
}
}
}
\description{
Convert between state names, postal abbreviations, and FIPS codes
}
\examples{
as_fips_code("california")
as_state_abbreviation("california")

# Character vector matches ignore case
as_fips_code(c("DC", "Md", "va"))
as_state_abbreviation(c("district of columbia", "Maryland", "VIRGINIA"))

# Note that `3` and `7` are standardized to `NA`,
# because no state is assigned those FIPS codes
as_fips_code(1:10)
as_state_abbreviation(1:10)

# You can even mix methods in the same vector
as_fips_code(c("utah", "NM", 8, "04"))
as_state_abbreviation(c("utah", "NM", 8, "04"))

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_census_data.R
\name{get_census_data}
\alias{get_census_data}
\title{Multilevel Census data download function.}
\usage{
get_census_data(
  key = Sys.getenv("CENSUS_API_KEY"),
  states,
  age = FALSE,
  sex = FALSE,
  year = "2020",
  census.geo = c("tract", "block", "block_group", "county", "place", "zcta"),
  retry = 3,
  county.list = NULL
)
}
\arguments{
\item{key}{A character string containing a valid Census API key,
which can be requested from the
\href{https://api.census.gov/data/key_signup.html}{U.S. Census API key signup page}.

By default, attempts to find a census key stored in an
\link[=Sys.getenv]{environment variable} named \code{CENSUS_API_KEY}.}

\item{states}{which states to extract Census data for, e.g., \code{c("NJ", "NY")}.}

\item{age}{A \code{TRUE}/\code{FALSE} object indicating whether to condition on
age or not. If \code{FALSE} (default), function will return Pr(Geolocation | Race).
If \code{TRUE}, function will return Pr(Geolocation, Age | Race).
If \code{\var{sex}} is also \code{TRUE}, function will return Pr(Geolocation, Age, Sex | Race).}

\item{sex}{A \code{TRUE}/\code{FALSE} object indicating whether to condition on
sex or not. If \code{FALSE} (default), function will return Pr(Geolocation | Race).
If \code{TRUE}, function will return Pr(Geolocation, Sex | Race).
If \code{\var{age}} is also \code{TRUE}, function will return Pr(Geolocation, Age, Sex | Race).}

\item{year}{A character object specifying the year of U.S. Census data to be downloaded.
Use \code{"2010"}, or \code{"2020"}. Default is \code{"2020"}.
Warning: 2020 U.S. Census data is downloaded only when \code{\var{age}} and
\code{\var{sex}} are both \code{FALSE}.}

\item{census.geo}{An optional character vector specifying what level of
geography to use to merge in U.S. Census 2010 geographic data. Currently
\code{"county"}, \code{"tract"}, \code{"block"}, and \code{"place"} are supported.}

\item{retry}{The number of retries at the census website if network interruption occurs.}

\item{county.list}{A named list of character vectors of counties present in your \var{voter.file}, per state.}
}
\value{
Output will be an object of class \code{list} indexed by state.
Output will contain a subset of the following elements:
\code{state}, \code{age}, \code{sex},
\code{county}, \code{tract}, \code{block_group}, \code{block}, and \code{place}.
}
\description{
\code{get_census_data} returns county-, tract-, and block-level Census data
for specified state(s). Using this function to download Census data in advance
can save considerable time when running \code{predict_race} and \code{census_helper}.
}
\examples{
\dontrun{get_census_data(states = c("NJ", "NY"), age = TRUE, sex = FALSE)}
\dontrun{get_census_data(states = "MN", age = FALSE, sex = FALSE, year = "2020")}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ebisg_utils.R
\name{resolve_ebisg_model}
\alias{resolve_ebisg_model}
\title{Resolve an ebisg.model argument into a config list}
\usage{
resolve_ebisg_model(ebisg.model)
}
\arguments{
\item{ebisg.model}{A character string giving the HuggingFace model ID
of a built-in model (e.g., \code{"intfloat/multilingual-e5-large"}),
or a named list with elements \code{transformer}, \code{dim},
\code{surname_mlp}, and optionally \code{firstname_mlp}.}
}
\value{
A list with elements \code{transformer}, \code{dim},
\code{surname_mlp}, \code{firstname_mlp}, \code{fullname_mlp}, \code{tag}.
}
\description{
Resolve an ebisg.model argument into a config list
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/census_helper_v2.R
\name{census_helper_new}
\alias{census_helper_new}
\title{Census helper function.}
\usage{
census_helper_new(
  key = Sys.getenv("CENSUS_API_KEY"),
  voter.file,
  states = "all",
  geo = c("tract", "block", "block_group", "county", "place", "zcta"),
  age = FALSE,
  sex = FALSE,
  year = "2020",
  census.data = NULL,
  retry = 3,
  use.counties = FALSE,
  skip_bad_geos = FALSE
)
}
\arguments{
\item{key}{A character string containing a valid Census API key,
which can be requested from the
\href{https://api.census.gov/data/key_signup.html}{U.S. Census API key signup page}.

By default, attempts to find a census key stored in an
\link[=Sys.getenv]{environment variable} named \code{CENSUS_API_KEY}.}

\item{voter.file}{An object of class \code{data.frame}. Must contain field(s) named
\code{\var{county}}, \code{\var{tract}}, \code{\var{block}}, and/or \code{\var{place}}
specifying geolocation. These should be character variables that match up with
U.S. Census categories. County should be three characters (e.g., "031" not "31"),
tract should be six characters, and block should be four characters.
Place should be five characters if it is included.}

\item{states}{A character vector specifying which states to extract
Census data for, e.g. \code{c("NJ", "NY")}. Default is \code{"all"}, which extracts
Census data for all states contained in user-input data.}

\item{geo}{A character object specifying what aggregation level to use.
Use \code{"county"}, \code{"tract"}, \code{"block"}, or \code{"place"}.
Default is \code{"tract"}. Warning: extracting block-level data takes very long.}

\item{age}{A \code{TRUE}/\code{FALSE} object indicating whether to condition on
age or not. If \code{FALSE} (default), function will return Pr(Geolocation | Race).
If \code{TRUE}, function will return Pr(Geolocation, Age | Race).
If \code{\var{sex}} is also \code{TRUE}, function will return Pr(Geolocation, Age, Sex | Race).}

\item{sex}{A \code{TRUE}/\code{FALSE} object indicating whether to condition on
sex or not. If \code{FALSE} (default), function will return Pr(Geolocation | Race).
If \code{TRUE}, function will return Pr(Geolocation, Sex | Race).
If \code{\var{age}} is also \code{TRUE}, function will return Pr(Geolocation, Age, Sex | Race).}

\item{year}{A character object specifying the year of U.S. Census data to be downloaded.
Use \code{"2010"}, or \code{"2020"}. Default is \code{"2020"}.}

\item{census.data}{A optional census object of class \code{list} containing
pre-saved Census geographic data. Can be created using \code{get_census_data} function.
If \code{\var{census.data}} is provided, the \code{\var{year}} element must
have the same value as the \code{\var{year}} option specified in this function
(i.e., \code{"2010"} in both or \code{"2020"} in both).
If \code{\var{census.data}} is provided, the \code{\var{age}} and the \code{\var{sex}}
elements must be \code{FALSE}. This corresponds to the defaults of \code{census_geo_api}.
If \code{\var{census.data}} is missing, Census geographic data will be obtained via Census API.}

\item{retry}{The number of retries at the census website if network interruption occurs.}

\item{use.counties}{A logical, defaulting to FALSE. Should census data be filtered by counties
available in \var{census.data}?}

\item{skip_bad_geos}{Logical. Option to have the function skip any geolocations that are not present
in the census data, returning a partial data set. Default is set to \code{FALSE}, which case it will
break and provide error message with a list of offending geolocations.}
}
\value{
Output will be an object of class \code{data.frame}. It will
consist of the original user-input data with additional columns of
Census data.
}
\description{
\code{census_helper_new} links user-input dataset with Census geographic data.
}
\details{
This function allows users to link their geocoded dataset (e.g., voter file)
with U.S. Census data (2010 or 2020). The function extracts Census Summary File data
at the county, tract, block, or place level. Census data calculated are
Pr(Geolocation | Race) where geolocation is county, tract, block, or place.
}
\examples{
\dontshow{data(voters)}
\dontrun{census_helper_new(voter.file = voters, states = "nj", geo = "block")}
\dontrun{census_helper_new(voter.file = voters, states = "all", geo = "tract")}
\dontrun{census_helper_new(voter.file = voters, states = "all", geo = "place",
 year = "2020")}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/voters.R
\docType{data}
\name{voters}
\alias{voters}
\title{Example voter file.}
\format{
A data frame with 10 rows and 12 variables:
\describe{
\item{VoterID}{Voter identifier (numeric)}
\item{surname}{Surname}
\item{state}{State of residence}
\item{CD}{Congressional district}
\item{county}{Census county (three-digit code)}
\item{first}{First name}
\item{last}{Last name or surname}
\item{tract}{Census tract (six-digit code)}
\item{block}{Census block (four-digit code)}
\item{precinct}{Voting precinct}
\item{place}{Voting place}
\item{age}{Age in years}
\item{sex}{0=male, 1=female}
\item{party}{Party registration (character)}
\item{PID}{Party registration (numeric)}
#' }
}
\usage{
voters
}
\description{
An example dataset containing voter file information.
}
\examples{
data(voters)
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ebisg_utils.R
\name{map_6class_to_5class}
\alias{map_6class_to_5class}
\title{Map 6-class eBISG output to wru's 5-class system}
\usage{
map_6class_to_5class(probs_6)
}
\arguments{
\item{probs_6}{A matrix with 6 columns named whi, bla, his, asi, aian, oth.}
}
\value{
A matrix with 5 columns named c_whi, c_bla, c_his, c_asi, c_oth.
}
\description{
The MLP outputs probabilities for 6 categories:
(whi, bla, his, asi, aian, oth) -- see \code{EBISG_RACE_COLS} in the
Python helper. wru uses 5 categories: (whi, bla, his, asi, oth).
This function sums \code{aian} and \code{oth} into \code{oth}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/census_geo_api.R
\name{census_geo_api}
\alias{census_geo_api}
\title{Census Data download function.}
\usage{
census_geo_api(
  key = Sys.getenv("CENSUS_API_KEY"),
  state,
  geo = c("tract", "block", "block_group", "county", "place", "zcta"),
  age = FALSE,
  sex = FALSE,
  year = c("2020", "2010"),
  retry = 3,
  save_temp = NULL,
  counties = NULL
)
}
\arguments{
\item{key}{A character string containing a valid Census API key,
which can be requested from the
\href{https://api.census.gov/data/key_signup.html}{U.S. Census API key signup page}.

By default, attempts to find a census key stored in an
\link[=Sys.getenv]{environment variable} named \code{CENSUS_API_KEY}.}

\item{state}{A required character object specifying which state to extract Census data for,
e.g., \code{"NJ"}.}

\item{geo}{A character object specifying what aggregation level to use.
Use \code{"block"}, \code{"block_group"}, \code{"county"}, \code{"place"}, \code{"tract"}, or \code{"zcta"}.
Default is \code{"tract"}. Warning: extracting block-level data takes very long.}

\item{age}{A \code{TRUE}/\code{FALSE} object indicating whether to condition on
age or not. If \code{FALSE} (default), function will return Pr(Geolocation | Race).
If \code{TRUE}, function will return Pr(Geolocation, Age | Race).
If \code{\var{sex}} is also \code{TRUE}, function will return Pr(Geolocation, Age, Sex | Race).}

\item{sex}{A \code{TRUE}/\code{FALSE} object indicating whether to condition on
sex or not. If \code{FALSE} (default), function will return Pr(Geolocation | Race).
If \code{TRUE}, function will return Pr(Geolocation, Sex | Race).
If \code{\var{age}} is also \code{TRUE}, function will return Pr(Geolocation, Age, Sex | Race).}

\item{year}{A character object specifying the year of U.S. Census data to be downloaded.
Use \code{"2010"}, or \code{"2020"}. Default is \code{"2020"}.
Warning: 2020 U.S. Census data is downloaded only when \code{\var{age}} and
\code{\var{sex}} are both \code{FALSE}.}

\item{retry}{The number of retries at the census website if network interruption occurs.}

\item{save_temp}{File indicating where to save the temporary outputs.
Defaults to NULL. If specified, the function will look for an .RData file
with the same format as the expected output.}

\item{counties}{A vector of counties contained in your data. If \code{NULL}, all counties are pulled.
Useful for smaller predictions where only a few counties are considered. Must be zero padded.}
}
\value{
Output will be an object of class \code{list}, indexed by state names. It will
consist of the original user-input data with additional columns of Census geographic data.
}
\description{
\code{census_geo_api} retrieves U.S. Census geographic data for a given state.
}
\details{
This function allows users to download U.S. Census geographic data (2010 or 2020),
at either the county, tract, block, or place level, for a particular state.
}
\examples{
\dontshow{data(voters)}
\dontrun{census_geo_api(states = c("NJ", "DE"), geo = "block")}
\dontrun{census_geo_api(states = "FL", geo = "tract", age = TRUE, sex = TRUE)}
\dontrun{census_geo_api(states = "MA", geo = "place", age = FALSE, sex = FALSE,
 year = "2020")}

}
\references{
Relies on \code{get_census_api()}, \code{get_census_api_2()}, and \code{vec_to_chunk()} functions authored by Nicholas Nagle,
available \href{https://rstudio-pubs-static.s3.amazonaws.com/19337_2e7f827190514c569ea136db788ce850.html}{here}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ebisg_utils.R
\name{get_ebisg_module}
\alias{get_ebisg_module}
\title{Load the Python eBISG helper module}
\usage{
get_ebisg_module()
}
\description{
Imports \code{inst/python/ebisg_helper.py} as a private module via
\code{reticulate::import_from_path()} and caches the module object on
\code{.ebisg_env}. Using \code{import_from_path} keeps the helper's symbols
encapsulated rather than dumping them into Python's global namespace
as \code{source_python} would.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_census_api.R
\name{get_census_api}
\alias{get_census_api}
\title{Census API function.}
\usage{
get_census_api(
  data_url,
  key = Sys.getenv("CENSUS_API_KEY"),
  var.names,
  region,
  retry = 0
)
}
\arguments{
\item{data_url}{URL root of the API,
e.g., \code{"https://api.census.gov/data/2020/dec/pl"}.}

\item{key}{A character string containing a valid Census API key,
which can be requested from the
\href{https://api.census.gov/data/key_signup.html}{U.S. Census API key signup page}.

By default, attempts to find a census key stored in an
\link[=Sys.getenv]{environment variable} named \code{CENSUS_API_KEY}.}

\item{var.names}{A character vector of variables to get,
e.g., \code{c("P2_005N", "P2_006N", "P2_007N", "P2_008N")}.
If there are more than 50 variables, then function will automatically
split variables into separate queries.}

\item{region}{Character object specifying which region to obtain data for.
Must contain "for" and possibly "in",
e.g., \code{"for=block:1213&in=state:47+county:015+tract:*"}.}

\item{retry}{The number of retries at the census website if network interruption occurs.}
}
\value{
If successful, output will be an object of class \code{data.frame}.
If unsuccessful, function prints the URL query that caused the error.
}
\description{
\code{get_census_api} obtains U.S. Census data via the public API.
}
\details{
This function obtains U.S. Census data via the public API. User
can specify the variables and region(s) for which to obtain data.
}
\examples{
\dontrun{
get_census_api(
  data_url = "https://api.census.gov/data/2020/dec/pl",
  var.names = c("P2_005N", "P2_006N", "P2_007N", "P2_008N"), region = "for=county:*&in=state:34"
)
}

}
\references{
Based on code authored by Nicholas Nagle, which is available
\href{https://rstudio-pubs-static.s3.amazonaws.com/19337_2e7f827190514c569ea136db788ce850.html}{here}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/census_geo_api_names.R
\name{census_geo_api_names}
\alias{census_geo_api_names}
\alias{census_geo_api_url}
\title{Census geo API helper functions}
\usage{
census_geo_api_names(
  year = c("2020", "2010", "2000"),
  age = FALSE,
  sex = FALSE
)

census_geo_api_url(year = c("2020", "2010", "2000"))
}
\arguments{
\item{year}{A character object specifying the year of U.S. Census data to be downloaded.
Use \code{"2010"}, or \code{"2020"}. Default is \code{"2020"}.
Warning: 2020 U.S. Census data is downloaded only when \code{\var{age}} and
\code{\var{sex}} are both \code{FALSE}.}

\item{age}{A \code{TRUE}/\code{FALSE} object indicating whether to condition on
age or not. If \code{FALSE} (default), function will return Pr(Geolocation | Race).
If \code{TRUE}, function will return Pr(Geolocation, Age | Race).
If \code{\var{sex}} is also \code{TRUE}, function will return Pr(Geolocation, Age, Sex | Race).}

\item{sex}{A \code{TRUE}/\code{FALSE} object indicating whether to condition on
sex or not. If \code{FALSE} (default), function will return Pr(Geolocation | Race).
If \code{TRUE}, function will return Pr(Geolocation, Sex | Race).
If \code{\var{age}} is also \code{TRUE}, function will return Pr(Geolocation, Age, Sex | Race).}
}
\value{
\describe{
\item{\code{census_geo_api_names()}}{
A named list of \link{character} vectors whose values correspond to columns
of a Census API table and whose names represent the new columns they are
used to calculate in \code{\link[=census_geo_api]{census_geo_api()}}.
}
\item{\code{census_geo_api_url()}}{
A \link{character} string containing the base of the URL to a
Census API table.
}
}
}
\description{
Census geo API helper functions
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/format_legacy_data.R
\name{format_legacy_data}
\alias{format_legacy_data}
\title{Legacy data formatting function.}
\usage{
format_legacy_data(legacyFilePath, state, outFile = NULL)
}
\arguments{
\item{legacyFilePath}{A character vector giving the location of a legacy census data folder,
sourced from https://www2.census.gov/programs-surveys/decennial/2020/data/01-Redistricting_File--PL_94-171/.
These file names should end in ".pl".}

\item{state}{The two letter state postal code.}

\item{outFile}{Optional character vector determining whether the formatted RData object should be saved. The
filepath should end in ".RData".}
}
\description{
\code{format_legacy_data} formats legacy data from the U.S. census to allow
for Bayesian name geocoding.
}
\details{
This function allows users to construct datasets for analysis using the census legacy data format.
These data are available for the 2020 census at
https://www2.census.gov/programs-surveys/decennial/2020/data/01-Redistricting_File--PL_94-171/.
This function returns data structured analogously to data from the Census API, which is not yet
available for the 2020 Census as of September 2021.
}
\examples{
\dontrun{
gaCensusData <- format_legacy_data(PL94171::pl_url('GA', 2020))
predict_race_new(ga.voter.file, namesToUse = 'last, first, mid', census.geo = 'block',
     census.data = gaCensusData)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ebisg_utils.R
\name{ensure_ebisg_python}
\alias{ensure_ebisg_python}
\title{Check that Python dependencies for eBISG are available}
\usage{
ensure_ebisg_python()
}
\description{
Check that Python dependencies for eBISG are available
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predict_race.R
\name{predict_race}
\alias{predict_race}
\title{Race prediction function.}
\usage{
predict_race(
  voter.file,
  census.surname = TRUE,
  surname.only = FALSE,
  census.geo = c("tract", "block", "block_group", "county", "place", "zcta"),
  census.key = Sys.getenv("CENSUS_API_KEY"),
  census.data = NULL,
  age = FALSE,
  sex = FALSE,
  year = "2020",
  party = NULL,
  retry = 3,
  impute.missing = TRUE,
  skip_bad_geos = FALSE,
  use.counties = FALSE,
  model = "BISG",
  race.init = NULL,
  name.dictionaries = NULL,
  names.to.use = "surname",
  control = NULL,
  ebisg.model = "intfloat/multilingual-e5-large"
)
}
\arguments{
\item{voter.file}{An object of class \code{data.frame}.
Must contain a row for each individual being predicted,
as well as a field named \code{\var{surname}} containing each individual's surname.
If using geolocation in predictions, \code{\var{voter.file}} must contain a field named
\code{\var{state}}, which contains the two-character abbreviation for each individual's
state of residence (e.g., \code{"nj"} for New Jersey).
If using Census geographic data in race/ethnicity predictions,
\code{\var{voter.file}} must also contain at least one of the following fields:
\code{\var{county}}, \code{\var{tract}}, \code{\var{block_group}}, \code{\var{block}},
and/or \code{\var{place}}.
These fields should contain character strings matching U.S. Census categories.
County is three characters (e.g., \code{"031"} not \code{"31"}),
tract is six characters, block group is usually a single character and block
is four characters. Place is five characters.
See below for other optional fields.}

\item{census.surname}{A \code{TRUE}/\code{FALSE} object. If \code{TRUE},
function will call \code{merge_surnames} to merge in Pr(Race | Surname)
from U.S. Census Surname List (2000, 2010, or 2020) and Spanish Surname List.
If \code{FALSE}, user must provide a \code{name.dictionary} (see below).
Default is \code{TRUE}.}

\item{surname.only}{A \code{TRUE}/\code{FALSE} object. If \code{TRUE}, race predictions will
only use surname data and calculate Pr(Race | Surname). Default is \code{FALSE}.}

\item{census.geo}{An optional character vector specifying what level of
geography to use to merge in U.S. Census geographic data. Currently
\code{"county"}, \code{"tract"}, \code{"block_group"}, \code{"block"}, and \code{"place"}
are supported.
Note: sufficient information must be in user-defined \code{\var{voter.file}} object.
If \code{\var{census.geo} = "county"}, then \code{\var{voter.file}}
must have column named \code{county}.
If \code{\var{census.geo} = "tract"}, then \code{\var{voter.file}}
must have columns named \code{county} and \code{tract}.
And if \code{\var{census.geo} = "block"}, then \code{\var{voter.file}}
must have columns named \code{county}, \code{tract}, and \code{block}.
If \code{\var{census.geo} = "place"}, then \code{\var{voter.file}}
must have column named \code{place}.
If \code{census.geo = "zcta"}, then \code{voter.file} must have column named \code{zcta}.
Specifying \code{\var{census.geo}} will call \code{census_helper} function
to merge Census geographic data at specified level of geography.}

\item{census.key}{A character object specifying user's Census API key.
Required if \code{census.geo} is specified, because a valid Census API key is
required to download Census geographic data.

If \code{\link{NULL}}, the default, attempts to find a census key stored in an
\link[=Sys.getenv]{environment variable} named \code{CENSUS_API_KEY}.}

\item{census.data}{A list indexed by two-letter state abbreviations,
which contains pre-saved Census geographic data.
Can be generated using \code{get_census_data} function.}

\item{age}{An optional \code{TRUE}/\code{FALSE} object specifying whether to
condition race predictions on age (in addition to surname and geolocation).
Default is \code{FALSE}. Must be same as \code{\var{age}} in \code{\var{census.data}} object.
May only be set to \code{TRUE} if \code{census.geo} option is specified.
If \code{TRUE}, \code{\var{voter.file}} should include a numerical variable \code{\var{age}}.}

\item{sex}{optional \code{TRUE}/\code{FALSE} object specifying whether to
condition race predictions on sex (in addition to surname and geolocation).
Default is \code{FALSE}. Must be same as \code{\var{sex}} in \code{\var{census.data}} object.
May only be set to \code{TRUE} if \code{census.geo} option is specified.
If \code{TRUE}, \code{\var{voter.file}} should include a numerical variable \code{\var{sex}},
where \code{\var{sex}} is coded as 0 for males and 1 for females.}

\item{year}{An optional character vector specifying the year of U.S. Census geographic
data to be downloaded. Use \code{"2010"}, or \code{"2020"}. Default is \code{"2020"}.}

\item{party}{An optional character object specifying party registration field
in \code{\var{voter.file}}, e.g., \code{\var{party} = "PartyReg"}.
If specified, race/ethnicity predictions will be conditioned
on individual's party registration (in addition to geolocation).
Whatever the name of the party registration field in \code{\var{voter.file}},
it should be coded as 1 for Democrat, 2 for Republican, and 0 for Other.}

\item{retry}{The number of retries at the census website if network interruption occurs.}

\item{impute.missing}{Logical, defaults to TRUE. Should missing be imputed?}

\item{skip_bad_geos}{Logical. Option to have the function skip any geolocations that are not present
in the census data, returning a partial data set. Default is set to \code{FALSE}, in which case it
will break and provide error message with a list of offending geolocations.}

\item{use.counties}{A logical, defaulting to FALSE. Should census data be filtered by counties
available in \var{census.data}?}

\item{model}{Character string: "BISG" (default), "fBISG" (fully-Bayesian with
error correction), or "eBISG" (embedding-supplemented BISG, which uses
text embeddings to predict race probabilities for names not found
in Census surname lists). The eBISG model requires Python with
sentence-transformers and torch; run \code{\link{setup_ebisg}} to configure.}

\item{race.init}{Vector of initial race for each observation in voter.file.
Must be an integer vector, with 1=white, 2=black, 3=hispanic, 4=asian, and
5=other. Defaults to values obtained using \code{model="BISG_surname"}.}

\item{name.dictionaries}{Optional named list of \code{data.frame}'s
containing counts of names by race. Any of the following named elements
are allowed: "surname", "first", "middle". When present, the objects must
follow the same structure as \code{last_c}, \code{first_c},
\code{mid_c}, respectively.}

\item{names.to.use}{One of 'surname', 'surname, first', or  'surname, first,
middle'. Defaults to 'surname'.}

\item{control}{List of control arguments only used when \code{model="fBISG"}, including
\describe{
\item{iter}{Number of MCMC iterations. Defaults to 1000.}
\item{burnin}{Number of iterations discarded as burnin. Defaults to half of \code{iter}.}
\item{verbose}{Print progress information. Defaults to \code{TRUE}.}
\item{me.correct}{Boolean. Should the model correct measurement error for \code{races|geo}? Defaults to \code{TRUE}.}
\item{seed}{RNG seed. If \code{NULL}, a seed is generated and returned as an attribute for reproducibility.}
}}

\item{ebisg.model}{Character string (HuggingFace model ID) or named list
specifying which embedding model to use when \code{model = "eBISG"}.
The only built-in option is \code{"intfloat/multilingual-e5-large"}
(default; 1024-dim), which is keyed by its full HuggingFace ID for
consistency with the Python side. To use a different sentence-transformer,
pass a named list with elements \code{transformer} (HuggingFace model ID),
\code{dim} (embedding dimension), and \code{surname_mlp},
\code{firstname_mlp} (paths to custom .pt checkpoints).}
}
\value{
Output will be an object of class \code{data.frame}. It will
consist of the original user-input \code{voter.file} with additional columns with
predicted probabilities for each of the five major racial categories:
\code{\var{pred.whi}} for White,
\code{\var{pred.bla}} for Black,
\code{\var{pred.his}} for Hispanic/Latino,
\code{\var{pred.asi}} for Asian/Pacific Islander, and
\code{\var{pred.oth}} for Other/Mixed.
}
\description{
\code{predict_race} makes probabilistic estimates of individual-level race/ethnicity.
}
\details{
This function implements the Bayesian race prediction methods outlined in
Imai and Khanna (2015). The function produces probabilistic estimates of
individual-level race/ethnicity, based on surname, geolocation, and party.
}
\examples{
\donttest{
#' data(voters)
try(predict_race(voter.file = voters, surname.only = TRUE))
\dontrun{
try(predict_race(voter.file = voters, census.geo = "tract"))
}
\dontrun{
try(predict_race(
  voter.file = voters, census.geo = "place", year = "2020"))
}
\dontrun{
CensusObj <- try(get_census_data(state = c("NY", "DC", "NJ")))
try(predict_race(
  voter.file = voters, census.geo = "tract", census.data = CensusObj, party = "PID")
  )
}
\dontrun{
CensusObj2 <- try(get_census_data(state = c("NY", "DC", "NJ"), age = T, sex = T))
try(predict_race(
  voter.file = voters, census.geo = "tract", census.data = CensusObj2, age = T, sex = T))
}
\dontrun{
CensusObj3 <- try(get_census_data(state = c("NY", "DC", "NJ"), census.geo = "place"))
try(predict_race(voter.file = voters, census.geo = "place", census.data = CensusObj3))
}
}
}
